package jmel;

import jmel.util.*;
import jmel.code.*;
import jmel.comp.*;
import jmel.code.Symbol.*;
import jmel.parser.*;
import jmel.tree.*;
import jmel.tree.Tree.*;
import jmel.util.*;
import java.io.*;
import java.lang.reflect.Method;

/**
 * This class could be the main entry point for GJC when GJC is used as a
 *  component in a larger software system. It provides operations to
 *  construct a new compiler, and to run a new compiler on a set of source
 *  files.
 */
public class JMelCompiler implements ClassReader.SourceCompleter {
	
	/**
	* The log to be used for error reporting.
	*/
	public Log log;

	/**
	* The symbol table module.
	*/
	public Symtab syms;
	
	/**
	* The module for attribution checking.
	*/
	public Check chk;
	
	/**
	* The module for identifier resolution.
	*/
	public Resolve rs;
	
	/**
	* The tree factory module.
	*/
	public TreeMaker make;
	
	/**
	* The module for the symbol table entry phases.
	*/
	public Enter enter;
	
	/**
	* The module for attribution.
	*/
	public Attr attr;
	
	/**
	* The module for code generation.
	*/
	public Gen gen;
	
	private static boolean isFullOptim = false;
	/**
	* Construct a new interpreter from a log and a options table.
	*/
	public JMelCompiler(Log log, Symtab syms, Hashtable options) {
		this.log = log;
		this.syms = syms;
		this.syms.reader.sourceCompleter = this;
		this.chk = new Check(log, syms, options);
		this.rs = new Resolve(log, syms, chk);
		this.make = new TreeMaker();
		this.enter = new Enter(log, syms, rs, chk, make, null, todo);
		this.attr = new Attr(log, syms, rs, chk, make, this.enter, options);
		this.gen = new Gen(log, syms, chk, rs, make, options);
		deprecation = options.get("-deprecation") != null;
		warnunchecked = options.get("-warnunchecked") != null;
		verbose = options.get("-verbose") != null;
		optim1 = options.get("-O1") != null;
		optim2 = options.get("-O2") != null;
		if (optim2) isFullOptim = true;
		optim1 |= optim2;
 	}

	/**
	* Construct a new compiler from a log and an options table.
	*  This will create a new symbol table module.
	*/
	public static JMelCompiler make(Hashtable options) {
		return make(new Log(), options);
	}
	public static JMelCompiler make(Log log, Hashtable options) {
		try {
			if (isFullOptim) options.put("-O2", "");	
			options.put("-classpath", System.getProperty("java.class.path"));	
			return new JMelCompiler(log, new Symtab(new ClassReader(options), new ClassWriter(options)), options);
		} catch (CompletionFailure ex) {
			log.error(Position.NOPOS, ex.getMessage());
			return null;
		}
	}

	/**
	* Give detailed deprecation warnings.
	*/
	public boolean deprecation;
	
	/**
	* Give detailed unchecked warnings.
	*/
	public boolean warnunchecked;
	
	public boolean verbose;

	public boolean optim1, optim2;
	
	/**
	* A queue of all as yet unattributed classes.
	*/
	public ListBuffer todo = new ListBuffer();
	
	/**
	* The set of currently compiled inputfiles, needed to ensure
	*  we don't accidentally overwrite an input file when -s is set.
	*  initialized by `compile'.
	*/
	public Set inputFiles = Set.make();
	
	/**
	* The number of errors reported so far.
	*/
	public int errorCount() {
		return log.nerrors;
	}

	/**
	* Try to open input stream with given name.
	*  Report an error if this fails.
	*  @param filename   The file name of the input stream to be opened.
	*/
	public InputStream openSource(String filename) {
		try {
			File f = new File(filename);
			inputFiles.put(f);
			return new FileInputStream(f);
		} catch (IOException e) {
			log.error(Position.NOPOS, "cant.read.file", filename);
			return null;
		}
	}
	
	/**
	* Parse contents of input stream.
	*  @param filename     The name of the file from which input stream comes.
	*  @param input        The input stream to be parsed.
	*/
	public TopLevel parse(String filename, InputStream input) {
		Name curr = Name.fromString(filename == null ? "<input>" : filename);
		Name prev = log.useSource(curr);
		long msec = System.currentTimeMillis();
		TopLevel tree = make.TopLevel(null, Tree.emptyList);
		if (input != null) {
			printVerbose("parsing.started", filename);
			try {
				Scanner scanner = new Scanner(input, log, null);
				if (filename != null) input.close();
				Parser parser = new Parser(scanner, make, log);
				tree = parser.compilationUnit(filename == null);
				printVerbose("parsing.done", Long.toString(System.currentTimeMillis() - msec));
			} catch (IOException e) {
				log.error(Position.NOPOS, "error.reading.file", filename, e.toString());
			}
		}
		if (Boolean.getBoolean("debug")) {
			tree.visit(new Pretty()); 
			tree.visit(new Core()); 
		}
		tree.sourcefile = curr;
		log.useSource(prev);
		return tree;
	}
	
	public TopLevel parse(String filename, String source) {
		InputStream input = new StringBufferInputStream(source);
		MelTranslator transl = new MelTranslator(optim2);
		TopLevel topLevel = parse(null, input);
		topLevel.sourcefile = Name.fromString(filename);
		if (log.nerrors != 0) return null;
		topLevel.visit(transl);
		return (TopLevel)transl.toplevels.toList().head;
	}
	
    /**
	* Main method: compile a list of files, return all compiled classes
	*  @param filenames     The names of all files to be compiled.
	*/
	public TopLevel compile(String filename, InputStream input) throws Error {
		long msec = System.currentTimeMillis();
		TopLevel myclass = null;
		try { myclass = parse(filename, input); } 
		catch (Abort ex) {}
		printVerbose("total", Long.toString(System.currentTimeMillis() - msec));
		return myclass;
	}
	public List compile(List filenames) throws Error {
		long msec = System.currentTimeMillis();
		ListBuffer classes = new ListBuffer();
		TopLevel myclass = null;
		try {
			inputFiles.reset();
			ListBuffer trees = new ListBuffer();
			for (List l = filenames; l.nonEmpty(); l = l.tail) {
				String filename = (String) l.head;
				Name prev = log.useSource(Name.fromString(filename));
				myclass = parse( filename, openSource(filename) );
				if (log.nerrors == 0) classes.append(myclass);
				log.useSource(prev);
			}
		} catch (Abort ex) {
		}
		printVerbose("total", Long.toString(System.currentTimeMillis() - msec));
		int errCount = errorCount();
		if (errCount == 1)
			printCount("error", errCount);
		else
			printCount("error.plural", errCount);
		if (log.nwarnings == 1)
			printCount("warn", log.nwarnings);
		else
			printCount("warn.plural", log.nwarnings);
		if (optim1) return generate(classes.toList());
		return classes.toList();
	}
	
	public List generate(List cl) throws Error {
		
		ListBuffer roots = new ListBuffer();
		MelTranslator transl = new MelTranslator(optim2);
		for (List l = cl; l.nonEmpty(); l = l.tail) {
			TopLevel topLevel = (TopLevel) l.head;
			topLevel.visit(transl);
			roots.appendList(transl.toplevels.toList());		
		}
		long msec = System.currentTimeMillis();
		ListBuffer classes = new ListBuffer();
		try {
			inputFiles.reset();
			Symbol.reset();
			chk.compiled.reset();
			enter.main(roots.toList());
			while (todo.nonEmpty()) {
				Env env = (Env) todo.first();
				todo.remove();
				Tree untranslated = env.tree;
				printVerbose("checking.attribution", env.enclClass.sym.toJava());
				Name prev = log.useSource(env.enclClass.sym.sourcefile);
				attr.attribClass(env.enclClass.sym);
				if (errorCount() == 0) {
					new Flow(log, syms, chk).analyzeDef(env.tree);
				}
				TreeMaker make = new TreeMaker(env.toplevel);
				if (errorCount() == 0) {
					env.tree = new TransTypes(log, syms, make).translateTopLevelClass(env.tree);
				}
				if (errorCount() == 0) {
					ClassDef cdef = null;
					try {
						List cdefs = new TransInner(log, syms, rs, chk, make).translateTopLevelClass(env, env.tree);
						for (List l = cdefs; errorCount() == 0 && l.nonEmpty(); l = l.tail) {
							cdef = (ClassDef) l.head;
							genCode(env, cdef);
							classes.append(cdef.sym);
						}
					} catch (IOException ex) {
						log.error(cdef.pos, "class.cant.write", cdef.sym.toJava(), ex.getMessage());
					}
				}
				log.useSource(prev);
			}
		} catch (Abort ex) {
		}
		printVerbose("total", Long.toString(System.currentTimeMillis() - msec));
		int errCount = errorCount();
		if (errCount == 1)
			printCount("error", errCount);
		else
			printCount("error.plural", errCount);
		if (log.nwarnings == 1)
			printCount("warn", log.nwarnings);
		else
			printCount("warn.plural", log.nwarnings);
		return classes.toList();
	}
	
	public void generate(TopLevel topLevel) throws Error {
		
		long msec = System.currentTimeMillis();
		ListBuffer classes = new ListBuffer();
		try {
			inputFiles.reset();
			Symbol.reset();
			chk.compiled.reset();
			enter.main(Tree.emptyList.append(topLevel));
			while (todo.nonEmpty()) {
				Env env = (Env) todo.first();
				todo.remove();
				Tree untranslated = env.tree;
				printVerbose("checking.attribution", env.enclClass.sym.toJava());
				Name prev = log.useSource(env.enclClass.sym.sourcefile);
				attr.attribClass(env.enclClass.sym);
				if (errorCount() == 0) {
					new Flow(log, syms, chk).analyzeDef(env.tree);
				}
				TreeMaker make = new TreeMaker(env.toplevel);
				if (errorCount() == 0) {
					env.tree = new TransTypes(log, syms, make).translateTopLevelClass(env.tree);
				}
				if (errorCount() == 0) {
					ClassDef cdef = null;
					try {
						List cdefs = new TransInner(log, syms, rs, chk, make).translateTopLevelClass(env, env.tree);
						for (List l = cdefs; errorCount() == 0 && l.nonEmpty(); l = l.tail) {
							cdef = (ClassDef) l.head;
							genCode(env, cdef);
							classes.append(cdef.sym);
						}
					} catch (IOException ex) {
						log.error(cdef.pos, "class.cant.write", cdef.sym.toJava(), ex.getMessage());
					}
				}
				log.useSource(prev);
			}
		} catch (Abort ex) {
		}
		printVerbose("total", Long.toString(System.currentTimeMillis() - msec));
		int errCount = errorCount();
		if (errCount == 1)
			printCount("error", errCount);
		else
			printCount("error.plural", errCount);
		if (log.nwarnings == 1)
			printCount("warn", log.nwarnings);
		else
			printCount("warn.plural", log.nwarnings);
	}
	
	/**
	* Generate code and emit a class file for a given class
	*  @param env    The attribution environment of the outermost class
	*                containing this class.
	*  @param cdef   The class definition from which code is generated.
	*/
	public void genCode(Env env, ClassDef cdef) throws IOException {
		gen.genClass(env, cdef);
		writeClass(cdef.sym);
	}

	/**
	* Emit a class file for a given class.
	*  @param c      The class from which a class file is generated.
	*/
	public void writeClass(ClassSymbol c) throws IOException {
		File outFile = syms.writer.outputFile(c, ".class");
		OutputStream out = new FileOutputStream(outFile);
		try {
			syms.writer.writeClassFile(out, c);
			printVerbose("wrote.file", outFile.getPath());
		}
		finally { 
			out.close();
		} 
	}

	/**
      * Output for "-verbose" option.
      *  @param key The key to look up the correct internationalized string.
      *  @param arg An argument for substitution into the output string.
      */
    private void printVerbose(String key, String arg) {
        if (verbose) System.err.println(log.getLocalizedString("verbose." + key, arg));
    }

	/**
	* Print numbers of errors and warnings.
	*/
	void printCount(String kind, int count) {
		if (count != 0)
			log.println( log.getLocalizedString("count." + kind, Integer.toString(count)));
	}
	
    public static String version() {
        return System.getProperty("java.version");
    }

    public static String date() {
        return "DD-MMMMM-YY";
    }
    
    
	/**
	* Complete compiling a source file that has been accessed
	*  by the class file reader.
	*  @param c          The class the source file of which needs to be compiled.
	*  @param filename   The name of the source file.
	*  @param f          An input stream that reads the source file.
	*/
	public void complete(ClassSymbol c, String filename, InputStream f) throws CompletionFailure {
		Tree tree = parse(filename, f);
		enter.main(List.make(tree));
		if (c.members() == null) {
			throw syms.reader.new BadClassFile(c, filename,	log.getLocalizedString("file.doesnt.contain.class",	c.fullname.toJava()));
		}
	}
}
