package jmel;

import jmel.interp.*;
import jmel.script.Script;
import jmel.script.Commandable;
import jmel.script.Commands;
import jmel.tree.*;
import jmel.util.*;
import jmel.code.*;
import jmel.tree.*;
import jmel.parser.*;
import jmel.tree.Tree.*;
import java.util.Vector;
import java.io.*;


public class JMelInterpreter implements ConsoleInterface {
	/* --- Begin static stuff --- */
	/*
 	Debug utils are static so that they are reachable by code that doesn't		
 	necessarily have an interpreter reference (e.g. tracing in utils).
	In the future we may want to allow debug/trace to be turned on on
	a per interpreter basis, in which case we'll need to use the parent 
	reference in some way to determine the scope of the command that
	turns it on or off...
	*/
	public static boolean DEBUG, TRACE;

	// This should be per instance
	static PrintStream debug;
	static {
 		staticInit();
	}

	/** Strict Java mode */
	// --- end static stuff --- 
	
	// --- Instance data --- 
	
	InputStream in;
	PrintStream out, err;
	NameSpace globalNameSpace;
    ConsoleInterface console; 

 	// If this interpeter is a child of another, the parent 
	JMelInterpreter parent;

	// The source compiler
	JMelCompiler compiler;
	
	// The name of the file or other source that this interpreter is reading
	String sourceFileInfo;
	
	//	Do we override exit on EOF as normally done in iteractive mode?
	// (This is used by Sessiond)
	public boolean noExitOnEOF;
	
	private boolean
 		evalOnly, 		// Interpreter has no input stream, use eval() only
		interactive;	// Interpreter has a user, print prompts, etc.

	// --- End instance data ---
	
	/*		The main constructor.	
	All constructors should now pass through here.
	@param namespace If namespace is non-null then this interpreter's
 		root namespace will be set to the one provided.  If it is null a new
 		one will be created for it.		
 	@param parent The parent interpreter if this interpreter is a child
 			of another.  May be null.
	@param sourceFileInfo An informative string holding the filename
 		or other description of the source from which this interpreter is
		reading... used for debugging.  May be null.
	*/ 

	public JMelInterpreter(
		InputStream in, PrintStream out, PrintStream err,
 		boolean interactive, NameSpace namespace,
		JMelInterpreter parent, String sourceFileInfo )
	{
		long t1=System.currentTimeMillis();
        this.in = in;
        this.out = out;
        Commands.out = out;
        this.err = err;
        this.interactive = interactive;
		debug = err;
		this.parent = parent;
		this.sourceFileInfo = sourceFileInfo;
		
		if ( namespace == null )
        	this.globalNameSpace = new NameSpace("global");
		else
			this.globalNameSpace = namespace;
		
		compiler = JMelCompiler.make(new jmel.util.Hashtable());;
		
		long t2=System.currentTimeMillis();
    }

    public JMelInterpreter(
		InputStream in, PrintStream out, PrintStream err,
 		boolean interactive,
 		NameSpace namespace )
 	{
		this( in, out, err, interactive, namespace, null, null );
	}

    public JMelInterpreter(
		InputStream in, PrintStream out, PrintStream err, boolean interactive )    {
        this(in, out, err, interactive, null);
    }

	/**
		Construct a new interactive interpreter attached to the specified 
		console using the specified parent namespace.
	*/
    public JMelInterpreter(ConsoleInterface console, NameSpace globalNameSpace ) {

        this( console.getIn(), console.getOut(), console.getErr(), true, globalNameSpace );

		setConsole( console );
    }

	/**
		Construct a new interactive interpreter attached to the specified 
		console.
	*/
    public JMelInterpreter(ConsoleInterface console ) {
        this(console, null);
    }
    
	//	Create an interpreter for evaluation only.
    public JMelInterpreter( )
    {
		this( new StringBufferInputStream(""), System.out, System.err, false, null );
        evalOnly = true;	
	}
	// End constructors
	
	/**
		Attach the console thusly... ;)
	*/
	public void setConsole( ConsoleInterface console ) {
		this.console = console;
	}

	/*
		Set the global namespace for this interpreter.
		<p>
		
		Note: This is here for completeness.  If you're using this a lot
 		it may be an indication that you are doing more work than you have
 		to.  For example, caching the interpreter instance rather than the
 		namespace should not add a significant overhead.  No state other
 		than the debug status is stored in the interpreter.
		<p>
		All features of the namespace can also be accessed using the
 		interpreter via eval() and the script variable 'this.namespace'
		(or global.namespace as necessary).	
	*/

	public void setNameSpace( NameSpace globalNameSpace ) {
		this.globalNameSpace = globalNameSpace;
	}

	/*
		Get the global namespace of this interpreter.
		<p>
		
		Note: This is here for completeness.  If you're using this a lot
 		it may be an indication that you are doing more work than you have
 		to.  For example, caching the interpreter instance rather than the
 		namespace should not add a significant overhead.  No state other than
 		the debug status is stored in the interpreter.
  		<p>
  		
		All features of the namespace can also be accessed using the
 		interpreter via eval() and the script variable 'this.namespace'
		(or global.namespace as necessary).	
	*/
	public NameSpace getNameSpace() {
		return globalNameSpace;
	}

	/**
		Run interactively.  (printing prompts, etc.)
	*/
    public void run() {
        if(evalOnly)
            throw new RuntimeException("JMel Interpreter: No stream");
		Script.interactive.set(true);

        /*
          We'll print our banner using eval(String) in order to
          exercise the parser and get the basic expression classes loaded...
          This ameliorates the delay after typing the first statement.
        */
        if ( interactive )
			try { 
				eval("printBanner();"); 
			} catch ( EvalError e ) {
			}

        boolean eof = false;

		// init the callstack.  
		CallStack callstack = new CallStack();
		callstack.push( globalNameSpace );
		TopLevel node = null;

        while(!eof)
        {
            try
            {
                // try to sync up the console
                System.out.flush();
                System.err.flush();
                Thread.yield();  // this helps a little
                if(interactive)
                    print("jmel> ");

				compiler.log = new Log();
                node = compiler.compile(sourceFileInfo, in);
                //eof = node.defs.length() == 0;
                if (compiler.errorCount() == 0)
                {
					Commandable.returnValue.set(null);
					Object ret = (new Eval( callstack, this )).eval(node);
					if (Commandable.returnValue.get() != null)
						ret = Commandable.returnValue.get();
					
					// sanity check during development
					if ( callstack.depth() > 1 )
						throw new InterpreterError(
							"Callstack growing: "+callstack);

                    if(ret instanceof ReturnControl)
                        ret = ((ReturnControl)ret).value;
                    if(ret != Primitive.VOID)
                    {
                         if(TRACE)
                            println("// Result: " + ret + " //");
                    }
                }
            }
            catch(InterpreterError e)
            {
                error(e.getMessage());
             if(DEBUG)
                e.printStackTrace();
                if(!interactive)
                    eof = true;
            }
            catch(TargetError e)
            {
                error(e.getMessage() );
             if(DEBUG)
				if ( e.inNativeCode() )
					e.printStackTrace( DEBUG, err );
                if(!interactive)
                    eof = true;
           }
            catch (EvalError e)
            {
				if ( interactive )
					error( e.getMessage() );
				else
					error( e.getMessage() );
                if(DEBUG)
                    e.printStackTrace();
                if(!interactive)
                    eof = true;
            }
            catch(Throwable e)
            {
                error(e.getMessage());
             if(DEBUG)
                e.printStackTrace();
                if(!interactive)
                    eof = true;
            }
            finally
            {
				// reinit the callstack
				if ( callstack.depth() > 1 ) {
					callstack.clear();
					callstack.push( globalNameSpace );
				}
            }
        }

		if ( interactive && !noExitOnEOF ) 
			System.exit(0);
    }

	// begin source and eval

	// Read text from fileName and eval it.
    public Object source( String filename, NameSpace nameSpace )
 		throws FileNotFoundException, IOException, EvalError
 	{
		File file = pathToFile( filename );
		debug("Sourcing file: "+file);
		try {
			jmel.script.Script.interactive.set(false);
			Object obj = eval( new FileInputStream(file), nameSpace, filename );
			return obj;
		} catch (FileNotFoundException e) {
			throw e;
		} catch (IOException e) {
			throw e;
		} catch (EvalError e) {
			throw e;
		} finally {
			jmel.script.Script.interactive.set(true);
		}
	}

	/*
		Read text from fileName and eval it.
		Convenience method.  Use the global namespace.
	*/
    public Object source( String filename )
 		throws FileNotFoundException, IOException, EvalError
 	{
		return source( filename, globalNameSpace );
	}

    /*   
        Spawn a non-interactive local interpreter to evaluate text in the
 		specified namespace.
  		Return value is the evaluated object (or corresponding primitive
 		wrapper).
		@param sourceFileInfo is for information purposes only.  
		It is used to display error messages (and in the future may be made 
		available to the script).
		@throws EvalError on script problems
		@throws TargetError on unhandled exceptions from the script
    */

	// Note: we need a form of eval that passes the callstack through...	

	/*
	Can't this be combined with run() ?
	run seems to have stuff in it for
 interactive vs. non-interactive...
	compare them side by side and see what they do differently, aside from the	exception
 handling.
	*/
    public Object eval(
 		InputStream in, NameSpace nameSpace, String sourceFileInfo )
 		throws EvalError
 	{
		Object retVal = null;

		/*
 			Create non-interactive local interpreter for this namespace
			with source from the input stream and out/err same as
 			this interpreter.
		*/
        JMelInterpreter localInterpreter = new JMelInterpreter( in, out, err, false, nameSpace, this, sourceFileInfo );

		CallStack callstack = new CallStack();
		callstack.push( nameSpace );
 
        boolean eof = false;
		TopLevel node = null;

        try
        {
            node = compiler.compile(sourceFileInfo, in);
            eof = node.defs.length() == 0 || compiler.errorCount() != 0;
            if (!eof)
            {
				// nodes remember from where they were sourced
                retVal = (new Eval( callstack, localInterpreter )).eval(node);
				// sanity check during development
				if ( callstack.depth() > 1 )
					throw new InterpreterError(
						"Callstack growing: "+callstack);
                if ( retVal instanceof ReturnControl ) {
                    retVal = ((ReturnControl)retVal).value;
				}
            }
        } catch(InterpreterError e) {
             if(DEBUG)
                e.printStackTrace();
       		throw new EvalError(
				"Sourced file: "+sourceFileInfo+" internal Error:\n\t" + e.getMessage(), node);
        } catch( TargetError e ) {
           if(DEBUG)
                e.printStackTrace();
			// failsafe, set the Line as the origin of the error.
			if ( e.getTree()==null )
				e.setTree( node );
//				e.reThrow("Sourced file: "+sourceFileInfo);
        } catch(EvalError e) {
            if(DEBUG)
                e.printStackTrace();
			// failsafe, set the Line as the origin of the error.
			if ( e.getTree()==null )
				e.setTree( node );
			e.reThrow( "Sourced file: "+sourceFileInfo );
        } catch(Throwable e) {
           if(DEBUG)
                e.printStackTrace();
           throw new EvalError(
				"Sourced file: "+sourceFileInfo+" unknown error:\n\t"
				+ e.getMessage(), node);
        } finally {
		// reinit the callstack
			if ( callstack.depth() > 1 ) {
				callstack.clear();
				callstack.push( nameSpace );
			}            
		}
	
		return node; //Primitive.unwrap( retVal );
    }

	// Evaluate the inputstream in this interpreter's global namespace.
    public Object eval( InputStream in ) throws EvalError
 	{
		return eval( in, globalNameSpace, "eval stream" );
	}

	/**
		Evaluate the string in this interpreter's global namespace.
	*/
    public Object eval( String statement ) throws EvalError {
		debug("eval(String): "+statement);
		return eval(statement, globalNameSpace);
	}

	/**
		Evaluate the string in the specified namespace.
	*/
    public Object eval( String statement, NameSpace nameSpace )
 		throws EvalError {
 		String s = ( statement.endsWith(";") ? statement : statement+";" );
        return eval( new StringBufferInputStream(s), nameSpace, "<Inline eval of: "+s+" >" );
    }
	// end source and eval

	/*
		Print an error message in a standard format on the output stream
		associated with this interpreter. On the GUI console this will appear
 		in red, etc.
	*/

    public final void error(String s) {
		if ( console != null )
				console.error( "// Error: " + s +"\n" );
		else {
			err.println("// Error: " + s);
			err.flush();
		}
     }

	// ConsoleInterface
	// The interpreter reflexively implements the console interface that it
 	// uses.  Should clean this up by using an inner class to implement the
	// console for us.
	/*
 		Get the input stream associated with this interpreter.
		This may be be stdin or the GUI console.
	*/
	public InputStream getIn() { return in; }

	/*
 		Get the outptut stream associated with this interpreter.
		This may be be stdout or the GUI console.
	*/
	public PrintStream getOut() { return out; }

	/*
 		Get the error output stream associated with this interpreter.
		This may be be stderr or the GUI console.
	*/
	public PrintStream getErr() { return err; }

    public final void println(String s)
    {
        print(s + "\n");
    }

    public final void print(String s)
    {
		if (console != null) {
            console.print(s);
        } else {
            out.print(s);
            out.flush();
        }
    }
	// End ConsoleInterface
	/*
		Print a debug message on debug stream associated with this interpreter
		only if debugging is turned on.
	*/
    public final static void debug(String s)
    {
        if(DEBUG)
            debug.println("// Debug: " + s);
    }

	/* 
		Primary interpreter set and get variable methods
		Note: These are squeltching errors... should they?
	*/

	/**
		Get the value of the name.
		name may be any value. e.g. a variable or field
	*/
    public Object get( String name ) throws EvalError {
		Object ret = globalNameSpace.get( name, this );
		return Primitive.unwrap( ret );
	}

	/**
		Assign the value to the name.	
		name may evaluate to anything assignable. e.g. a variable or field.
	*/
    public void set(String name, Object value) 
		throws EvalError 
	{
		// map null to Primtive.NULL coming in...
		if ( value == null )
			value = Primitive.NULL;

		CallStack callstack = new CallStack();
		LHS lhs = globalNameSpace.getNameResolver( name ).toLHS( callstack, this );
		lhs.assign( value );
	}
	
    public void set(String name, int value) throws EvalError {
        set(name, new Primitive(value));
	}
    public void set(String name, double value) throws EvalError {
        set(name, new Primitive(value));
	}
    public void set(String name, float value) throws EvalError {
        set(name, new Primitive(value));
	}
	// end primary set and get methods


	// Methods for interacting with Parser

	/*
		Localize a path to the file name based on the bsh.cwd interpreter
 		working directory.
	*/
    public File pathToFile( String fileName )
 		throws IOException
	{
		File file = new File( fileName );

		// if relative, fix up to bsh.cwd
		if ( !file.isAbsolute() ) {
			String cwd = System.getProperty("user.dir");
			file = new File( cwd + File.separator + fileName );
		}

		return new File( file.getCanonicalPath() );
	}
	
	public static void redirectOutputToFile( String filename )
 	{
		try {
			PrintStream pout = new PrintStream(
 				new FileOutputStream( filename ) );
			System.setOut( pout );
			System.setErr( pout );
		} catch ( IOException e ) {
			System.err.println("Can't redirect output to file: "+filename );
		}
	}

	static void staticInit() {
	/*
 		Apparently in some environments you can't catch the security exception
		at all...  e.g. as an applet in IE  ... will probably have to work
 		around
 	*/
		try {
			Script.interpreted.set(true);
    		debug = System.err;
    		DEBUG = Boolean.getBoolean("debug");
    		TRACE = Boolean.getBoolean("trace");
			String outfilename = System.getProperty("outfile");
			if ( outfilename != null )
				redirectOutputToFile( outfilename );
		} catch ( SecurityException e ) {
 			System.err.println("Could not init static:"+e);
		} catch ( Exception e ) {
			System.err.println("Could not init static(2):"+e);
		} catch ( Throwable e ) {
 			System.err.println("Could not init static(3):"+e);
		}
	}

	/*
		Specify the source of the text from which this interpreter is reading.
		Note: there is a difference between what file the interrpeter is
 		sourcing and from what file a method was originally parsed.  One
		file may call a method sourced from another file.  See SimpleNode
		for origination file info.
		@see SimpleNode.getSourceFile
 	*/
	public String getSourceFileInfo() {
 		if ( sourceFileInfo != null )
			return sourceFileInfo;
		else
			return "<unknown source>";
	}
	public JMelInterpreter getParent() {
		return parent;
	}
}
