package jmel;

import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import jmel.script.Commands;
import jmel.util.*;
import jmel.code.*;
import jmel.tree.*;
import jmel.comp.*;
import java.io.*;
import java.util.Properties;

/**
 * This class provides a command-line interface to the GJC compiler.
 */
public class Main {

    /**
     * For testing: enter any options you want to be set implicitly
     *  here.
     */
    public static String[] forcedOpts = {};

    /**
     * The name of the compiler, for use in diagnostics.
     */
    String ownName;

    /**
     * Result codes.
     */
    public static final int EXIT_OK = 0;

    /**
     * Result codes.
     */
    public static final int EXIT_ERROR = 1;

    /**
     * Result codes.
     */
    public static final int EXIT_CMDERR = 2;

    /**
     * Result codes.
     */
    public static final int EXIT_SYSERR = 3;

    /**
     * Result codes.
     */
    public static final int EXIT_ABNORMAL = 4;

    /**
     * A three-column matrix describing all possible options.
     *  First column:   The option's name
     *  Second column:  The key to locate the option's argument, "" if there
     *                  is none.
     *  Third column:   The key to locate the internationalized string which
     *                  describes the option.
     *
     *  The first and second columns are important internally, since they
     *  are used to construct the table of actual options that's passed
     *  as an argument to the JavaCompiler constructor.
     */
    static String[][] standardOptions = {
    	{"-test", "", ""}, 
    	{"-O1", "", "opt.O1"}, {"-O2", "", "opt.O2"}, 
    	{"-g", "", "opt.g"}, {"-g:none", "", "opt.g.none"}, {"-g:{lines,vars,source}", "", "opt.g.lines.vars.source"}, 
    	{"-verbose", "", "opt.verbose"}, 
    	{"-deprecation", "", "opt.deprecation"}, 
    	{"-classpath", "opt.arg.path", "opt.classpath"}, 
    	{"-sourcepath", "opt.arg.path", "opt.sourcepath"}, 
    	{"-bootclasspath", "opt.arg.path", "opt.bootclasspath"}, 
    	{"-extdirs", "opt.arg.dirs", "opt.extdirs"}, 
    	{"-d", "opt.arg.directory", "opt.d"}, 
    	{"-target", "opt.arg.release", "opt.target"}
    };

    /**
     * Command-line options currently available.
     */
    String[][] enabledOptions;

	private String fileExtension;
	private List files = null;
	private String[] extraOpts = null;

    public static void main(String[] args) {
        new Main(args);
     }

    /**
     * Construct a compiler instance.
     */
     private String[] args;
     public Main(String[] args) {

		// set up new properties object
		// from file "jmel.properties"
		try {
			FileInputStream propFile = new FileInputStream("jmel.properties");
			Properties p = new Properties(System.getProperties());
			p.load(propFile);
			// set the system properties
			System.setProperties(p);
    	} catch (Throwable t) {
    	}    
         	
     	this.args = args;
		ownName = "jmel";
		enabledOptions = standardOptions;
		fileExtension = ".mel";
		processOpts(forcedOpts);
     	
		try {
			args = CommandLine.parse(args);
        } catch (IOException ex) {
			ioMessage(ex);
			return;
        }
		processOpts(args);
        if (extraOpts != null && options.get("-test") != null) {
			if (extraOpts.length == 0) {
				System.err.println("Syntax: Commands <cmd> -<opt> <args>");
				return;
			}
			try {
				String[] tmp = new String[extraOpts.length-1];
				System.arraycopy(extraOpts, 1, tmp, 0, extraOpts.length-1);
				Object value = Commands.get(extraOpts[0]).execAsExpr(tmp);
				System.out.println("Return value: ");Commands.print(value);
			} catch (Throwable t) {
				t.printStackTrace();
			}
		} else
	        if (options.get("-O1") == null && options.get("-O2") == null)
	        	interprete();
	        else
	        	compile();
    }

    /**
      * A table of all options that's passed to the JavaCompiler constructor.
      *  If an option has an argument, the option name is mapped to the argument.
      *  If a set option has no argument, it is mapped to itself.
      */
    public Hashtable options = Hashtable.make();

    /**
     * Print a string that explains usage.
     */
    public void help() {
        System.err.println(getLocalizedString("msg.usage.header", ownName));
        for (int i = 0; i < enabledOptions.length; i++) {
            if (!enabledOptions[i][0].startsWith("-X")) {
                String s = "  " + enabledOptions[i][0] + " ";
                if (!enabledOptions[i][1].equals(""))
                    s += getLocalizedString(enabledOptions[i][1]);
                System.err.print(s);
                for (int j = s.length(); j < 28; j++)
                    System.err.print(" ");
                System.err.println(getLocalizedString(enabledOptions[i][2]));
            }
        }
        System.err.println();
    }

    /**
      * Report a usage error.
      */
    void error(String key, String arg) {
        System.err.println(ownName + ": " + getLocalizedString(key, arg));
        help();
    }

    /**
      * Does argument string match option pattern?
      *  @param pattern    The option pattern, as found in the first column of
      *                    the enabledOptions matrix.
      *  @param arg        The command line argument string.
      */
    private boolean matches(String pattern, String arg) {
        int plen = pattern.length();
        int alen = arg.length();
        int ppos = Util.pos(pattern, ':', 0);
        int apos = Util.pos(arg, ':', 0);
        if (pattern.substring(0, ppos).equals(arg.substring(0, apos))) {
            if (ppos == plen && apos == alen) {
                return true;
            } else if (ppos < plen && apos < alen) {
                apos++;
                ppos++;
                if (ppos < plen && pattern.charAt(ppos) == '{') {
                    ppos++;
                    pattern = pattern.substring(0, pattern.length() - 1).concat(",");
                }
                int apos1 = apos;
                boolean found = true;
                while (found & apos1 < alen) {
                    int apos2 = Util.pos(arg, ',', apos1);
                    String option = arg.substring(apos1, apos2);
                    int ppos1 = ppos;
                    found = false;
                    while (!found && ppos1 < plen) {
                        int ppos2 = Util.pos(pattern, ',', ppos1);
                        if (pattern.substring(ppos1, ppos2).equals(option))
                            found = true;
                        ppos1 = ppos2 + 1;
                    }
                    apos1 = apos2 + 1;
                }
                return found;
            }
        }
        return false;
    }

    /**
      * Process command line arguments: store all command line options
      *  in `options' table and return all source filenames.
      *  @param args    The array of command line arguments.
      */
    public void processOpts(String[] args) {
		ListBuffer filenames = new ListBuffer();
        int ac = 0;
        while (ac < args.length) {
            String arg = args[ac];
            ac++;
            if (arg.startsWith("-")) {
                int j = 0;
                while (j < enabledOptions.length &&
                        !matches(enabledOptions[j][0], arg))
                    j++;
                if (j == enabledOptions.length) {
                    error("err.invalid.flag", arg);
                    return;
                }
                String operand = enabledOptions[j][1];
                if (operand.length() != 0) {
                    if (ac == args.length) {
                        error("err.req.arg", arg);
                        return;
                    }
                    operand = args[ac];
                    ac++;
                }
                if (arg.equals("-moreinfo")) {
                    Type.moreInfo = true;
                } else if (arg.equals("-version")) {
                    System.err.println(ownName + " " + JMelCompiler.version() +
                            ", " + JMelCompiler.date());
                } else {
                    if (arg.equals("-target")) {
                        if (!ClassFile.isValidTargetRelease(operand)) {
                            error("err.invalid.target", operand);
                            return;
                        }
                    } else if (arg.length() >= 3 && arg.charAt(2) == ':') {
                        operand = arg.substring(3);
                        arg = arg.substring(0, 3);
                    }
                    options.put(arg, operand);
                }
			} else if (arg.endsWith(fileExtension)) {
				filenames.append(arg);
			} else {
				extraOpts = new String[args.length-ac+1];
				System.arraycopy(args, ac-1, extraOpts, 0, args.length-ac+1);
				//error("err.invalid.arg", arg);
				return;
			}
		}
        files = filenames.toList();
    }

     /**
      * Programmatic interface for main function.
      * @param args    The command line parameters.
      */
    public int compile() {    	
        JMelCompiler comp = JMelCompiler.make(options);
        if (comp == null)
            return EXIT_SYSERR;

        try {
            if (files != null && !files.isEmpty()) {
		        comp.compile(files);
	            if (comp.errorCount() != 0)
	                return EXIT_ERROR;
				System.err.flush();
			}
            if (Util.statistics)
                printStats();
 		}
		catch (OutOfMemoryError ex) {
            {
                resourceMessage(ex);
                return EXIT_SYSERR;
            }
        }
        catch (FatalError ex) {
            {
                feMessage(ex);
                return EXIT_SYSERR;
            }
        }
        catch (Throwable ex) {
            {
				if (Boolean.getBoolean("debug"))
					ex.printStackTrace();            	
                feMessage(ex);
                return EXIT_ABNORMAL;
            }
        }
        return EXIT_OK;
    }
    
    public int interprete() {
        try {
        	JMelInterpreter interp;
            if (files != null && !files.isEmpty()) {
				interp = new JMelInterpreter();
				interp.source( (String)files.head, interp.globalNameSpace );
			} 
			interp = new JMelInterpreter( System.in, System.out, System.err, true );
			interp.run();
		}
		catch (OutOfMemoryError ex) {
            {
                resourceMessage(ex);
                return EXIT_SYSERR;
            }
        }
        catch (FatalError ex) {
            {
                feMessage(ex);
                return EXIT_SYSERR;
            }
        }
        catch (Throwable ex) {
            {
ex.printStackTrace();            	
                bugMessage(ex);
                return EXIT_ABNORMAL;
            }
        }
        return EXIT_OK;
    }

    /**
      * Print a message reporting an internal error.
      */
    protected void bugMessage(Throwable ex) {
        System.err.println(getLocalizedString("msg.bug", JMelCompiler.version()));
    }

    /**
      * Print a message reporting an fatal error.
      */
    protected void feMessage(Throwable ex) {
        System.err.println(ex.getMessage());
    }

    /**
      * Print a message reporting an input/output error.
      */
    protected void ioMessage(Throwable ex) {
        System.err.println(getLocalizedString("msg.io"));
        ex.printStackTrace();
    }

    /**
      * Print a message reporting an out-of-resources error.
      */
    protected void resourceMessage(Throwable ex) {
        System.err.println(getLocalizedString("msg.resource"));
        ex.printStackTrace();
    }

    /**
      * Print statistic information.
      */
    protected void printStats() {
        String[][] stats = {{"stats.tree", Integer.toString(Tree.count)}, {"stats.symbol",
        Integer.toString(Symbol.count)}, {"stats.type",
        Integer.toString(Type.count)}, {"stats.scope",
        Integer.toString(Scope.count)}, {"stats.scope.entry",
        Integer.toString(Scope.Entry.count)}, {"stats.item",
        Integer.toString(Items.count)}};
        if (Util.statistics) {
            for (int i = 0; i < stats.length; i++) {
                System.err.println(getLocalizedString(stats[i][0], stats[i][1]));
            }
        }
    }

    /**
      * Find a localized string in the resource bundle.
      *  @param key     The key for the localized string.
      */
    private String getLocalizedString(String key) {
        return getText(ownName + "." + key, null);
    }

    private String getLocalizedString(String key, String arg0) {
        return getText(ownName + "." + key, arg0);
    }
    private static final String javacRB = "jmel.resources.jmel";
    protected static ResourceBundle messageRB;

    /**
     * Initialize ResourceBundle.
     */
    protected void initResource() {
        try {
            {
                messageRB = ResourceBundle.getBundle(javacRB);
            }
        } catch (MissingResourceException e) {
            {
                System.out.println(e.getMessage());
                System.out.println(e.getClassName());
                throw new FatalError("Fatal Error: Resource for javac is missing");
            }
        }
    }

    /**
      * Get and format message string from resource.
      */
    private String getText(String key, String arg0) {
        if (messageRB == null)
            initResource();
        try {
            {
                String[] args = {arg0};
                return MessageFormat.format(messageRB.getString(key), args);
            }
        } catch (MissingResourceException e) {
            {
                if (arg0 == null)
                    arg0 = "null";
                String[] args = {key, arg0};
                String msg = "javac message file broken: key={0} arguments={1}";
                return MessageFormat.format(msg, args);
            }
        }
    }
}
