package jmel.interp;

import jmel.JMelInterpreter;
import jmel.code.*;
import jmel.script.*;
import jmel.util.*;
import jmel.tree.*;
import jmel.tree.Tree.*;
import java.io.*;
import java.util.Vector;
import java.util.Enumeration;
import java.lang.reflect.Array;

/**
 * Evaluates a Mel script.
 */
public class Eval extends GlobalVisitor  {

	private CallStack callstack;
	private boolean overrideNamespace;
	private JMelInterpreter interpreter;
	private Tree lhsTree = null;
	
	// The return value 
	public Object returnValue;

	/**
		@param overrideNamespace if set to true *no* new BlockNamespace will
		be swapped onto the stack and the eval will happen in the current
		top namespace.  This is used by ForStatement, etc.  which must 
		intialize the block with the for-init and also for those that 
		perform multiple passes in the same block.
	*/
	public Eval(CallStack callstack, JMelInterpreter interpreter, boolean overrideNamespace) {
		this.callstack = callstack;
		this.interpreter = interpreter;
		this.overrideNamespace = overrideNamespace;
	}
	public Eval(CallStack callstack, JMelInterpreter interpreter) {
		this(callstack,interpreter,false);
	}
	
	public Object eval(Tree tree) {
		returnValue = null;
		if (tree != null) tree.visit(this);
		return returnValue;
	}
		
	public void visitExpr(Tree tree) {
		EvalError.currentTree = tree;
		if (tree != null) tree.visit(this);
	}
		
	public void visitExprs(List trees) {
		if (trees.nonEmpty()) {
			visitExpr((Tree) trees.head);
			if (returnValue instanceof ReturnControl) return;
			for (List l = trees.tail; l.nonEmpty(); l = l.tail) {
				visitExpr((Tree) l.head);
				if (returnValue instanceof ReturnControl) return;
			}
		}
	}
	
	public void visitStats(List trees) {
		for (List l = trees; l.nonEmpty(); l = l.tail) {
			visitStat((Tree) l.head);
			if (returnValue instanceof ReturnControl) {
				return;
			}
		}
	}

	/**
	* visit unit consisting of package clause and import statements in toplevel,
	*  followed by class definition. if class definition == null,
	*  visit all definitions in toplevel.
	*  @param tree     The toplevel tree
	*  @param cdef     The class definition, which is assumed to be part of the
	*                  toplevel
	*/
	public void evalUnit(TopLevel tree) {
		for (List l = tree.defs; l.nonEmpty(); l = l.tail) {
			visitStat((Tree) l.head);
		}
	}

	/**
	* Visitor methods
	*/
	public void _case(TopLevel tree) {
		// Look for global vars
		tree.visit(new EvalGlobal(callstack, interpreter));
		evalUnit(tree);
	}
	
	public void _case(MethodDef tree) {
		NameSpace namespace = callstack.top();
		int length = tree.params.length(), i = 0;
		Class[] argTypes = new Class[length];
		for (List l = tree.params; l.nonEmpty(); l = l.tail, i++) {
			VarDef varDef = (VarDef)l.head;
			argTypes[i] = Reflect.getType(varDef.vartype);
		}
		Class returnType = Reflect.getType(tree.restype);
		namespace.setMethod( tree.name.toString(), new JMelScriptMethod( tree, argTypes, returnType, namespace ) );
		returnValue = Primitive.VOID;
	}

	private static boolean canCastToDeclaredType( Object value, Class toType ) {
		if ( !(value instanceof Primitive) ) {
			if (toType == int.class || 
				toType == float.class || 
				toType == double.class)
				return value instanceof Number;
			return value.getClass().isAssignableFrom(toType);
		}
		Class fromType = ((Primitive)value).getType();
		if (toType == Integer.TYPE) {
			return true;
		}
		if (toType == Float.TYPE) {
			return true;
		}
		if (toType == Double.TYPE) {
			return true;
		}
		if (toType == String.class) {
			return true;
		}
		if (toType == MelVector.class) {
			return true;
		}
		if (toType == MelMatrix.class) {
			return true;
		}
		if (MelArray.class.isAssignableFrom(toType)) {
			return true;
		}
		return false;
	}

	static void castError(Class from, Class to) {		
		castError(Reflect.normalizeClassName(from), Reflect.normalizeClassName(to) );    
	}    
	static void castError(String from, String to) {		
		throw new InternalError( "Illegal cast. Cannot cast " + from + " to " + to );    
	}

	public static Primitive castPrimitive( Primitive primValue, Class toType ) throws EvalError {
		// can't cast void to anything
		if ( primValue == Primitive.VOID )
			castError( "void value", Reflect.normalizeClassName(toType) );

		// unwrap, etc.
		Object value = primValue.getValue();		
		Class fromType = primValue.getType();		
		
		// trying to do numeric promotion		
		// first promote char to Number type to avoid duplicating code		
		if (value instanceof Boolean) {		
			boolean flag = ((Boolean)value).booleanValue();
			if(toType == Integer.TYPE)			
				value = new Primitive(flag?1:0);			
			else if(toType == Float.TYPE)				
				value = new Primitive(flag?1.0F:0.0F);			
			else if(toType == Double.TYPE)				
				value = new Primitive(flag?1.0D:0.0D);			
			else if(toType == String.class)				
				value = new Primitive(flag?"1":"0");			
			else if(toType == MelVector.class)				
				value = new Primitive(new MelVector(flag?1:0));				
			else if(toType == MelMatrix.class)				
				value = new Primitive(ConvUtils.toMatrix(flag));				
			else				
				castError(fromType, toType);			
			return (Primitive)value;		
		}
				
		if (value instanceof Number) {		
			Number number = (Number)value;			
			if(toType == Integer.TYPE)			
				value = new Primitive(number.intValue());			
			else if(toType == Float.TYPE)				
				value = new Primitive(number.floatValue());			
			else if(toType == Double.TYPE)				
				value = new Primitive(number.doubleValue());			
			else if(toType == String.class)				
				value = new Primitive(number.toString());			
			else if(toType == MelVector.class)				
				value = new Primitive(ConvUtils.toVector(number.floatValue()));				
			else if(toType == MelMatrix.class)				
				value = new Primitive(ConvUtils.toMatrix(number.floatValue()));		
			else				
				castError(fromType, toType);			
			return (Primitive)value;		
		}
		
		if (value instanceof String) {		
			if(toType == Integer.TYPE)			
				value = new Primitive(ConvUtils.toInt((String)value));			
			else if(toType == Float.TYPE)				
				value = new Primitive(ConvUtils.toFloat((String)value));			
			else if(toType == Double.TYPE)				
				value = new Primitive(ConvUtils.toDouble((String)value));			
			else if(toType == String.class)				
				value = new Primitive((String)value);			
			else if(toType == MelVector.class)				
				value = new Primitive(ConvUtils.toVector((String)value));				
			else if(toType == MelMatrix.class)				
				value = new Primitive(ConvUtils.toMatrix((String)value));				
			else				
				castError(fromType, toType);			
			return (Primitive)value;		
		}
		
		if (value instanceof MelVector) {		
			if(toType == Integer.TYPE)			
				value = new Primitive(ConvUtils.toInt((MelVector)value));			
			else if(toType == Float.TYPE)				
				value = new Primitive(ConvUtils.toFloat((MelVector)value));			
			else if(toType == Double.TYPE)				
				value = new Primitive(ConvUtils.toDouble((MelVector)value));			
			else if(toType == String.class)				
				value = new Primitive(ConvUtils.toString((MelVector)value));			
			else if(toType == MelVector.class)				
				value = new Primitive((MelVector)value);				
			else if(toType == MelMatrix.class)				
				value = new Primitive(ConvUtils.toMatrix((MelVector)value));				
			else				
				castError(fromType, toType);			
			return (Primitive)value;		
		}
		
		if (value instanceof MelMatrix) {		
			if(toType == Integer.TYPE)			
				value = new Primitive(ConvUtils.toInt((MelMatrix)value));			
			else if(toType == Float.TYPE)				
				value = new Primitive(ConvUtils.toFloat((MelMatrix)value));			
			else if(toType == Double.TYPE)				
				value = new Primitive(ConvUtils.toDouble((MelMatrix)value));			
			else if(toType == String.class)				
				value = new Primitive(ConvUtils.toString((MelMatrix)value));			
			else if(toType == MelVector.class)				
				value = new Primitive(ConvUtils.toVector((MelMatrix)value));			
			else if(toType == MelMatrix.class)				
				value = new Primitive((MelMatrix)value);				
			else				
				castError(fromType, toType);			
			return (Primitive)value;		
		}
		
		if (value instanceof MelArray) {		
			if(MelArray.class.isAssignableFrom(toType))				
				value = new Primitive((MelArray)value);				
			else				
				castError(fromType, toType);			
			return (Primitive)value;		
		}

		throw new EvalError("unknown type in cast: " + value.getClass());	
	}

	public static Object castObject( Object fromValue, Class toType )
		throws EvalError
	{
        Class fromType = fromValue.getClass();

		// The compiler isn't smart enough to allow me to leave this unassigned
		// even though it is clearly assigned in all cases below.
        Object result = null;

		// Going to a primitive type
       if ( toType.isPrimitive() ) 
			if ( fromValue instanceof Primitive )
				result = castPrimitive( (Primitive)fromValue, toType );
			else {
				if (fromValue instanceof Integer)
					result = castPrimitive( new Primitive((Integer)fromValue), toType );
				else if (fromValue instanceof Float)
					result = castPrimitive( new Primitive((Float)fromValue), toType );
				else if (fromValue instanceof Double)
					result = castPrimitive( new Primitive((Double)fromValue), toType );
				else if (fromValue instanceof String)
					result = castPrimitive( new Primitive((String)fromValue), toType );
				else if (fromValue instanceof MelVector)
					result = castPrimitive( new Primitive((MelVector)fromValue), toType );
				else if (fromValue instanceof MelMatrix)
					result = castPrimitive( new Primitive((MelMatrix)fromValue), toType );
				else
					castError(fromType, toType);
			}
       else 
			// Going to an object type
			if ( fromValue instanceof Primitive )
				// let castPrimitive handle trivial but legit case of NULL
				result = castPrimitive( (Primitive)fromValue, toType );
			else
				// Could probably add getAssignableForm here to allow 
				// special bsh widening converions... wrappers to wrappers
				if ( toType.isInstance(fromValue ) )
					result = fromValue;
				else
					castError(fromType, toType);

		if ( result == null )
			throw new InternalError("bad construct somewhere...");

		return result;
	}

	public void _case(VarDef tree) {
		try {
			NameSpace namespace = callstack.top();
			Class type = Reflect.getType(tree.vartype);

			// Type node is passed down the chain for array initializers
			// which need it under some circumstances
			Object value = Primitive.VOID;
	       	if (tree.init != null) {
				visitExpr(tree.init);
				value = returnValue;
			} else {
				if (tree.vartype instanceof TypeIdent) {
					TypeIdent t = (TypeIdent)tree.vartype;
					switch (t.typetag) {
						case Type.INT: value = new Primitive(0); break;
						case Type.FLOAT: value = new Primitive(0.0F); break;
						case Type.DOUBLE: value = new Primitive(0.0D); break;
						case Type.STRING: value = new Primitive(""); break;
						case Type.VECTOR: value = new Primitive(new MelVector()); break;
					}
				} else if (tree.vartype instanceof TypeArray) {
					TypeArray ta = (TypeArray)tree.vartype;
					Tree elemtype = ta.elemtype;
					TypeIdent ti = (TypeIdent)elemtype;
					switch (ti.typetag) {
						case Type.INT: value = new MelArrayInt(new int[ta.size]); break;
						case Type.FLOAT: value = new MelArrayFloat(new float[ta.size]); break;
						case Type.DOUBLE: value = new MelArrayDouble(new double[ta.size]); break;
						case Type.STRING: {
							String t[] = new String[ta.size];
							for (int i = 0; i < ta.size; i++)
								t[i] = "";
							value = new MelArrayString(t); 
							break;
						}
						case Type.VECTOR: {
							MelVector t[] = new MelVector[ta.size];
							for (int i = 0; i < ta.size; i++)
								t[i] = new MelVector();
							value = new MelArrayVector(t); 
							break;
						}
						case Type.MATRIX: {
							MelMatrix t[] = new MelMatrix[ta.size];
							for (int i = 0; i < ta.size; i++)
								t[i] = new MelMatrix();
							value = new MelArrayMatrix(t); 
							break;
						}
					}
					value = new Primitive((MelArray)value);
				} else if (tree.vartype instanceof TypeMatrix) {
					TypeMatrix t = (TypeMatrix)tree.vartype;
					value = new Primitive(new MelMatrix(new float[t.rows][t.cols]));
				}
			}

			// simple declaration with no value, e.g. int a;
			// null in value will prompt defaulting in setTypedVariable
			if ( value == Primitive.VOID ) 
				value = null;
			else if ( value == Primitive.NULL) {
 				// true null value being assigned
 				// leave as Primitive.NULL
			} else if ( canCastToDeclaredType( value, type ) )
				// allow specific numeric conversions on declaration
				value = castObject( value, type );
			else {
				throw new EvalError("Cannot affect " + value + " to " + type);
			}

			if ( value == null ) 
				throw new EvalError("Cannot affect " + value + " to " + type);

			if ((tree.flags & PUBLIC) != 0)
				namespace.setTypedVariable( tree.name.toString(), type, Primitive.ANY, true );
			else
				namespace.setTypedVariable( tree.name.toString(), type, value, (tree.flags & PUBLIC) != 0 );
		} catch ( EvalError e ) {
			e.reThrow( "Typed variable declaration", tree );
		}

        returnValue = Primitive.VOID;
	}
	
	public void _case(Block tree) {
		NameSpace enclosingNameSpace = null;
		if ( !overrideNamespace ) 
		{
			enclosingNameSpace= callstack.top();
			BlockNameSpace bodyNameSpace = new BlockNameSpace( enclosingNameSpace );
			callstack.swap( bodyNameSpace );
		} 
		visitBlock(tree.stats);
		if ( !overrideNamespace ) 
			callstack.swap( enclosingNameSpace );
	}

	public void _case(DoLoop tree) {
		boolean doOnceFlag = true;        
		while( doOnceFlag || evaluateCondition(tree.cond)) {			
			if ( tree.body == null ) // no body?				
				continue;			

			visitStat(tree.body); 
			Object ret = returnValue;			
			boolean breakout = false;			
			
			if(ret instanceof ReturnControl) {				
				switch(((ReturnControl)ret).kind ) {					
					case Tree.RETURN:	
						returnValue = ret;					
						return;					
						
					case Tree.CONTINUE:						
						continue;					
							
					case Tree.BREAK:						
						breakout = true;						
						break;				
				}			
			}		
			
			if(breakout) break;			
				
			doOnceFlag = false;		
		}        
		
		returnValue = Primitive.VOID;    
	}

	public void _case(WhileLoop tree) {
		boolean doOnceFlag = false;        
		while( doOnceFlag || evaluateCondition(tree.cond)) {			
			if ( tree.body == null ) // no body?				
				continue;			

			visitStat(tree.body); 
			Object ret = returnValue;			
			boolean breakout = false;			
			
			if(ret instanceof ReturnControl) {				
				switch(((ReturnControl)ret).kind ) {					
					case Tree.RETURN:	
						returnValue = ret;					
						return;					
						
					case Tree.CONTINUE:						
						continue;					
							
					case Tree.BREAK:						
						breakout = true;						
						break;				
				}			
			}		
			
			if(breakout) break;			
				
			doOnceFlag = false;		
		}        
		
		returnValue = Primitive.VOID;    
	}
	
	public void _case(ForLoop tree) {
		// Do the for init
		visitExprs(tree.init); 

		Object returnControl = Primitive.VOID;
		while(true)
		{
			if (tree.cond != null) {
				boolean cond = evaluateCondition(tree.cond);
				if ( !cond ) break;
			}

			boolean breakout = false; // switch eats a multi-level break here?
			if ( tree.body != null ) // not empty statement
			{
				// do *not* invoke special override for block... (see above)
				visitStat(tree.body);
				Object ret = returnValue;

				if (ret instanceof ReturnControl)
				{
					switch(((ReturnControl)ret).kind)
					{
						case Tree.RETURN:
							returnControl = ret;
							breakout = true;
							break;

						case Tree.CONTINUE:
							break;

						case Tree.BREAK:
							breakout = true;
							break;
					}
				}
			}

			if ( breakout ) break;

			visitExprs(tree.step);
		}
		returnValue = returnControl;
	}
	
	public void _case(ForIn tree) {
		Operation op = (Operation)tree.coll;
		Tree coll = (Tree)op.args.tail.head;
		visitExpr(coll); 

		// 		ForUtils coll = new ForUtils(<col>);
		// 		for (
		//			<var> = AssignUtils.assign(<var>, coll.firstElement());
		//			coll.hasMoreElement();
		//			<var> = AssignUtils.assign(<var>, coll.nextElement())
		// 		)

		// Note: it's important that there is only one exit point from this
		// method so that we can swap back the namespace.

		Ident var = (Ident)op.args.head;
		String varName = var.name.toString();

		Object theColl = Primitive.unwrap(returnValue);
		ForUtils _coll = new ForUtils(theColl);
		
		Object returnControl = Primitive.VOID;
loop:			
		while(_coll.hasMoreElement())
		{
			(callstack.top()).setVariable(varName, new Primitive(_coll.nextElement()));

			if ( tree.body != null ) // not empty statement
			{
				// do *not* invoke special override for block... (see above)
				visitStat(tree.body);
				Object ret = returnValue;

				if (ret instanceof ReturnControl)
				{
					switch(((ReturnControl)ret).kind)
					{
						case Tree.RETURN:
							returnControl = ret;
							break loop;

						case Tree.CONTINUE:
							break;

						case Tree.BREAK:
							break loop;
					}
				}
			}
		}
		returnValue = returnControl;
	}
		
	public void _case(Switch tree) {
		visitExpr(tree.selector);
		Object switchVal = returnValue;

		boolean found = false;
		ReturnControl returnControl = null;
		Case defaultCase = null;

		// while more labels or blocks and haven't hit return control
		for (List l = tree.cases; l.nonEmpty() && returnControl == null; l = l.tail) {
			Case pat = (Case) l.head;
			if (pat.pat != null) {
				visitExpr(pat.pat);

				if (switchVal.equals(returnValue) || found) {
					visitStat(pat);
					Object value = returnValue; 

					// should check to disallow continue here?
					if ( value instanceof ReturnControl ) {
						returnControl = (ReturnControl)value;
						break;
					}

					found = true;
				}
			}
			else
				defaultCase = pat;
		}

		if (!found && defaultCase != null) {
			visitStat(defaultCase);
			Object value = returnValue; 

			// should check to disallow continue here?
			if ( value instanceof ReturnControl ) {
				returnControl = (ReturnControl)value;
			}
		}

		if ( returnControl != null && returnControl.kind == Tree.RETURN )
			returnValue = returnControl;
		else
			returnValue = Primitive.VOID;
	}

	public void _case(Case tree) {
		visitStats(tree.stats);
	}
	
	public void _case(Catch tree) {
		try {
			visitExpr(tree.body);
			returnValue = new Primitive(0);
		} catch (Throwable t) {
			if(Boolean.getBoolean("debug"))
				t.printStackTrace();
			returnValue = new Primitive(1);
		}
	}

	public boolean evaluateCondition(Tree condExp) 
		throws EvalError
	{
		visitExpr(condExp);
		Object obj = returnValue;

		if(obj instanceof Primitive) {
			if ( obj == Primitive.VOID )
				throw new EvalError("Condition evaluates to void type", condExp );
			obj = ((Primitive)obj).getValue();
		}

		if(obj instanceof Boolean)
			return ((Boolean)obj).booleanValue();
		if(obj instanceof Integer)
			return ((Integer)obj).intValue() != 0;
		if(obj instanceof Float)
			return ((Float)obj).intValue() != 0;
		if(obj instanceof Double)
			return ((Double)obj).intValue() != 0;
		if(obj instanceof MelVector)
			return (int)((MelVector)obj).length() != 0;
		if(obj instanceof MelMatrix)
			return (int)((MelMatrix)obj).length() != 0;
		throw new EvalError("Condition must evaluate to a Boolean or boolean.", condExp );
	}

	public void _case(Conditional tree) {
		if (tree.tag == Tree.CONDSTAT) {
			Object ret = null;
			if( evaluateCondition(tree.cond) ) {
				visitStat(tree.thenpart);
				ret = returnValue;
			} else if (tree.elsepart != null) {
				visitStat(tree.elsepart);
				ret = returnValue;
			}
	
			if(ret instanceof ReturnControl)
				returnValue = ret;
			else    
				returnValue = Primitive.VOID;
		} else {
			Tree eval = evaluateCondition( tree.cond ) ? tree.thenpart : tree.elsepart;
			visitStat(eval);
		}
	}

	public void _case(Exec tree) {
		visitExpr(tree.expr);
		if (tree.expr instanceof Assign || tree.expr instanceof Assignop)
			returnValue = Primitive.VOID;
	}
	
	public void _case(Tree.Command tree) {
		visitExpr(tree.cmd);
	}
	
	public void _case(Alias tree) {
		NameSpace namespace = callstack.top();
		if (tree.cmd != null) namespace.setAlias( tree.name.toString(), tree.cmd );
		else if (tree.args != null) namespace.setAlias( tree.name.toString(), tree.args );
	}
	
	public void _case(Break tree) {
		returnValue = new ReturnControl( Tree.BREAK, Primitive.VOID );
	}
	
	public void _case(Continue tree) {
		returnValue = new ReturnControl( Tree.CONTINUE, Primitive.VOID );
	}
	
	public void _case(Return tree) {
		Object ret = Primitive.VOID;
		if (tree.expr != null) {
			visitExpr(tree.expr);
			ret = returnValue;
		}
		returnValue = new ReturnControl( Tree.RETURN, ret );
	}

	
	Object[] getArguments( List args )
	{
		// evaluate each child
		Object[] args_ = new Object[args.length()];
		int i = 0;
		for (List l = args; l.nonEmpty(); l = l.tail, i++) {
			visitExpr( (Tree) l.head );
			args_[i] = returnValue;
		}
		return args_;
	}

	private String cmdName;
	private Object[] cmdArgs;
	
	public boolean processNestedAliases(Apply tree) {
		cmdName = "";
		
		if (tree.meth instanceof Ident)
			cmdName = ((Ident)tree.meth).name.toString();
		else {
			Select t = (Select)tree.meth;
			Tree selected = t.selected;
			cmdName = t.name.toString();
			while (true) {
				if (selected instanceof Ident) {
					cmdName = ((Ident)selected).name.toString() + "." + cmdName;
					break;
				} else if (selected instanceof Select) {
					cmdName = ((Select)selected).name.toString() + "." + cmdName;
					selected = ((Select)selected).selected;
				}
			}
		}
		
		NameSpace ns = callstack.top();
		
		// save cmdArgs and cmdName (for the case when commands are invoked into the arguments)
		String tmpName = cmdName;
		Object[] tmp = cmdArgs;
		cmdArgs = null;
		Object[] args = getArguments(tree.args);
		cmdArgs = tmp;
		cmdName = tmpName;
		
		// check for aliases
		Object listArgs = ns.getAlias(cmdName);
		if (listArgs != null) {
			if (listArgs instanceof List) {
				if ( args.length != 0 )
					throw new InterpreterError("No argument expected for alias '" + cmdName + "'.");
				visitStats( (List)listArgs );
				return false;
			}

			Apply apply = (Apply) listArgs;
			if (!processNestedAliases(apply)) return false;
		} 
		
		if (cmdArgs != null) {
			tmp = new Object[cmdArgs.length + args.length];
			System.arraycopy(cmdArgs, 0, tmp, 0, cmdArgs.length);
			System.arraycopy(args, 0, tmp, cmdArgs.length, args.length);
			cmdArgs = tmp;
		} else
			cmdArgs = args;
			
		return true;
	}
	
	public void _case(Apply tree) {
		cmdArgs = null;
		if (!processNestedAliases(tree)) return;
		
		Name nm = new Name(callstack.top(), cmdName);
		try {
			returnValue = nm.invokeMethod(interpreter, cmdArgs, callstack, tree, !tree.statement);
		} catch (ReflectError e) {
			throw new EvalError("Error in method invocation: " + e.getMessage(), tree);
		} catch ( EvalError ee ) {
			ee.reThrow( tree );
			throw new Error("should be unreachable...");
		}
	}

	public void _case(NewArray tree) {
		Object initializers = null;
		Class baseType = null;
		
		Tree firstType = null;
		int typetag = -1;
		int i = 0;
		for (List l = tree.elems; l.nonEmpty(); l = l.tail, i++)
		{
			if (initializers == null) {
				firstType = (Tree) l.head;

				if (firstType instanceof Literal) typetag = ((Literal)firstType).typetag;
				else if (firstType instanceof TypeCast) typetag = ((TypeIdent)((TypeCast)firstType).clazz).typetag;
				
				switch (typetag) {
					case Type.CLASS:
						baseType = String.class;
						break;
							
					case Type.VECTOR:
						baseType = MelVector.class;
						break;
									
					case Type.INT:
						baseType = int.class;
						break;
						
					case Type.FLOAT:
						baseType = float.class;
						break;
										
					case Type.DOUBLE:
						baseType = double.class;
						break;		
				}
				
				if (baseType == null) {
					if (firstType instanceof NewVector)
						baseType = MelVector.class;
				}
				
				initializers = Array.newInstance( baseType, tree.elems.length() );
			}

			Tree node = (Tree) l.head;
			visitExpr(node);

			if ( returnValue == Primitive.VOID )
				throw new EvalError("Void in array initializer", node);

			// unwrap primitive to the wrapper type
			Object value;
			if ( returnValue instanceof Primitive )
				value = ((Primitive)returnValue).getValue();
			else
				value = returnValue;

			// store the value in the array
			try {
				Array.set(initializers, i, value);
			} catch( IllegalArgumentException e ) {
				JMelInterpreter.debug("illegal arg"+e);
				throw new EvalError( "illegal arg: " + baseType + " expected, " + value.getClass() + " found.", node  );
			} catch( ArrayStoreException e ) { // I think this can happen
				JMelInterpreter.debug("arraystore"+e);
				throw new EvalError( "Array store exception", node  );
			}
		}

		if (initializers == null) {
			returnValue = new Primitive(new MelArrayInt(new int[] {}));
			return;
		}
		
		returnValue = initializers;
			
		switch (typetag) {
			case Type.CLASS:
				returnValue = new Primitive(new MelArrayString(returnValue));
				break;
					
			case Type.VECTOR:
				returnValue = new Primitive(new MelArrayVector(returnValue));
				break;
							
			case Type.INT:
				returnValue = new Primitive(new MelArrayInt(returnValue));
				break;
				
			case Type.FLOAT:
				returnValue = new Primitive(new MelArrayFloat(returnValue));
				break;
								
			case Type.DOUBLE:
				returnValue = new Primitive(new MelArrayDouble(returnValue));
				break;		
		}
		
		if (baseType == MelVector.class) {
			returnValue = new Primitive(new MelArrayVector(returnValue));
		} else if (baseType == null) {
			if (firstType instanceof NewVector)
				returnValue = new Primitive(new MelArrayVector(returnValue));
		}
	}

	public void assign(Tree tree, Object rhs) {
		LHS lhs = toLHS(tree);
			
		if (lhs == null) throw new InterpreterError( "Error, null LHS" );

		returnValue = lhs.assign(rhs);
		JMelInterpreter.debug("lhs assign: "+returnValue);
	}
	
	public void _case(Assign tree) {
		try {
			visitExpr(tree.rhs);
			Object rhs = returnValue;
			if ( rhs == Primitive.VOID )
				throw new EvalError("Void assignment.", tree);
			if (tree.lhs instanceof GlobalRef) {
				GlobalRef globalRef = (GlobalRef)tree.lhs;
				String name = globalRef.name.toString();
				NameSpace namespace = callstack.top();
				Class type = Reflect.getType(globalRef.type);
				namespace.setTypedVariable( name, type, Primitive.ANY, true);

				returnValue = Primitive.VOID;
				return;
			}
			assign(tree.lhs, rhs);
		} catch ( EvalError e ) {
			e.reThrow(tree);
		}
	}
	
	public void _case(GlobalRef tree) {
		String name = tree.name.toString();
		NameSpace namespace = callstack.top();
		Class type = Reflect.getType(tree.type);
		namespace.setTypedVariable( name, type, Primitive.ANY, true);
	}

	private Object operation(Tree tree, Object lhs, Object rhs, int kind) 
		throws EvalError
	{
		/*
			Implement String += value;
			According to the JLS, value may be anything.
			In BeanShell, we'll disallow VOID (undefined) values.
			(or should we map them to the empty string?)
		*/
		if ( lhs instanceof String && rhs != Primitive.VOID ) {
			if ( kind != Tree.PLUS )
			throw new EvalError(
					"Use of non + operator with String LHS", tree);     

			return (String)lhs + rhs;
		}

		if ( lhs instanceof Primitive || rhs instanceof Primitive )
			if(lhs == Primitive.VOID || rhs == Primitive.VOID)
				throw new EvalError(
					"Illegal use of undefined object or 'void' literal", tree);
			else if ( lhs == Primitive.NULL || rhs == Primitive.NULL )
				throw new EvalError(
					"Illegal use of null object or 'null' literal", tree);


		return Primitive.binaryOperation(lhs, rhs, kind);
	}

	public void _case(Assignop tree) {
		LHS lhs = toLHS(tree.lhs);
			
		if (lhs == null)
			throw new InterpreterError( "Error, null LHS" );

		try {
			visitExpr(tree.rhs);
			Object rhs = returnValue;
			if ( rhs == Primitive.VOID )
				throw new EvalError("Void assignment.", tree);

			switch(tree.tag)
			{
				case Tree.PLUS_ASG:
					returnValue = lhs.assign(operation(tree, lhs.getValue(), rhs, Tree.PLUS));
					break;

				case Tree.MINUS_ASG:
					returnValue = lhs.assign(operation(tree, lhs.getValue(), rhs, Tree.MINUS));
					break;

				case Tree.MUL_ASG:
					returnValue = lhs.assign(operation(tree, lhs.getValue(), rhs, Tree.MUL));
					break;

				case Tree.DIV_ASG:
					returnValue = lhs.assign(operation(tree, lhs.getValue(), rhs, Tree.DIV));
					break;

				case Tree.BITAND_ASG:
					returnValue = lhs.assign(operation(tree, lhs.getValue(), rhs, Tree.BITAND));
					break;

				case Tree.BITOR_ASG:
					returnValue = lhs.assign(operation(tree, lhs.getValue(), rhs, Tree.BITOR));
					break;

				case Tree.BITXOR_ASG:
					returnValue = lhs.assign(operation(tree, lhs.getValue(), rhs, Tree.BITXOR));
					break;

				case Tree.MOD_ASG:
					returnValue = lhs.assign(operation(tree, lhs.getValue(), rhs, Tree.MOD));
					break;

				case Tree.SL_ASG:
					returnValue = lhs.assign(operation(tree, lhs.getValue(), rhs, Tree.SL));
					break;

				case Tree.SR_ASG:
					returnValue = lhs.assign(operation(tree, lhs.getValue(), rhs, Tree.SR));
					break;

				case Tree.USR_ASG:
					returnValue = lhs.assign(operation(tree, lhs.getValue(), rhs, Tree.USR));
					break;

				default:
					throw new InterpreterError("unimplemented operator in assignment BSH");
			}
		} catch ( EvalError e ) {
			e.reThrow(tree);
		}
		JMelInterpreter.debug("lhs assignop: "+returnValue);
	}

	private Object primitiveWrapperUnaryOperation(Object val, int kind)
	throws EvalError
	{
		Class operandType = val.getClass();
		Object operand = Primitive.promoteToInteger(val);

		if(operand instanceof Integer)
		{
			int result = Primitive.intUnaryOperation((Integer)operand, kind);
			return new Integer(result);
		}
		else if(operand instanceof Float) {
			float result = Primitive.floatUnaryOperation((Float)operand, kind);
			return new Float(result);
		} else if(operand instanceof Double) {
			double result = Primitive.doubleUnaryOperation((Double)operand, kind);
			return new Double(result);
		} else
			throw new InterpreterError("An error occurred.  Please call technical support."+operand.getClass());
	}

	/*
		object is a non-null and non-void Primitive type
	*/
	private boolean isPrimitiveValue( Object obj ) {
        return ( (obj instanceof Primitive) && (obj != Primitive.VOID) && (obj != Primitive.NULL) );
	}

	/*
		object is a java.lang wrapper for boolean, char, or number type
	*/
	private boolean isWrapper( Object obj ) {
        return ( obj instanceof Number );
	}

	private Object unaryOperation(Operation tree, Object op)
	{
		if(op instanceof Number)
			return primitiveWrapperUnaryOperation(op, tree.tag);

		if(!(op instanceof Primitive))
			throw new EvalError("Unary operation " + TreeInfo.operatorName(tree.tag).toString()
				+ " inappropriate for object", tree);

		return Primitive.unaryOperation((Primitive)op, tree.tag);
	}

	private Object lhsUnaryOperation(Operation tree, LHS lhs)
	{
		JMelInterpreter.debug("lhsUnaryOperation");
		Object prevalue, postvalue;
		prevalue = lhs.getValue();
		postvalue = unaryOperation(tree, prevalue);

		Object retVal;
		if ( tree.tag >= Tree.POSTINC)
			retVal = prevalue;
		else
			retVal = postvalue;

		lhs.assign(postvalue);
		return retVal;
	}

	public LHS toLHS(Tree tree)
	{
		if (lhsTree != null)
			throw new InterpreterError( "Error, not null LHS" );
		
		lhsTree = tree;
		visitExpr(tree); 
		lhsTree = null;

		if (returnValue instanceof LHS)
			return (LHS)returnValue;
		return null;
	}

	public void _case(Operation tree) {
		if (tree.tag <= Tree.COMPL) {
			visitExpr((Tree) tree.args.head );
			
			if(returnValue instanceof Number) {
				returnValue = primitiveWrapperUnaryOperation(returnValue, tree.tag);
				return;
			}

			if(!(returnValue instanceof Primitive))
				throw new EvalError("Unary operation " + TreeInfo.operatorName(tree.tag).toString() + " inappropriate for object", tree);

			returnValue = Primitive.unaryOperation((Primitive)returnValue, tree.tag);
		} else if (tree.tag <= Tree.POSTDEC) {
            returnValue = lhsUnaryOperation(tree, toLHS((Tree) tree.args.head ));
		} else {
			visitExpr((Tree) tree.args.head);
			Object lhs = returnValue;
			if (lhs instanceof Primitive) lhs = ((Primitive)lhs).getValue();

			// The following two boolean checks were tacked on.
			// This could probably be smoothed out.

			/*
				Look ahead and short circuit evaluation of the rhs if:
				we're a boolean AND and the lhs is false.
			*/
			if ( tree.tag == Tree.BITAND ) {
				Object obj = lhs;
				if ( obj instanceof Boolean && ( ((Boolean)obj).booleanValue() == false ) )
				returnValue = new Primitive(false);
				return;
			}

			/*
				Look ahead and short circuit evaluation of the rhs if:
				we're a boolean AND and the lhs is false.
			*/
			if ( tree.tag == Tree.BITOR ) {
				Object obj = lhs;
				if ( obj instanceof Boolean && ( ((Boolean)obj).booleanValue() == true ) )
				returnValue = new Primitive(true);
				return;
			}

			// end stuff that was tacked on for boolean short-circuiting.
			/*
				Are both the lhs and rhs either wrappers or primitive values?
				do binary op
			*/
			boolean isLhsWrapper = isWrapper( lhs );
			visitExpr((Tree) tree.args.tail.head);
		    Object rhs = returnValue;
			boolean isRhsWrapper = isWrapper( rhs );
			try {
				returnValue = new Primitive(Primitive.binaryOperation(lhs, rhs, tree.tag));
				return;
			} catch ( TargetError e ) {
				// this doesn't really help...  need to catch it higher?
				e.reThrow( tree );
			}
		}
	}

	public void _case(TypeCast tree) {
		NameSpace namespace = callstack.top();
		Class toType = Reflect.getType(tree.clazz);

		// evaluate the expression
		visitExpr(tree.expr);

		Object fromValue = returnValue;
		Class fromType = fromValue.getClass();

		try {
			returnValue = castObject( fromValue, toType );
		} catch ( EvalError e ) {
			e.reThrow( tree );
			throw new InterpreterError("can't happen"); // help the compiler
		}
	}
	
/*	public void _case(TypeTest tree) {
	}
*/
	public void _case(Indexed tree) {
		if (tree.indexed.tag == Tree.INDEXED)
			visitExpr(((Indexed)tree.indexed).indexed);
		else
			visitExpr(tree.indexed);
		Object obj = returnValue;
		
		if (!(obj instanceof Primitive) || 
			!(((Primitive)obj).getValue() instanceof MelArray) )
			throw new EvalError("Not an array", tree);
			
		MelArray array = (MelArray)((Primitive)obj).getValue();

		int index1, index2 = -1;
		try {
			if (tree.indexed.tag == Tree.INDEXED) {
				visitExpr(((Indexed)tree.indexed).index);
				Primitive val = (Primitive)returnValue;
				index1 = val.intValue();
				
				visitExpr(tree.index);
				val = (Primitive)returnValue;
				index2 = val.intValue();
				
				if (lhsTree == tree)
					returnValue = Reflect.getLHSIndex(obj, index1, index2);
				else
					returnValue = Reflect.getIndex(obj, index1, index2);
			} else {
				visitExpr(tree.index);
				Primitive val = (Primitive)returnValue;
				index1 = val.intValue();
				
				if (lhsTree == tree)
					returnValue = Reflect.getLHSIndex(obj, index1);
				else
					returnValue = Reflect.getIndex(obj, index1);
			}
		} catch(Exception e) {
			throw new EvalError("You can only index arrays by integer types", tree.index);
		}
	}
	
	public void _case(Select tree) {
	 	visitExpr(tree.selected);
		String name = tree.name.toString();
			
		NameSpace namespace = callstack.top();
		Object obj = (new Name(namespace, name)).resolveThisFieldReference( 
			callstack, namespace, interpreter, name, true, returnValue instanceof This );

		if (lhsTree == tree) 
			returnValue = Reflect.getLHSObjectField(returnValue, name);
		else
			returnValue = Reflect.getObjectField(returnValue, name);
	}
	
	public void _case(Ident tree) {
		String name = tree.name.toString();
		if (name.equals("true") || name.equals("false"))
			returnValue = new Primitive(new Boolean(name));				
		else if (name.equals("null"))
			returnValue = null;
		else 
			if (lhsTree == tree)
				returnValue = ( callstack.top() ).getNameResolver( name ).toLHS( callstack, interpreter );
			else {
				returnValue = ( callstack.top() ).getNameResolver( name ).toObject( callstack, interpreter );
				if (returnValue == null) returnValue = new Primitive(name);
			}
	}

	public void _case(Literal tree) {
		switch (tree.typetag) {
			case Type.INT:
				returnValue = new Primitive((Integer)tree.value);				
				break;

			case Type.FLOAT:
				returnValue = new Primitive((Float)tree.value);				
				break;

			case Type.DOUBLE:
				returnValue = new Primitive((Double)tree.value);				
				break;

			case Type.CLASS:
				returnValue = new Primitive((String)tree.value);				
				break;
			
			default:
				throw new EvalError("unknown literal " + tree.value);
		}
	}

	public void _case(TypeIdent tree) {
		returnValue = tree;
	}
/*	
	public void _case(TypeArray tree) {
	}
	
	public void _case(TypeMatrix tree) {
	}
*/	
	public void _case(NewVector tree) {
		Primitive x = new Primitive(0.0F), y = new Primitive(0.0F), z = new Primitive(0.0F);
		int i = 0;
		for (List l = tree.elems; l.nonEmpty(); l = l.tail, i++) {
			visitExpr((Tree) l.head);
			if (i == 0) x = (Primitive)castObject(returnValue, float.class);
			else if (i == 1) y = (Primitive)castObject(returnValue, float.class);
			else if (i == 2) z = (Primitive)castObject(returnValue, float.class);
		}
		returnValue = new Primitive(new MelVector((Float)x.getValue(), (Float)y.getValue(), (Float)z.getValue()));
	}
	
	public void _case(NewMatrix tree) {
		
		float f[][] = new float[tree.rows][tree.cols];
		for (int i = 0; i < tree.rows; i++) {
			ListBuffer elems2 = new ListBuffer();
			for (int j = 0; j < tree.cols; j++) { 
				visitExpr(tree.elements[i][j]);
				f[i][j] = ((Float)((Primitive)castObject(returnValue, float.class)).getValue()).floatValue();
			}
		}
		returnValue = new Primitive(new MelMatrix(f));
	}
/*	
	public void _case(Erroneous tree) {
	}
*/	
	public void _case(Tree tree) {
		throw new EvalError("(INTERP UNKNOWN: " + tree + ")");
	}
}
