package jmel.interp;

import jmel.util.Position;
import jmel.tree.Tree;

/**
	EvalError indicates that we cannot continue evaluating the script
	or the script has thrown an exception.

	EvalError may be thrown for a script syntax error, an evaluation 
	error such as referring to an undefined variable, an internal error.
	
	If the script has thrown an exception the exception will be wrapped
	in a TargetError.  
	@see TargetError
*/
public class EvalError extends Error {
	public static Tree currentTree;
	Tree tree;

	// Note: no way to get at Throwable message, must maintain our own
	String message;

	public EvalError(String s) {
		setMessage(s);
	}

	public EvalError(String s, Tree tree) {
		setMessage(s);
		this.tree = tree;
	}

	/**
	*/
	public String toString() {
		String trace;
		tree = currentTree;
		if ( tree != null ) {
			int line = Position.line(tree.pos);
			int col = Position.column(tree.pos);
			trace = "\n\tat Line: "+ line + " Col: " + col;
		} else
			// users should not see this, in the worst case the interpreter
			// should insert the Line() AST node.
			trace = "\n\t<at unknown location>";

		//return super.toString() + trace;
		return getMessage();
		//return getMessage() + trace;
	}

	/**
		Re-throw the eval error, prefixing msg to the message.
		<p>
		Unfortunately at the moment java.lang.Exception's message isn't 
		mutable so we just make a new one... could do something about this 
		later.
	*/
	public void reThrow( String msg ) 
		throws EvalError 
	{
		reThrow( msg, null );
	}

	/**
		Re-throw the eval error, specifying the node.
		If a node already exists the node is ignored.
		@see setNode()
		<p>

		Unfortunately at the moment java.lang.Exception's message isn't 
		mutable so we just make a new one... could do something about this 
		later.
	*/
	public void reThrow( Tree tree ) 
		throws EvalError 
	{
		reThrow( null, tree );
	}

	/**
		Re-throw the eval error, prefixing msg to the message and specifying
		the node.
		If a node already exists the addNode is ignored.
		@see setNode()
		<p>
		@param msg may be null for no additional message.

		Unfortunately at the moment java.lang.Exception's message isn't 
		mutable so we just make a new one... could do something about this 
		later.
	*/
	public void reThrow( String addMsg, Tree addTree ) 
		throws EvalError 
	{
		prependMessage( addMsg );
		addTree( addTree );
		throw this;
	}

	/**
		Set the AST node for trace info.
		@see reThrow
	
		This is useful for the interpreter if it detects that there is no
		trace info and wants to supply the Line() AST before printing.
	*/
	public void setTree( Tree tree ) {
		this.tree = tree;
	}

	/**
		The error has trace info associated with it. 
		i.e. It has an AST node that can print its location and source text.
	*/
	public Tree getTree() {
		return tree;
	}

	public String getErrorText() { 
		if ( tree != null )
			return tree.toString() ;
		else
			return "<unknown error>";
	}

	public int getErrorLineNumber() { 
		if ( tree != null )
			return tree.pos ;
		else
			return -1;
	}

	public String getErrorSourceFile() {
	/*	if ( tree != null )
			return node.getSourceFile() ;
		else*/
			return "<unknown file>";
	}

	public String getMessage() { return message; }

	public void setMessage( String s ) { message = s; }

	/**
		Prepend the message if it is non-null.
	*/
	protected void prependMessage( String s ) { 
		if ( s != null )
			message = s + " : "+ message;
	}

	protected void addTree( Tree addTree  ) {
		Tree tree = this.tree;
		if ( tree == null && addTree != null )
			tree = addTree;
	}
}

