package jmel.interp;

import jmel.JMelInterpreter;
import jmel.tree.*;
import jmel.tree.Tree.*;
import jmel.code.*;

/**
	This represents an *instance* of a bsh method declaration in a particular
	namespace.  This is a thin wrapper around the BSHMethodDeclaration
	with a pointer to the declaring namespace.

	The issue is that when a method is located in a subordinate namespace or
	invoked from an arbitrary namespace it must nontheless execute with its
	'super' as the context in which it was declared.

	i.e.
	The local method context is a child namespace of the declaring namespace.
*/
public class JMelScriptMethod implements java.io.Serializable
{
	public MethodDef method;

	/* 
		I believe this is always the namespace in which the method is
		defined...  It is a back-reference for the node, which needs to
		execute under this namespace.
		So it is not necessary to declare this transient, because we can
		only be saved as part of our namespace anyway... (currently).
	*/
	NameSpace declaringNameSpace;

	private String name;
	private Class[] argTypes;
	private Class returnType;
	private String[] argNames;

	public JMelScriptMethod( 
		MethodDef method, Class[] argTypes, Class returnType, NameSpace declaringNameSpace ) 
		throws EvalError
	{
		this.method = method;
		this.declaringNameSpace = declaringNameSpace;
		this.name = method.name.toString();
		this.argTypes = argTypes;
		this.returnType = returnType;
		
		VarDef[] vars = new VarDef[method.params.length()];
		argNames = new String[method.params.length()];
		method.params.toArray(vars);
		for (int i = 0; i < method.params.length(); i++)
			argNames[i] = vars[i].name.toString();
	}

	public Class[] getArgTypes() { return argTypes; }
	
	public Class getReturnType() { return returnType; }
	
	public boolean isPublic() { return (method.flags & Flags.PUBLIC) != 0; }

	/**
		Invoke the bsh method with the specified args, interpreter ref,
		and callstack.
		callerInfo is the node representing the method invocation
		It is used primarily for debugging in order to provide access to the 
		text of the construct that invoked the method through the namespace.
		@param callerInfo is the node representing the method invocation
			This is used primarily for debugging and may be null.
		@param callstack is the callstack of course.  If you are using a 
		hacked version of BeanShell that exposed this method take a look
		at NameSpace invokeMethod to see how to make a fake callstack...
	*/
	public Object invokeDeclaredMethod( 
		Object[] argValues, JMelInterpreter interpreter, CallStack callstack,
			Tree callerInfo ) 
		throws EvalError 
	{
		if ( argValues == null )
			argValues = new Object [] { };

		// Cardinality (number of args) mismatch
		if ( argValues.length != argTypes.length ) {
			throw new EvalError( 
					"Wrong number of arguments for local method: " 
					+ name, callerInfo);
		}

		// Make the local namespace for the method invocation
		MethodNameSpace localNameSpace = new MethodNameSpace( declaringNameSpace );
		localNameSpace.setTree( callerInfo );

		// set the method parameters in the local namespace
		for(int i=0; i<argTypes.length; i++)
		{
			try {
				argValues[i] = NameSpace.getAssignableForm(argValues[i], argTypes[i]);
			}
			catch(EvalError e) {
				throw new EvalError(
					"Invalid argument: " 
					+ "`"+argNames[i]+"'" + " for method: " 
					+ name + " : " + 
					e.getMessage(), callerInfo);
			}
			localNameSpace.setTypedVariable( argNames[i], argTypes[i], argValues[i], false);
		} 

		// Push the new namespace on the call stack
		callstack.push( localNameSpace );
		// Invoke the method
		Object ret = (new Eval(callstack, interpreter, true)).eval(method.body);
		// pop back to caller namespace
		callstack.pop();

		if ( ret instanceof ReturnControl )
		{
			ReturnControl rs = (ReturnControl)ret;
			if(rs.kind == Tree.RETURN)
				ret = ((ReturnControl)ret).value;
			else 
				// This error points to the method, should it?
				throw new EvalError("continue or break in method body", method);
		}

		// there should be a check in here for an explicit value return 
		// from a void type method... (throw evalerror)

		if(returnType != null)
		{
			// if void return type throw away any value
			// ideally, we'd error on an explicit 'return' of value
			if(ret == Primitive.VOID)
				return ret;
	
			// return type is a class
			try {
				ret = NameSpace.getAssignableForm(ret, (Class)returnType);
			}
			catch(EvalError e) {
				// This error points to the method, should it?
				throw new EvalError(
					"Incorrect type returned from method: " 
					+ name + e.getMessage(), method);
			}
		}
	
		return ret;
	}

	public String toString() {
		return "Script Method: "+name;
	}
}
