package jmel.interp;

import jmel.JMelInterpreter;
import jmel.script.*;
import jmel.util.List;
import jmel.tree.Tree;
import java.lang.reflect.*;
import java.util.Hashtable;
import java.io.*;
import java.lang.reflect.InvocationTargetException;

/**
	What's in a name?  I'll tell you...
	Name() is a somewhat ambiguous thing in the grammar and so is this.
	<p>
	
	This class is a name resolver.  It holds a possibly ambiguous dot 
	separated name and reference to a namespace in which it allegedly lives.  
	It provides methods that attempt to resolve the name to various types of 
	entities: e.g. an Object, a Class, a localy declared bsh method.
	<p>

	*** implementing ***
	Name objects are not to be constructed arbitrarily, but are to be 
	factoried by NameSpace.getNameResolver, which caches them subject to
	a namespace change.  This means that we can cache information about various
	types of resolution here.
*/
/*
	<strong>Implementation notes</strong>
	<pre>

	*** implementing ***
	Name objects are not to be constructed arbitrarily, but are to be 
	factoried by NameSpace.getNameResolver, which caches them subject to
	a namespace change.  This means that we can cache information about various
	types of resolution here.
	Note that we cannot simply cache any result, we must be smart about it.
	For example, the value of a variable may change between calls.  So we could	
	cache the knowledge that it is a variable, but must re-fetch the value
	each time.  We could even do cool optimizations such as knowing about
	'final' variables ;)
	
	Questions: how much will this actually buy us?  The simple cases are not
	expensive here, are they?  How often are long chains of names evaluated?
	*** implementing ***

	Threads:
	Thread safety: all of the work methods in this class must be synchronized
	because they share the internal intermediate evaluation state.

	Random note:
	In some ways Name wants to be a private inner class of NameSpace... 
	However it is used elsewhere as an absraction for objects which haven't
	been pinned down yet.  So it is exposed.

	Note on this.caller resolution:
	Although references like these do work:

		this.caller.caller.caller...   // works

	the equivalent using successive calls:

		// does *not* work
		for( caller=this.caller; caller != null; caller = caller.caller );

	is prohibited by the restriction that you can only call .caller on a 
	literal	this or caller reference.
	The effect is that magic caller reference only works through the current 
	'this' reference.
	The real explanation is that This referernces do not really know anything
	about their depth on the call stack.  It might even be hard to define
	such a thing...

	For those purposes we provide :

		this.callstack

	</pre>
*/
public class Name implements java.io.Serializable
{
	// These do not change during evaluation
	public NameSpace namespace;
	public String value = null;
	
	// ---------------------------------------------------------
	// The following instance variables mutate during evaluation and should
	// be reset by the reset() method where necessary

	// For evaluation
	private String evalName;		// text left to eval
	private Object evalBaseObject;	// base object for current eval

	private int callstackDepth;		// number of times eval hit 'this.caller'
	/*
		The last round consume the literal 'this' reference (not super, 
		global, or another This type var).  We use this flag to support magic
		variables that can only be referenced through 'this.xxx', e.g.
		this.interpreter and this.caller;
	*/
	private boolean literalThisReference;
	/**
		The last round consume the literal 'caller' reference (not super, 
		global, or another This type var).  This is used to limit references
		to .caller to only after a literal 'this' or compound '.caller'.
	*/
	private boolean literalCallerReference;

	//  
	//  End mutable instance variables.
	// ---------------------------------------------------------


	private void reset() {
		evalName = value;
		evalBaseObject = null;
		callstackDepth = 0;
		literalThisReference=false;
		literalCallerReference=false;
	}

	/**
		This constructor should *not* be used in general. 
		Use NameSpace getNameResolver() which supports caching.
		I wish I could make this "friendly" to just that class.
		@see NameSpace getNameResolver().
	*/
	public Name(NameSpace namespace, String s)
	{
		this.namespace = namespace;
		value = s;
	}
	
	/**
		Resolve possibly complex name to an object value.

		Throws EvalError on various failures.
		A null object value is indicated by a Primitive.NULL.
		A return type of Primitive.VOID comes from attempting to access
		an undefined variable.

		Some cases:
			myVariable
			myVariable.foo
			myVariable.foo.bar
			java.awt.GridBagConstraints.BOTH
			my.package.stuff.MyClass.someField.someField...

		Interpreter reference is necessary to allow resolution of 
		"this.interpreter" magic field.
		CallStack reference is necessary to allow resolution of 
		"this.caller" magic field.
		"this.callstack" magic field.
	*/
	synchronized public Object toObject( CallStack callstack, JMelInterpreter interpreter ) 
		throws EvalError
	{
		reset();
		Object obj = consumeNextObjectField( callstack, interpreter );
		return obj;
	}

	/**
		Get next prefixed object field component
	*/
	private Object consumeNextObjectField( 	
		CallStack callstack, JMelInterpreter interpreter ) 
		throws EvalError
	{
		/*
			Is it a simple variable name?
			Doing this first gives the correct Java precedence for vars 
			vs. imported class names (at least in the simple case - see
			tests/precedence1.bsh).  It should also speed things up a bit.
		*/
		if ( evalBaseObject == null ) 
		{
			Object obj = resolveThisFieldReference( 
				callstack, namespace, interpreter, evalName, false );

			if ( obj != Primitive.VOID ) {
				evalName = null; // finished
				return evalBaseObject = obj;  // convention
			}
		}

		/*
			Is it a bsh script variable reference?
			If we're just starting the eval of name (no base object)
			or we're evaluating relative to a This reference check.
		*/
		if ( ( evalBaseObject == null || evalBaseObject instanceof This  )) 
		{
			String varName = evalName;
			Object obj;
			if ( evalBaseObject == null ) {
				obj = resolveThisFieldReference( 
					callstack, namespace, interpreter, varName, false );
			} else {
				// null callstack, cannot be caller reference
				obj = resolveThisFieldReference( 
					callstack, ((This)evalBaseObject).namespace, 
					interpreter, varName, true );
			}

			if ( obj != Primitive.VOID ) 
			{
				// Resolved the variable
				return evalBaseObject = obj;
			}
		}

		/*
			If we didn't find a class or variable name (or prefix) above
			there are two possibilities:

			- If we are a simple name then we can pass as a void variable 
			reference.
			- If we are compound then we must fail at this point.
		*/
		if ( evalBaseObject == null ) {
			return null;
		}

		/*
			--------------------------------------------------------
			After this point we're definitely evaluating relative to
			a base object.
			--------------------------------------------------------
		*/

		/*
			Do some basic validity checks.
		*/

		if(evalBaseObject == Primitive.NULL) // previous round produced null
			throw new TargetError( "Null Pointer while evaluating: "
				+value, new NullPointerException() );

		if(evalBaseObject == Primitive.VOID) // previous round produced void
			throw new EvalError(
				"Undefined variable or class name while evaluating: "+value);

		if(evalBaseObject instanceof Primitive)
			throw new EvalError("Can't treat primitive like an object. "+
			"Error while evaluating: "+value);


		/* 
			Some kind of field access?
		*/

		String field = evalName;

		// check for field on object 
		// Note: could eliminate throwing the exception somehow
		try
		{
			Object obj = Reflect.getObjectField(evalBaseObject, field);
			return (evalBaseObject = obj);
		}
		catch(ReflectError e) {} // not a field
	
		// if we get here we have failed
		throw new EvalError(
			"Cannot access field: " + field + ", on object: " + evalBaseObject);
	}

	/**
		Resolve a variable relative to a This reference.

		This is the general variable resolution method, accomodating special
		fields from the This context.  Together the namespace and interpreter
		comprise the This context.  The callstack, if available allows for the
		this.caller construct.  
		Optionally interpret special "magic" field names: e.g. interpreter.

		@param callstack may be null, but this is only legitimate in special
		cases where we are sure resolution will not involve this.caller.

		@param namespace the namespace of the this reference (should be the
		same as the top of the stack?
	*/
	Object resolveThisFieldReference( 
		CallStack callstack, NameSpace thisNamespace, JMelInterpreter interpreter, 
		String varName, boolean specialFieldsVisible ) 
		throws EvalError
	{
		return resolveThisFieldReference( callstack, thisNamespace, interpreter, varName, specialFieldsVisible, false );
	}
	public Object resolveThisFieldReference( 
		CallStack callstack, NameSpace thisNamespace, JMelInterpreter interpreter, 
		String varName, boolean specialFieldsVisible, boolean isThis ) 
		throws EvalError
	{
		Object obj = null;
		// preserve the state of the last round flags until the end
		boolean 
			wasThis = isThis,		
			wasCaller = false;
		literalThisReference = isThis;

		obj = thisNamespace.getVariable(varName);

		literalThisReference = wasThis;
		literalCallerReference = wasCaller;
		return obj;
	}

	/*
	*/
	synchronized public LHS toLHS( 
		CallStack callstack, JMelInterpreter interpreter )
		throws EvalError
	{
		reset();
		return new LHS(namespace,evalName);
	}
	
	public Object getAlias( )
	{
		return namespace.getAlias( value );
	}
	
	
	public JMelScriptMethod toLocalMethod( Class [] sig )
	{
		return namespace.getMethod( value, sig );
	}
	
	private JMelScriptMethod toLocalMethod( Object [] args )
	{
		Class [] sig = Reflect.getTypes( args );
		return namespace.getMethod( value, sig );
	}

    /*
		Invoke the method identified by name.

        Name contains a wholely unqualfied messy name; resolve it to 
		( object | static prefix ) + method name and invoke.

        The interpreter is necessary to support 'this.interpreter' references
		in the called code. (e.g. debug());

        Some cases:

            // dynamic
            local();
            myVariable.foo();
            myVariable.bar.blah.foo();
            // static
            java.lang.Integer.getInteger("foo");

    */
    public Object invokeMethod( JMelInterpreter interpreter, Object[] args, CallStack callstack, Tree callerInfo	)
        throws EvalError, ReflectError
    {
		return invokeLocalMethod(interpreter, args, callstack, callerInfo, true);
	}
    public Object invokeMethod( JMelInterpreter interpreter, Object[] args, CallStack callstack, Tree callerInfo, boolean expr )
        throws EvalError, ReflectError
    {
		return invokeLocalMethod(interpreter, args, callstack, callerInfo, expr);
	}

    public Object invokeLocalMethod( JMelInterpreter interpreter, Object[] args, CallStack callstack, Tree callerInfo )
        throws EvalError, ReflectError
    {
		return invokeLocalMethod(interpreter, args, callstack, callerInfo, true);
	}
    public Object invokeLocalMethod( JMelInterpreter interpreter, Object[] args, CallStack callstack, Tree callerInfo, boolean expr )
        throws EvalError, ReflectError
    {
		if ("warning".equals(value)) {
			if (args == null || (args.length != 1 &&args.length != 2))
 	      		throw new EvalError("Wrong number or type of args for command "+value);
         	Reflect.unwrapPrimitives(args);
			if (!(args[0] instanceof String) ||
				(args.length == 2 && !(args[1] instanceof String)))
 	      		throw new EvalError("Wrong number or type of args for command "+value);
 	      	if (args.length == 1)
 	      		Commands.warning((String)args[0], jmel.util.Position.line(callerInfo.pos));
 	      	else
 	      		Commands.warning((String)args[0], (String)args[1], jmel.util.Position.line(callerInfo.pos));
 			return null;
		}
		if ("size".equals(value)) {
			if (args == null ||args.length != 1)
        		throw new EvalError("Wrong number or type of args for command "+value);
        	Reflect.unwrapPrimitives(args);
			return new Primitive(Commands.size(args[0]));
		}
		if ("clear".equals(value)) {
			if (args == null ||args.length != 1)
        		throw new EvalError("Wrong number or type of args for command "+value);
        	Reflect.unwrapPrimitives(args);
			Commands.clear(args[0]);
			return null;
		}
		if ("print".equals(value)) {
			if (args != null && args.length > 1)
       		throw new EvalError("Wrong number or type of args for command "+value);
       		
        	Reflect.unwrapPrimitives(args);
			if (args == null || args.length == 0) 
				Commands.print();
			else
 				Commands.print(args[0]);
			return null;
		}
		if ("eval".equals(value)) {
			if (args == null)
       			throw new EvalError("Wrong number or type of args for command "+value);
       		
        	Reflect.unwrapPrimitives(args);
        	String s = "";
        	for (int i = 0; i < args.length; i++) {
        		s += args[i].toString();
        		if (i < args.length-1) s += " ";
        	}
        	s += ";";
			return eval(interpreter, s);
		}
		if ("exists".equals(value)) {
			if (args == null ||args.length != 1)
        		throw new EvalError("Wrong number or type of args for command "+value);
        	Reflect.unwrapPrimitives(args);
			String s = args[0].toString();
			return new Primitive(getStream(s) != null);
		}
		if ("source".equals(value)) {
			if (args == null ||args.length != 1)
        		throw new EvalError("Wrong number or type of args for command "+value);
        	Reflect.unwrapPrimitives(args);
			String s = args[0].toString();
			InputStream in = getStream(s);
			if (in == null)
				throw new InternalError("File '"+s+"' not found.");
			try {
				jmel.script.Script.interactive.set(false);
				Object obj = interpreter.eval( in, namespace, s );
				return obj;
			} catch (Throwable t) {
				throw new InternalError(t.getMessage());
			} finally {
				jmel.script.Script.interactive.set(true);
			}
		}
		if ("bye".equals(value)) {
			System.exit(0);
		}
		
        // Check for locally declared method
        JMelScriptMethod meth = toLocalMethod( args );
        if ( meth != null )
            return meth.invokeDeclaredMethod( args, interpreter, callstack, callerInfo );
 
		/*
		Look for scripted command as resource
		*/
		String commandName = value.endsWith(".mel") ? value : value + ".mel";
		InputStream in = getStream(value);
		if (in != null)
		{
			if ( interpreter == null )
				throw new InterpreterError("2234432 interpreter = null");
			
			interpreter.eval( in, namespace, commandName);
			
			// try again
			meth = toLocalMethod( args );
			if(meth != null)
				return meth.invokeDeclaredMethod( args, interpreter, callstack, callerInfo );
			else
				throw new EvalError("Loaded resource: " + commandName + "had an error or did not contain the correct method");
		}
		
		commandName = value;
		
        // add interpreter and namespace to args list
		// check for compiled bsh command class
    	Reflect.unwrapPrimitives(args);
    	ObjectCommands oc = ObjectCommands.get((Object)commandName);
    	CommandObjectInvoker invoker = oc.getInvoker();
    	invoker.setScriptVars( namespace, interpreter, callstack, callerInfo );
		Object ret = oc.execAsExpr(args);
		if (ret == null) return Primitive.NULL;
		return new Primitive(ret);
 	}

	private Object eval(JMelInterpreter interpreter, String arg) {
		return interpreter.eval( arg, namespace);
	}

	public static final String pathSep = System.getProperty("path.separator");
	public static final String fileSep = System.getProperty("file.separator");
	private InputStream getStream(String arg) {
		String commandName = arg.endsWith(".mel") ? arg : arg + ".mel";
		String extdirs = System.getProperty("java.class.path");
		extdirs = extdirs.endsWith(pathSep) ? extdirs : extdirs + pathSep;
		int i = 0;
		int edlen = extdirs.length();
		while (i < edlen) {
			int end = extdirs.indexOf(pathSep, i);
			String extdir = extdirs.substring(i, end);
			File f = new File(extdir + fileSep + commandName);
			try {
				return new FileInputStream(f);
			} catch (FileNotFoundException e) {}
			i = end + 1;
		}
		return null;
	}

	// end compound name routines

	public String toString() { return value; }
}

