package	jmel.interp;

import jmel.JMelInterpreter;
import jmel.util.List;
import jmel.tree.*;
import jmel.script.*;
import jmel.tree.Tree;
import java.util.*;
import java.io.InputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;

/**
    A namespace	in which methods and variables live.  This is package public 
	because it is used in the implementation of some bsh commands.  However
	for normal use you should be using methods on bsh.Interpreter to interact
	with your scripts.
	<p>

	A bsh.This object is a thin layer over a NameSpace.  Together they 
	comprise a bsh scripted object context.
	<p>

	Note: I'd really like to use collections here, but we have to keep this
	compatible with JDK1.1 
*/
public class NameSpace implements NameSource
{
	public static final NameSpace JAVACODE = new NameSpace("Called from compiled Java code");
    static TreeMap definedGlobalVariables = new TreeMap();

	public String name; 
	private NameSpace parent;
	private Hashtable variables;
	private Hashtable aliases;
	private Hashtable globalVariables = new Hashtable();
	private Hashtable methods;
	private This thisReference;
	//String debugInfo;

    public NameSpace( String name ) { 
		this( null, name );
	}

    public NameSpace( NameSpace parent, String name ) {
		setName(name);
		setParent(parent);
    }

	public void setName( String name ) {
		this.name = name;
	}
	public String getName() {
		return this.name;
	}

	Tree callerInfoTree;
	/**
		Set the node associated with the creation of this namespace.
		This is used in debugging.
	*/
	void setTree( Tree tree ) {
		this.callerInfoTree= tree;
	}
	Tree getTree() {
		return this.callerInfoTree;
	}

	/**
		Resolve name to an object through this namespace.
	*/
	public Object get( String name, JMelInterpreter interpreter ) 
		throws EvalError 
	{
		CallStack callstack = new CallStack();
		return getNameResolver( name ).toObject( callstack, interpreter );
	}


	/**
		Get the names of variables defined in this namespace.
		(This does not show variables in parent namespaces).
	*/
	public String [] getVariableNames() {
		if ( variables == null )
			return new String [0];
		else
			return enumerationToStringArray( variables.keys() );
	}

	/**
		Get the names of methods defined in this namespace.
		(This does not show methods in parent namespaces).
	*/
	public String [] getMethodNames() {
		if ( methods == null )
			return new String [0];
		else
			return enumerationToStringArray( methods.keys() );
	}

	private String [] enumerationToStringArray( Enumeration e ) {
		Vector v = new Vector();
		while ( e.hasMoreElements() )
			v.addElement( e.nextElement() );
		String [] sa = new String [ v.size() ];
		v.copyInto( sa );
		return sa;
	}

	/**
		Get the parent namespace.
		Note: this isn't quite the same as getSuper().
		getSuper() returns 'this' if we are at the root namespace.
	*/
	public NameSpace getParent() {
		return parent;
	}

    public NameSpace getSuper()
    {
		if(parent != null)
			return parent;
		else
			return this;
    }

    public NameSpace getGlobal()
    {
		if(parent != null)
			return parent.getGlobal();
		else
			return this;
    }

	
	/**
		A This object is a thin layer over a namespace, comprising a bsh object
		context.  We create it here only if needed for the namespace.

		Note: that This is factoried for different capabilities.  When we
		add classpath modification we'll have to have a listener here to
		uncache the This reference and allow it to be refactoried.
	*/
    public This getThis( JMelInterpreter declaringInterpreter ) {

		if ( thisReference == null )
			thisReference = This.getThis( this, declaringInterpreter );
 
    	return thisReference;
    }

	/**
		Used for serialization
	*/
	public void prune() {
		parent = null;

	/*
	Do we need this?
	If so, fix the loop... can get Vectors of methods as well as methods

		if ( methods != null )
			// Prune the methods of this namespace - detach the nodes
			// from their parent nodes. 
			for( Enumeration e=methods.elements(); e.hasMoreElements(); )
				((BshMethod)e.nextElement()).method.prune();
	*/
	}

	public void setParent( NameSpace parent ) {
		this.parent = parent;
	}

   public Object getAlias( String name ) {
		return getAlias( name, true );
	}

   public Object getAlias( String name, boolean recurse ) {
		Object args = null;

		if(aliases != null)
			args = aliases.get(name);

		if ( recurse && (args == null) && (parent != null) )
			args = parent.getAlias(name, recurse);

		return args;
	}

   public void	setAlias(String	name, Tree cmd) 
		throws EvalError 
	{
		if (aliases == null)
			aliases =	new Hashtable();

		// does the alias already exist?
		if ( aliases.containsKey(name) ) 
		{
			throw new EvalError( "Redefinition of alias: "+name);
		} 

		// add the new alias
		aliases.put(name, cmd);
   }

   public void	setAlias(String	name, List args) 
		throws EvalError 
	{
		if (aliases == null)
			aliases =	new Hashtable();

		// does the alias already exist?
		if ( aliases.containsKey(name) ) 
		{
			throw new EvalError( "Redefinition of alias: "+name);
		} 

		// add the new alias
		aliases.put(name, args);
   }
    
    /**
		Get the specified variable in this namespace or a parent namespace.
		<p>
		Note: this method is primarily intended for use internally.  If you use
		this method outside of the bsh package you will have to use 
		Primitive.unwrap() to get primitive values.
		@see Primitive.unwrap()

		@return The variable value or Primitive.VOID if it is not defined.
	*/
    public Object getVariable( String name ) {
		return getVariable( name, true );
	}

	/**
		Get the specified variable in this namespace.
		If recurse is true extend search through parent namespaces.
		<p>
		Note: this method is primarily intended for use internally.  If you use
		this method outside of the bsh package you will have to use 
		Primitive.unwrap() to get primitive values.
		@see Primitive.unwrap()

		@return The variable value or Primitive.VOID if it is not defined.
	*/
    public Object getVariable( String name, boolean recurse ) {
		Object val = getVariableImpl( name, recurse );
		return unwrapVariable( val );
    }

	/**
		Unwrap a typed variable to its value.
		Turn null into Primitive.VOID
	*/
	protected Object unwrapVariable( Object val ) {
		if (val instanceof TypedVariable)
			val	= ((TypedVariable)val).getValue();

		return (val == null) ? Primitive.VOID :	val;
	}

	/**
		Return the raw variable retrieval (TypedVariable object or for untyped
		the simple value) with optional recursion.
		@return the raw variable value or null if it is not defined
	*/
    protected Object getVariableImpl( String name, boolean recurse ) {
		Object val = null;

		if(variables != null)
			val	= variables.get(name);

		if(val == null && globalVariables.get(name) != null) 
			val	= definedGlobalVariables.get(name);

		if ( recurse && (val == null) && (parent != null) )
			val	= parent.getVariable(name, recurse);

		return val;
    }

	/**
		Set a variable in this namespace.
		<p>
		Note: this method is primarily intended for use internally.  If you use
		this method outside of the bsh package and wish to set variables with
		primitive values you will have to wrap them using bsh.Primitive.
		@see bsh.Primitive

		@param value a value of null will remove the variable definition.
	*/
    public void	setVariable(String name, Object	value) throws EvalError 
	{
		if ( variables == null )
			variables =	new Hashtable();

		// hack... should factor this out...
		if ( value == null ) {
			variables.remove(name);
			return;
		}

		// Locate the variable definition if it exists
		Object current = getVariableImpl( name, false );

		// found a typed variable
		if ( (current != null) && (current instanceof TypedVariable) )
		{
			try {
				((TypedVariable)current).setValue(value);
			} catch(EvalError e) {
				throw new EvalError(
					"Typed variable: " + name + ": " + e.getMessage());
			} 
		} else if (globalVariables.get(name) == null)
			variables.put(name, value);
			
		if (globalVariables.get(name) != null) {
			globalVariables.put(name, value);
			definedGlobalVariables.put(name, value);
		}
    }

    /**
		Set the typed variable with the value.  
		An existing typed variable may only be set to the same type.
		If an untyped variable exists it will be overridden with the new
		typed var.
		The set will perform a getAssignableForm() on the value if necessary.

		<p>
		Note: this method is primarily intended for use internally.  If you use
		this method outside of the bsh package and wish to set variables with
		primitive values you will have to wrap them using bsh.Primitive.
		@see bsh.Primitive

		@param value If value is null, you'll get the default value for the type
    */
    public void	setTypedVariable(
		String	name, Class type, Object value,	boolean	isGlobal) 
		throws EvalError 
	{
		if (variables == null)
			variables =	new Hashtable();

		if (value == null)
		{
			// initialize variable to appropriate default value	- JLS 4.5.4
			if(type.isPrimitive())
			{
				if (type == Integer.TYPE)
					value = new	Primitive((int)0);
				else if (type == Float.TYPE)
					value = new	Primitive(0.0f);
				else if (type == Double.TYPE)
					value = new	Primitive(0.0d);
			}
			else
				value =	Primitive.NULL;
		}

		// does the variable already exist?
		if ( variables.containsKey(name) ) 
		{
			throw new EvalError( "Redefinition of variable: "+name);
		} 

		// add the new typed var
		TypedVariable tvar;
		if (isGlobal) {
			if (value == Primitive.ANY) tvar = (TypedVariable)definedGlobalVariables.get(name);
			else tvar = new TypedVariable(type, value);
			globalVariables.put(name, tvar);
			definedGlobalVariables.put(name, tvar);
		} else {
			tvar = new TypedVariable(type, value);
			variables.put(name, tvar);
		}
    }

	/**
		Note: this is primarily for internal use.
		@see Interpreter.source()
		@see Interpreter.eval()
	*/
    public void	setMethod(String name, JMelScriptMethod method) 
	{
		if(methods == null)
			methods = new Hashtable();

		Object m = methods.get(name);

		if ( m == null )
			methods.put(name, method);
		else 
		if ( m instanceof JMelScriptMethod ) {
			Vector v = new Vector();
			v.addElement( m );
			v.addElement( method );
			methods.put( name, v );
		} else // Vector
			((Vector)m).addElement( method );
    }

	/*
		Get the bsh method matching the specified signature declared in 
		this name space or a parent.
		<p>
		Note: this method is primarily intended for use internally.  If you use
		this method outside of the bsh package you will have to be familiar
		with BeanShell's use of the Primitive wrapper class.
		@see bsh.Primitive
	*/
    public JMelScriptMethod getMethod( String name, Class [] sig ) 
	{
		JMelScriptMethod method = null;

		Object m = null;
		if ( methods != null )
			m = methods.get(name);

		if ( m instanceof Vector ) {
			Vector vm = (Vector)m;
			JMelScriptMethod [] ma = new JMelScriptMethod[ vm.size() ];
			vm.copyInto( ma );

			Class [][] candidates = new Class[ ma.length ][];
			for( int i=0; i< ma.length; i++ )
				candidates[i] = ma[i].getArgTypes();

			int match = Reflect.findMostSpecificSignature( sig, candidates );
			if ( match != -1 )
				method = ma[match];
		} else
			method = (JMelScriptMethod)m;
			
		if ((method == null) && (parent != null))
			return parent.getMethod( name, sig );

		return method;
    }

	/**
		Implements NameSource
		@return all class and variable names in this and all parent
		namespaces
	*/
	public String [] getAllNames() 
	{
		Vector vec = new Vector();
		getAllNamesAux( vec );
		String [] names = new String [ vec.size() ];
		vec.copyInto( names );
		return names;
	}

	/**
		Helper for implementing NameSource
	*/
	protected void getAllNamesAux( Vector vec ) 
	{
		Enumeration varNames = variables.keys();
		while( varNames.hasMoreElements() )
			vec.addElement( varNames.nextElement() );

		Enumeration methodNames = methods.keys();
		while( methodNames.hasMoreElements() )
			vec.addElement( methodNames.nextElement() );

		if ( parent != null )
			parent.getAllNamesAux( vec );
	}

	Vector nameSourceListeners;
	/**
		Implements NameSource
		Add a listener who is notified upon changes to names in this space.
	*/
	public void addNameSourceListener( NameSource.Listener listener ) {
		if ( nameSourceListeners == null )
			nameSourceListeners = new Vector();
		nameSourceListeners.addElement( listener );
	}
	
    static class TypedVariable implements java.io.Serializable 
    {
		Class type;
		Object value = null; // uninitiailized

		TypedVariable(Class type, Object value) throws EvalError
		{
			this.type =	type;
			if ( type == null )
				throw new InterpreterError("null type in typed var: "+value);
			if (value != Primitive.ANY) setValue( value );
		}

		/**
			Set the value of the typed variable.
		*/
		void setValue(Object val) throws EvalError
		{
			// do basic assignability check
			val = getAssignableForm(val, type);
			
			// If we are a numeric primitive type we want to convert to the 
			// actual numeric type of this variable...  Being assignable is 
			// not good enough.
			if ( val instanceof Primitive && ((Primitive)val).isNumber() )
				try {
					val = Eval.castPrimitive( (Primitive)val, type );
				} catch ( EvalError e ) {
					throw new InterpreterError("auto assignment cast failed");
				} catch (Throwable t) {
					throw new EvalError("Unknown error:" + t);
				}
				
			this.value= val;
		}

		Object getValue() { return value; }
		
		Class getType() { return type;	}
 
		public String toString() { 
			return "TypedVariable: "+type+", value:"+value;
		}
   }

	/**
		@deprecated name changed.
		@see getAssignableForm()
	*/
    public static Object checkAssignableFrom(Object rhs, Class lhsType)
		throws EvalError
    {
		return getAssignableForm( rhs, lhsType );
	}

	/**
		<p>
		Determine if the RHS object can be assigned to the LHS type (as is,
		through widening, promotion, etc. ) and if so, return the 
		assignable form of the RHS.  Note that this is *not* a cast operation.
		Only assignments which are always legal (upcasts, promotion) are 
		passed.
		<p>

		In normal cases this functions as a simple check for assignability
		and the value is returned unchanged.  e.g. a String is assignable to
		an Object, but no conversion is necessary.  Similarly an int is 
		assignable to a long, so no conversion is done.
		In this sense assignability is in terms of what the Java reflection API
		will allow since the reflection api will do widening conversions in the 
		case of sets on fields and arrays.
		<p>
		The primary purpose of the abstraction "returning the assignable form"			abstraction is to allow non standard bsh assignment conversions. e.g.
		the wrapper stuff.  I'm still not sure how much of that we should
		be doing.
		<p>

		This method is used in many places throughout bsh including assignment
		operations and method selection.
		<p>

		@returns an assignable form of the RHS or throws EvalError
		@throws EvalError on non assignable
		@see BSHCastExpression.castObject();
	*/
	/*
		Notes:
	
		Need to define the exact behavior here:
			Does this preserve Primitive types to Primitives, etc.?

		This is very confusing in general...  need to simplify and clarify the
		various places things are happening:
			Reflect.isAssignableFrom()
			Primitive?
			here?
	*/
    public static Object getAssignableForm(Object rhs, Class lhsType)
		throws EvalError
    {
		Class originalType;
		
		if ( lhsType == null )
			throw new InterpreterError(
				"Null value for type in getAssignableForm");

		if(rhs == Primitive.VOID)
			throw new EvalError( "Undefined variable or class name");

		if(rhs == Primitive.ANY)
			return rhs;

		if (rhs == null || rhs == Primitive.NULL)
			if(!lhsType.isPrimitive())
				return rhs;
			else
				throw new EvalError(
					"Can't assign null to primitive type " + lhsType.getName());

		Class rhsType;

		if ( rhs instanceof Primitive ) 
		{
			// set the rhsType to the type of the primitive
			rhsType = originalType = ((Primitive)rhs).getType();

			// check for primitive/non-primitive mismatch
			if ( lhsType.isPrimitive() ) {
				// not doing this yet...  leaving as the assignable orig type
				/*
					We have two primitive types.  If Reflect.isAssignableFrom()
					which knows about primitive widening conversions says they
					are assignable, we will do a cast to change the value
				if ( Reflect.isAssignableFrom(
					((Primitive)lhs).getType(), ((Primitive)rhs).getType() )
				*/
			} else
			{
				// attempt promotion to	a primitive wrapper
				// if lhs a wrapper type, get the rhs as wrapper value
				// else error
				if( Number.class.isAssignableFrom(lhsType) )
				{
					rhs	= ((Primitive)rhs).getValue();
					// type is the wrapper class type
					rhsType = rhs.getClass();
				}
				else if( boolean.class.isAssignableFrom(lhsType) ) {
					rhs	= ((Primitive)rhs).getValue();
					rhsType = Boolean.class;
				} else if( String.class.isAssignableFrom(lhsType) )
					rhsType = String.class;
				else if( MelVector.class.isAssignableFrom(lhsType) )
					rhsType = MelVector.class;
				else if( MelMatrix.class.isAssignableFrom(lhsType) )
					rhsType = MelMatrix.class;
				else if( MelArrayInt.class.isAssignableFrom(lhsType) )
					rhsType = MelArrayInt.class;
				else if( MelArrayFloat.class.isAssignableFrom(lhsType) )
					rhsType = MelArrayFloat.class;
				else if( MelArrayDouble.class.isAssignableFrom(lhsType) )
					rhsType = MelArrayDouble.class;
				else if( MelArrayVector.class.isAssignableFrom(lhsType) )
					rhsType = MelArrayVector.class;
				else if( MelArrayString.class.isAssignableFrom(lhsType) )
					rhsType = MelArrayString.class;
				else
					assignmentError(lhsType, originalType);
			}
		} else 
		{
			// set the rhs type
			rhsType = originalType = rhs.getClass();

			// check for primitive/non-primitive mismatch
			if ( lhsType.isPrimitive() ) {

				// attempt unwrapping wrapper class for assignment 
				// to a primitive

				if (Number.class.isAssignableFrom(rhsType))
				{
					rhs	= new Primitive((Number)rhs);
					rhsType = ((Primitive)rhs).getType();
				}
				else
					assignmentError(lhsType, originalType);
			}
		}

		// This handles both class types and primitive .TYPE types
		if ( Reflect.isAssignableFrom(lhsType, rhsType) )
			return rhs;
	
		if ( rhs instanceof Primitive ) 
			rhs	= ((Primitive)rhs).getValue();
			
		/* 
			bsh extension -
			Attempt widening conversions as defined in JLS 5.1.2
			except perform them on primitive wrapper objects.
		*/
		if(lhsType == Float.class || lhsType == float.class) {
			if(rhsType == Integer.class || rhsType == int.class)
				return new Float(((Number)rhs).floatValue());
		}
		if(lhsType == Integer.class || lhsType == int.class) {
			if(rhsType == Boolean.class || rhsType == boolean.class)
				return new Integer(((Boolean)rhs).booleanValue()?1:0);
		}

		if(lhsType == Double.class || lhsType == double.class) {
			if (rhsType == Integer.class || rhsType == int.class ||
				rhsType == Float.class || rhsType == float.class)
				return new Double(((Number)rhs).doubleValue());
		}

		assignmentError(lhsType, originalType);

		return rhs;
    }

    private static void	assignmentError(Class lhs, Class rhs) throws EvalError
    {
		String lhsType = Reflect.normalizeClassName(lhs);
		String rhsType = Reflect.normalizeClassName(rhs);
		throw new EvalError ("Can't assign " + rhsType + " to "	+ lhsType);
    }

	public String toString() {
		return
			"NameSpace: "
			+ ( name==null
				? super.toString()
				: name + " (" + super.toString() +")" );
	}

	/**
		Invoke a method in this namespace with the specified args and
		interpreter reference.  The caller namespace is set to this namespace.
		This is a convenience for users outside of this package.
		<p>
		Note: this method is primarily intended for use internally.  If you use
		this method outside of the bsh package and wish to use variables with
		primitive values you will have to wrap them using bsh.Primitive.
		@see bsh.Primitive
	*/
	public Object invokeMethod( 
		String methodName, Object [] args, JMelInterpreter interpreter ) 
		throws EvalError
	{
		return invokeMethod( methodName, args, interpreter, null, null );
	}

	/**
		invoke a method in this namespace with the specified args,
		interpreter reference, and callstack
		This is a convenience for users outside of this package.
		<p>
		Note: this method is primarily intended for use internally.  If you use
		this method outside of the bsh package and wish to use variables with
		primitive values you will have to wrap them using bsh.Primitive.
		@param if callStack is null a new CallStack will be created and
			initialized with this namespace.
		@see bsh.Primitive
	*/
	public Object invokeMethod( 
		String methodName, Object [] args, JMelInterpreter interpreter, 
		CallStack callstack, Tree callerInfo ) 
		throws EvalError
	{
		if ( callstack == null ) {
			callstack = new CallStack();
			callstack.push( this );
		}

		// Look for method in the bsh object
        JMelScriptMethod meth = getMethod( methodName, Reflect.getTypes( args ) );
        if ( meth != null )
           return meth.invokeDeclaredMethod( args, interpreter, callstack, callerInfo );

		// Look for a default exec() handler method
		meth = getMethod( "exec", null );

		// Call script "exec( );
		if ( meth != null )
			return meth.invokeDeclaredMethod( 
				new Object [] { methodName, args }, interpreter, callstack, callerInfo );

		throw new EvalError( "No locally declared method: " + methodName + " in namespace: " + this );
	}

	/**
		This is the factory for Name objects which resolve names within
		this namespace (e.g. toObject(), toClass(), toLHS()).
		This supports name resolver caching, allowing Name objects to 
		cache info about the resolution of names for performance reasons.
		(This would be called getName() if it weren't already used for the
		simple name of the NameSpace)
	*/
	public Name getNameResolver( String name ) {
		// no caching yet
		return new Name(this,name);
	}

	public int getInvocationLine() {
		Tree tree = getTree();
		if ( tree != null )
			return tree.pos;
		else
			return -1;
	}
	public String getInvocationText() {
		Tree tree = getTree();
		if ( tree != null )
			return tree.toString();
		else
			return "<invoked from Java code>";
	}
}

