package jmel.interp;

import jmel.tree.Tree;
import jmel.script.*;

/**
    Wrapper for primitive types in Bsh.  This is package public because it 
	is used in the implementation of some bsh commands.

    See the note in LHS.java about wrapping objects.
*/
public class Primitive
{
    // stored internally in java.lang. wrappers
    private Object value;

    private static class Special
    {
        private Special() { }

        public static final Special NULL_VALUE = new Special();
        public static final Special VOID_TYPE = new Special();
        public static final Special ANY_TYPE = new Special();
    }

    /*
        NULL means "no value".
        This ia a placeholder for primitive null value.
    */
    public static final Primitive NULL = new Primitive(Special.NULL_VALUE);

    /**
        VOID means "no type".
        Strictly speaking, this makes no sense here.  But for practical
        reasons we'll consider the lack of a type to be a special value.
    */
    public static final Primitive VOID = new Primitive(Special.VOID_TYPE);
    
    /*
        ANY means "any value".
        This ia a placeholder for primitive pre-value (pre-definition of global vars).
    */
    public static final Primitive ANY = new Primitive(Special.ANY_TYPE);

    public Primitive(Object value)
    {
        if(value == null)
            throw new InterpreterError(
				"Use Primitive.NULL instead of Primitive(null)");

		boolean ok = true;
    	if (!(value instanceof Number) &&
    		!(value instanceof Special) &&
	   		!(value instanceof MelVector) &&
    		!(value instanceof MelMatrix) &&
    		!(value instanceof String) &&
    		!(value instanceof MelArray) &&
    		!(value instanceof Boolean)) {
    		if (value.getClass().isArray()) {
    			Class type = value.getClass().getComponentType();
    			if (type == int.class) value = new MelArrayInt((int[])value);
    			else if (type == float.class) value = new MelArrayFloat((float[])value);
    			else if (type == double.class) value = new MelArrayDouble((double[])value);
    			else if (type == MelVector.class) value = new MelArrayVector((MelVector[])value);
	   			else if (type == String.class) value = new MelArrayString((String[])value);
    			else 
    				ok = false;
    		}
		}

    	if (!ok) 
	        throw new InterpreterError("Cannot create a primitive with object of type " + value.getClass().getName());
	
        this.value = value;
    }

    public Primitive(Number number) { this((Object)number); }

    public Primitive(Integer value) { this((Object)value); }
    public Primitive(Float value) { this((Object)value); }
    public Primitive(Double value) { this((Object)value); }
    public Primitive(MelVector value) { this((Object)value); }
    public Primitive(MelMatrix value) { this((Object)value); }
    public Primitive(String value) { this((Object)value); }
    public Primitive(MelArray value) { this((Object)value); }
    public Primitive(Boolean value) { this((Object)value); }

    public Primitive(int value) { this(new Integer(value)); }
    public Primitive(float value) { this(new Float(value)); }
    public Primitive(double value) { this(new Double(value)); }
    public Primitive(boolean value) { this(new Boolean(value)); }

    public Object getValue()
    {
        if(value == Special.NULL_VALUE)
            return null;
        else if(value == Special.VOID_TYPE)
                throw new InterpreterError("attempt to unwrap void type");
        else
            return value;
    }

    public String toString()
    {
        if(value == Special.NULL_VALUE)
            return "null";
        else if(value == Special.VOID_TYPE)
            return "void";
        else
            return value.toString();
    }

    public Class getType()
    {
        return getType(value);
    }

    private Class getType(Object o)
    {
        if(o instanceof Integer)
            return Integer.TYPE;
		else if(o instanceof Boolean)
			return Boolean.TYPE;
		else if(o instanceof Float)
            return Float.TYPE;
        else if(o instanceof Double)
            return Double.TYPE;
		else if(o instanceof String)
			return String.class;
		else if(o instanceof MelVector)
			return MelVector.class;
		else if(o instanceof MelMatrix)
			return MelMatrix.class;
		else if(o instanceof MelArrayInt)
			return MelArrayInt.class;
		else if(o instanceof MelArrayFloat)
			return MelArrayFloat.class;
		else if(o instanceof MelArrayDouble)
			return MelArrayDouble.class;
		else if(o instanceof MelArrayVector)
			return MelArrayVector.class;
		else if(o instanceof MelArrayString)
			return MelArrayString.class;

        return null;
    }

	/**
		Allow primitive operations on wrapper types such as Integer and Boolean.
		This is static so that it can be reached from wherever...
	*/
    public static Object binaryOperation(Object obj1, Object obj2, int kind)
        throws EvalError
    {
		// special primitive types
        if(obj1 == NULL || obj2 == NULL)
            throw new EvalError(
				"Null value or 'null' literal in binary operation");
        if(obj1 == VOID || obj2 == VOID)
            throw new EvalError(
			"Undefined variable, class, or 'void' literal in binary operation");

		// keep track of the original types
		Class lhsOrgType = obj1.getClass();
		Class rhsOrgType = obj2.getClass();

		// Unwrap primitives
        if(obj1 instanceof Primitive)
            obj1 = ((Primitive)obj1).getValue();
        if(obj2 instanceof Primitive)
            obj2 = ((Primitive)obj2).getValue();

        Object[] operands = promotePrimitives(obj1, obj2);
        Object lhs = operands[0];
        Object rhs = operands[1];

		Object result = null;
		try {
			result = binaryOperationImpl( lhs, rhs, kind );
		} catch ( ArithmeticException e ) {
			throw new TargetError("Arithemetic Exception in binary op", e);
		}

		// If both original args were Primitives return a Primitive result
		// else it was mixed (wrapper/primitive) return the wrapper type
		if ( lhsOrgType == Primitive.class && rhsOrgType == Primitive.class )
			return new Primitive( result );
		else
			return result;
    }

    static Object binaryOperationImpl( Object lhs, Object rhs, int kind )
        throws EvalError
	{
		if(lhs instanceof Boolean) {
			if(rhs instanceof Boolean)
				return intBinaryOperation( new Integer(((Boolean)lhs).booleanValue()?1:0), new Integer(((Boolean)rhs).booleanValue()?1:0), kind );
			else if(rhs instanceof Number)
				return intBinaryOperation( new Integer(((Boolean)lhs).booleanValue()?1:0), new Integer(((Number)rhs).intValue()), kind );
			else if(rhs instanceof MelVector)
				return intBinaryOperation( new Integer(((Boolean)lhs).booleanValue()?1:0), (MelVector)rhs, kind );
			else if(rhs instanceof MelMatrix)
				return intBinaryOperation( new Integer(((Boolean)lhs).booleanValue()?1:0), (MelMatrix)rhs, kind );
		} else if(lhs instanceof Integer) {
			if(rhs instanceof Boolean)
				return intBinaryOperation( (Integer)lhs, new Integer(((Boolean)rhs).booleanValue()?1:0), kind );
			else if(rhs instanceof Number)
				return intBinaryOperation( (Integer)lhs, new Integer(((Number)rhs).intValue()), kind );
			else if(rhs instanceof MelVector)
				return intBinaryOperation( (Integer)lhs, (MelVector)rhs, kind );
			else if(rhs instanceof MelMatrix)
				return intBinaryOperation( (Integer)lhs, (MelMatrix)rhs, kind );
		} else if(lhs instanceof Float) {
			if(rhs instanceof Boolean)
				return floatBinaryOperation( (Float)lhs, new Float(((Boolean)rhs).booleanValue()?1.0F:0.0F), kind );
			else if(rhs instanceof Number)
				return floatBinaryOperation( (Float)lhs, new Float(((Number)rhs).floatValue()), kind );
			else if(rhs instanceof MelVector)
				return floatBinaryOperation( (Float)lhs, (MelVector)rhs, kind );
			else if(rhs instanceof MelMatrix)
				return floatBinaryOperation( (Float)lhs, (MelMatrix)rhs, kind );
		} else if(lhs instanceof Double) {
			if(rhs instanceof Boolean)
				return doubleBinaryOperation( (Double)lhs, new Double(((Boolean)rhs).booleanValue()?1.0D:0.0D), kind );
			else if(rhs instanceof Number)
				return doubleBinaryOperation( (Double)lhs, new Double(((Number)rhs).doubleValue()), kind );
			else if(rhs instanceof MelVector)
				return floatBinaryOperation( new Float(((Double)lhs).floatValue()), (MelVector)rhs, kind );
			else if(rhs instanceof MelMatrix)
				return floatBinaryOperation( new Float(((Double)lhs).floatValue()), (MelMatrix)rhs, kind );
		} else if(lhs instanceof MelVector) {
			if(rhs instanceof Boolean)
				return vectorBinaryOperation( (MelVector)lhs, new Integer(((Boolean)rhs).booleanValue()?1:0), kind );
			else if(rhs instanceof Number)
				return vectorBinaryOperation( (MelVector)lhs, (Number)rhs, kind );
			else if(rhs instanceof MelVector)
				return vectorBinaryOperation( (MelVector)lhs, (MelVector)rhs, kind );
			else if(rhs instanceof MelMatrix)
				return vectorBinaryOperation( (MelVector)lhs, (MelMatrix)rhs, kind );
		} else if(lhs instanceof MelMatrix) {
			if(rhs instanceof Boolean)
				return matrixBinaryOperation( (MelMatrix)lhs, new Integer(((Boolean)rhs).booleanValue()?1:0), kind );
			else if(rhs instanceof Number)
				return matrixBinaryOperation( (MelMatrix)lhs, (Number)rhs, kind );
			else if(rhs instanceof MelVector)
				return matrixBinaryOperation( (MelMatrix)lhs, (MelVector)rhs, kind );
			else if(rhs instanceof MelMatrix)
				return matrixBinaryOperation( (MelMatrix)lhs, (MelMatrix)rhs, kind );
		} else if(lhs instanceof String)
			return stringBinaryOperation( (String)lhs, (String)rhs, kind);
		throw new EvalError("Invalid types in binary operator:" +lhs.getClass().getName()+"  "+rhs.getClass().getName());
	}

    // returns Object covering both Integer and Boolean return types
    static Object intBinaryOperation(Integer I1, Integer I2, int kind)
    {
        int lhs = I1.intValue();
        int rhs = I2.intValue();

        switch(kind)
        {
            // boolean
            case Tree.LT:
                return new Boolean(lhs < rhs);

            case Tree.GT:
                return new Boolean(lhs > rhs);

            case Tree.EQ:
                return new Boolean(lhs == rhs);

            case Tree.LE:
                return new Boolean(lhs <= rhs);

            case Tree.GE:
                return new Boolean(lhs >= rhs);

            case Tree.NE:
                return new Boolean(lhs != rhs);

            // arithmetic
            case Tree.PLUS:
                return new Integer(lhs + rhs);

            case Tree.MINUS:
                return new Integer(lhs - rhs);

            case Tree.MUL :
                return new Integer(lhs * rhs);

            case Tree.DIV:
                return new Integer(lhs / rhs);

            case Tree.MOD:
                return new Integer(lhs % rhs);

            // bitwise
            case Tree.SL:
                return new Integer(lhs << rhs);

            case Tree.SR:
                return new Integer(lhs >> rhs);

            case Tree.USR:
                return new Integer(lhs >>> rhs);

            case Tree.BITAND:
                return new Integer(lhs & rhs);

            case Tree.BITOR:
                return new Integer(lhs | rhs);

            case Tree.BITXOR:
                return new Integer(lhs ^ rhs);

            case Tree.AND:
                return new Integer((lhs!=0  && rhs!=0)?1:0);

            case Tree.OR:
                return new Integer((lhs!=0 || rhs!=0)?1:0);

            default:
                throw new InterpreterError("Unimplemented binary integer operator");
        }
    }

    static Object intBinaryOperation(Integer I, MelVector v, int kind)
    {
        int lhs = I.intValue();
        float rhs = v.length();
 
        switch(kind)
        {
            // boolean
            case Tree.LT:
                return new Boolean(lhs < rhs);

            case Tree.GT:
                return new Boolean(lhs > rhs);

            case Tree.EQ:
                return new Boolean(lhs == rhs);

            case Tree.LE:
                return new Boolean(lhs <= rhs);

            case Tree.GE:
                return new Boolean(lhs >= rhs);

            case Tree.NE:
                return new Boolean(lhs != rhs);

            // arithmetic
            case Tree.PLUS:
                return v.add(lhs);

            case Tree.MINUS:
                return (new MelVector(lhs)).sub(v);

            case Tree.MUL :
                return v.mul(lhs);

            case Tree.DIV:
                return new Float(lhs / rhs);

            case Tree.MOD:
                return new Float(lhs % rhs);

            case Tree.AND:
               return new Integer((lhs!=0  && (int)rhs!=0)?1:0);

            case Tree.OR:
                return new Integer((lhs!=0 || (int)rhs!=0)?1:0);

             default:
                throw new InterpreterError("Unimplemented binary integer/vector operator");
        }
    }

    static Object intBinaryOperation(Integer I, MelMatrix m, int kind)
    {
        int lhs = I.intValue();
        float rhs = m.length();
 
        switch(kind)
        {
            // boolean
            case Tree.LT:
                return new Boolean(lhs < rhs);

            case Tree.GT:
                return new Boolean(lhs > rhs);

            case Tree.EQ:
                return new Boolean(lhs == rhs);

            case Tree.LE:
                return new Boolean(lhs <= rhs);

            case Tree.GE:
                return new Boolean(lhs >= rhs);

            case Tree.NE:
                return new Boolean(lhs != rhs);

            // arithmetic
            case Tree.PLUS:
                return ConvUtils.toMatrix(lhs).add(m);

            case Tree.MINUS:
                return ConvUtils.toMatrix(lhs).sub(m);

            case Tree.MUL :
                return m.mul(lhs);

            case Tree.DIV:
                return new Float(lhs / rhs);

            case Tree.MOD:
                return new Float(lhs % rhs);

            case Tree.AND:
              return new Integer((lhs!=0  && (int)rhs!=0)?1:0);

            case Tree.OR:
                return new Integer((lhs!=0 || (int)rhs!=0)?1:0);

             default:
                throw new InterpreterError("Unimplemented binary integer/matrix operator");
        }
    }

    // returns Object covering both Double and Boolean return types
    static Object doubleBinaryOperation(Double D1, Double D2, int kind)
        throws EvalError
    {
        double lhs = D1.doubleValue();
        double rhs = D2.doubleValue();

        switch(kind)
        {
            // boolean
            case Tree.LT:
                return new Boolean(lhs < rhs);

            case Tree.GT:
                return new Boolean(lhs > rhs);

            case Tree.EQ:
                return new Boolean(lhs == rhs);

            case Tree.LE:
                return new Boolean(lhs <= rhs);

            case Tree.GE:
                return new Boolean(lhs >= rhs);

            case Tree.NE:
                return new Boolean(lhs != rhs);

            // arithmetic
            case Tree.PLUS:
                return new Double(lhs + rhs);

            case Tree.MINUS:
                return new Double(lhs - rhs);

            case Tree.MUL:
                return new Double(lhs * rhs);

            case Tree.DIV:
                return new Double(lhs / rhs);

            case Tree.MOD:
                return new Double(lhs % rhs);

            // can't shift floating-point values
            case Tree.SL:
            case Tree.SR:
            case Tree.USR:
                throw new EvalError("Can't shift doubles");

            case Tree.AND:
              return new Integer(((int)lhs!=0  && (int)rhs!=0)?1:0);

            case Tree.OR:
                return new Integer(((int)lhs!=0 || (int)rhs!=0)?1:0);

            default:
                throw new InterpreterError("Unimplemented binary double operator");
        }
    }
    static Object floatBinaryOperation(Float F, MelVector v, int kind)
    {
        float lhs = F.floatValue();
        float rhs = v.length();
 
        switch(kind)
        {
            // boolean
            case Tree.LT:
                return new Boolean(lhs < rhs);

            case Tree.GT:
                return new Boolean(lhs > rhs);

            case Tree.EQ:
                return new Boolean(lhs == rhs);

            case Tree.LE:
                return new Boolean(lhs <= rhs);

            case Tree.GE:
                return new Boolean(lhs >= rhs);

            case Tree.NE:
                return new Boolean(lhs != rhs);

            // arithmetic
            case Tree.PLUS:
                return v.add(lhs);

            case Tree.MINUS:
                return (new MelVector(lhs)).sub(v);

            case Tree.MUL :
                return v.mul(lhs);

            case Tree.DIV:
                return new Float(lhs / rhs);

            case Tree.MOD:
                return new Float(lhs % rhs);

            case Tree.AND:
               return new Integer(((int)lhs!=0  && (int)rhs!=0)?1:0);

            case Tree.OR:
                return new Integer(((int)lhs!=0 || (int)rhs!=0)?1:0);

             default:
                throw new InterpreterError("Unimplemented binary float/vector operator");
        }
    }


    static Object floatBinaryOperation(Float F, MelMatrix m, int kind)
    {
        float lhs = F.floatValue();
        float rhs = m.length();
 
        switch(kind)
        {
            // boolean
            case Tree.LT:
                return new Boolean(lhs < rhs);

            case Tree.GT:
                return new Boolean(lhs > rhs);

            case Tree.EQ:
                return new Boolean(lhs == rhs);

            case Tree.LE:
                return new Boolean(lhs <= rhs);

            case Tree.GE:
                return new Boolean(lhs >= rhs);

            case Tree.NE:
                return new Boolean(lhs != rhs);

            // arithmetic
            case Tree.PLUS:
                return ConvUtils.toMatrix(lhs).add(m);

            case Tree.MINUS:
                return ConvUtils.toMatrix(lhs).sub(m);

            case Tree.MUL :
                return m.mul(lhs);

            case Tree.DIV:
                return new Float(lhs / rhs);

            case Tree.MOD:
                return new Float(lhs % rhs);

            case Tree.AND:
              return new Integer(((int)lhs!=0  && (int)rhs!=0)?1:0);

            case Tree.OR:
                return new Integer(((int)lhs!=0 || (int)rhs!=0)?1:0);

             default:
                throw new InterpreterError("Unimplemented binary integer/matrix operator");
        }
    }

   // returns Object covering both Long and Boolean return types
    static Object floatBinaryOperation(Float F1, Float F2, int kind)
        throws EvalError
    {
        float lhs = F1.floatValue();
        float rhs = F2.floatValue();

        switch(kind)
        {
            // boolean
            case Tree.LT:
                return new Boolean(lhs < rhs);

            case Tree.GT:
                return new Boolean(lhs > rhs);

            case Tree.EQ:
                return new Boolean(lhs == rhs);

            case Tree.LE:
                return new Boolean(lhs <= rhs);

            case Tree.GE:
                return new Boolean(lhs >= rhs);

            case Tree.NE:
                return new Boolean(lhs != rhs);

            // arithmetic
            case Tree.PLUS:
                return new Float(lhs + rhs);

            case Tree.MINUS:
                return new Float(lhs - rhs);

            case Tree.MUL:
                return new Float(lhs * rhs);

            case Tree.DIV:
                return new Float(lhs / rhs);

            case Tree.MOD:
                return new Float(lhs % rhs);

            // can't shift floats
            case Tree.SL:
            case Tree.SR:
            case Tree.USR:
                throw new EvalError("Can't shift floats ");

            case Tree.AND:
              return new Integer(((int)lhs!=0  && (int)rhs!=0)?1:0);

            case Tree.OR:
                return new Integer(((int)lhs!=0 || (int)rhs!=0)?1:0);

            default:
                throw new InterpreterError("Unimplemented binary float operator");
        }
    }
    
    static Object vectorBinaryOperation(MelVector v, Number N, int kind)
    {
        float lhs = v.length();
        float rhs = N.floatValue();
 
        switch(kind)
        {
            // boolean
            case Tree.LT:
                return new Boolean(lhs < rhs);

            case Tree.GT:
                return new Boolean(lhs > rhs);

            case Tree.EQ:
                return new Boolean(lhs == rhs);

            case Tree.LE:
                return new Boolean(lhs <= rhs);

            case Tree.GE:
                return new Boolean(lhs >= rhs);

            case Tree.NE:
                return new Boolean(lhs != rhs);

            // arithmetic
            case Tree.PLUS:
                return v.add(rhs);

            case Tree.MINUS:
                return v.sub(new MelVector(rhs));

            case Tree.MUL :
                return v.mul(rhs);

            case Tree.DIV:
                return v.div(rhs);

            case Tree.MOD:
                return v.mod(rhs);
                
            case Tree.BITXOR:
               return v.pow(new MelVector(rhs));

            case Tree.AND:
              return new Integer(((int)lhs!=0  && (int)rhs!=0)?1:0);

            case Tree.OR:
                return new Integer(((int)lhs!=0 || (int)rhs!=0)?1:0);

             default:
                throw new InterpreterError("Unimplemented binary vector/float operator");
        }
    }
   
    static Object vectorBinaryOperation(MelVector v1, MelVector v2, int kind)
    {
        float lhs = v1.length();
        float rhs = v2.length();
 
        switch(kind)
        {
            // boolean
            case Tree.LT:
                return new Boolean(lhs < rhs);

            case Tree.GT:
                return new Boolean(lhs > rhs);

            case Tree.EQ:
                return new Boolean(v1.x==v2.x && v1.y==v2.y && v1.z==v2.z);

            case Tree.LE:
                return new Boolean(v1.x!=v2.x || v1.y!=v2.y || v1.z!=v2.z);

            case Tree.GE:
                return new Boolean(lhs >= rhs);

            case Tree.NE:
                return new Boolean(lhs != rhs);

            // arithmetic
            case Tree.PLUS:
                return v1.add(v2);

            case Tree.MINUS:
                return v1.sub(v2);

            case Tree.MUL :
                return new Float(v1.mul(v2));

            case Tree.DIV:
                return v1.div(v2);

            case Tree.MOD:
                return v1.mod(v2);
                
            case Tree.BITXOR:
               return v1.pow(v2);

            case Tree.AND:
              return new Integer(((int)lhs!=0  && (int)rhs!=0)?1:0);

            case Tree.OR:
                return new Integer(((int)lhs!=0 || (int)rhs!=0)?1:0);

             default:
                throw new InterpreterError("Unimplemented binary vector/vector operator");
        }
    }
  
    static Object vectorBinaryOperation(MelVector v, MelMatrix m, int kind)
    {
        float lhs = v.length();
        float rhs = m.length();
 
        switch(kind)
        {
            // boolean
            case Tree.LT:
                return new Boolean(lhs < rhs);

            case Tree.GT:
                return new Boolean(lhs > rhs);

            case Tree.EQ:
                return new Boolean(lhs == rhs);

            case Tree.LE:
                return new Boolean(lhs <= rhs);

            case Tree.GE:
                return new Boolean(lhs >= rhs);

            case Tree.NE:
                return new Boolean(lhs != rhs);

            // arithmetic
            case Tree.PLUS:
                return v.add(ConvUtils.toVector(m));

            case Tree.MINUS:
                return v.sub(ConvUtils.toVector(m));

            case Tree.MUL :
                return new Float(v.mul(ConvUtils.toVector(m)));

            case Tree.DIV:
                return v.div(ConvUtils.toVector(m));

            case Tree.MOD:
                return v.mod(ConvUtils.toVector(m));
                 
            case Tree.BITXOR:
               return v.pow(ConvUtils.toVector(m));

            case Tree.AND:
              return new Integer(((int)lhs!=0  && (int)rhs!=0)?1:0);

            case Tree.OR:
                return new Integer(((int)lhs!=0 || (int)rhs!=0)?1:0);

            default:
                throw new InterpreterError("Unimplemented binary vector/matrix operator");
        }
    }
   
    static Object matrixBinaryOperation(MelMatrix m, Number N, int kind)
    {
        float lhs = m.length();
        float rhs = N.floatValue();
 
        switch(kind)
        {
            case Tree.LT:
                return new Boolean(lhs < rhs);

            case Tree.GT:
                return new Boolean(lhs > rhs);

            case Tree.EQ:
                return new Boolean(lhs == rhs);

            case Tree.LE:
                return new Boolean(lhs <= rhs);

            case Tree.GE:
                return new Boolean(lhs >= rhs);

            case Tree.NE:
                return new Boolean(lhs != rhs);

            // arithmetic
            case Tree.PLUS:
                return ConvUtils.toMatrix(rhs).add(m);

            case Tree.MINUS:
                return m.sub(ConvUtils.toMatrix(rhs));

            case Tree.MUL :
                return m.mul(rhs);

            case Tree.DIV:
                return m.div(rhs);

            case Tree.MOD:
                return m.mod(rhs);

            case Tree.AND:
               return new Integer(((int)lhs!=0  && (int)rhs!=0)?1:0);

            case Tree.OR:
                return new Integer(((int)lhs!=0 || (int)rhs!=0)?1:0);

             default:
                throw new InterpreterError("Unimplemented binary matrix/float operator");
        }
    }
   
    static Object matrixBinaryOperation(MelMatrix m, MelVector v, int kind)
    {
        float lhs = m.length();
        float rhs = v.length();
 
        switch(kind)
        {
           case Tree.LT:
                return new Boolean(lhs < rhs);

            case Tree.GT:
                return new Boolean(lhs > rhs);

            case Tree.EQ:
                return new Boolean(lhs == rhs);

            case Tree.LE:
                return new Boolean(lhs <= rhs);

            case Tree.GE:
                return new Boolean(lhs >= rhs);

            case Tree.NE:
                return new Boolean(lhs != rhs);

            // arithmetic
            case Tree.PLUS:
                return m.add(ConvUtils.toMatrix(v));

            case Tree.MINUS:
                return m.sub(ConvUtils.toMatrix(v));

            case Tree.MUL :
                return m.mul(v);

            case Tree.DIV:
                return m.div(ConvUtils.toMatrix(v));

            case Tree.MOD:
                return m.mod(ConvUtils.toMatrix(v));

            case Tree.AND:
              return new Integer(((int)lhs!=0  && (int)rhs!=0)?1:0);

            case Tree.OR:
                return new Integer(((int)lhs!=0 || (int)rhs!=0)?1:0);

             default:
                throw new InterpreterError("Unimplemented binary matrix/vector operator");
        }
    }
  
    static Object matrixBinaryOperation(MelMatrix m1, MelMatrix m2, int kind)
    {
        float lhs = m1.length();
        float rhs = m2.length();
 
        switch(kind)
        {
            // boolean
            case Tree.LT:
                return new Boolean(lhs < rhs);

            case Tree.GT:
                return new Boolean(lhs > rhs);

            case Tree.EQ:
                return new Boolean(lhs == rhs);

            case Tree.LE:
                return new Boolean(lhs <= rhs);

            case Tree.GE:
                return new Boolean(lhs >= rhs);

            case Tree.NE:
                return new Boolean(lhs != rhs);

            // arithmetic
            case Tree.PLUS:
                 return m1.add(m2);

            case Tree.MINUS:
                return m1.sub(m2);

            case Tree.MUL :
                return m1.mul(m2);

            case Tree.DIV:
                return m1.div(m2);

            case Tree.MOD:
                return m1.mod(m2);

            case Tree.AND:
              return new Integer(((int)lhs!=0  && (int)rhs!=0)?1:0);

            case Tree.OR:
                return new Integer(((int)lhs!=0 || (int)rhs!=0)?1:0);

             default:
                throw new InterpreterError("Unimplemented binary matrix/matrix operator");
        }
    }

    // returns Object covering both Long and Boolean return types
    static Object stringBinaryOperation(String S1, String S2, int kind)
        throws EvalError
    {
         switch(kind)
        {
            // boolean
            case Tree.LT:
                return new Boolean(S1.length() < S2.length());

            case Tree.GT:
                return new Boolean(S1.length() > S2.length());

            case Tree.EQ:
                return new Boolean(S1.length() == S2.length());

            case Tree.LE:
                return new Boolean(S1.length() <= S2.length());

            case Tree.GE:
                return new Boolean(S1.length() >= S2.length());

            case Tree.NE:
                return new Boolean(S1.length() != S2.length());

            // arithmetic
            case Tree.PLUS:
                return S1+S2;

            default:
                throw new InterpreterError("Unimplemented binary string operator");
        }
    }

	/**
		Promote primitive wrapper type to to Integer wrapper type
		Can we use the castPrimitive() (in BSHCastExpression) for this?
	*/
    public static Object promoteToInteger(Object primitive)
    {
        if(primitive instanceof Character)
            return new Integer(((Character)primitive).charValue());
        else if((primitive instanceof Byte) || (primitive instanceof Short))
            return new Integer(((Number)primitive).intValue());

        return primitive;
    }

	/**
		Promote the pair of primitives to the maximum type of the two.
		e.g. [int,long]->[long,long]
	*/
    public static Object[] promotePrimitives(Object lhs, Object rhs)
    {
        lhs = promoteToInteger(lhs);
        rhs = promoteToInteger(rhs);

        if((lhs instanceof Number) && (rhs instanceof Number))
        {
            Number lnum = (Number)lhs;
            Number rnum = (Number)rhs;

            boolean b;

            if((b = (lnum instanceof Double)) || (rnum instanceof Double))
            {
                if(b)
                    rhs = new Double(rnum.doubleValue());
                else
                    lhs = new Double(lnum.doubleValue());
            }
            else if((b = (lnum instanceof Float)) || (rnum instanceof Float))
            {
                if(b)
                    rhs = new Float(rnum.floatValue());
                else
                    lhs = new Float(lnum.floatValue());
            }
        }

        if((lhs instanceof String) || (rhs instanceof String))
        {
            lhs = lhs.toString();
            rhs = rhs.toString();
        }

        return new Object[] { lhs, rhs };
    }

    public static Primitive unaryOperation(Primitive val, int kind)
        throws EvalError
    {
        if(val == NULL)
            throw new EvalError("illegal use of null object or 'null' literal");
        if(val == VOID)
            throw new EvalError("illegal use of undefined object or 'void' literal");

        Class operandType = val.getType();
        Object operand = val.getValue();

		if(operand instanceof Integer)
             return new Primitive(intUnaryOperation((Integer)operand, kind));
        else if(operand instanceof Float)
            return new Primitive(floatUnaryOperation((Float)operand, kind));
        else if(operand instanceof Double)
            return new Primitive(doubleUnaryOperation((Double)operand, kind));
        else if(operand instanceof MelVector)
            return new Primitive(vectorUnaryOperation((MelVector)operand, kind));
        else if(operand instanceof MelMatrix)
            return new Primitive(matrixUnaryOperation((MelMatrix)operand, kind));
        else if(operand instanceof Boolean)
            return new Primitive(booleanUnaryOperation((Boolean)operand, kind));
            throw new InterpreterError("Bad unaryOperation operation for primitive '" + val + "'.");
    }

    public static boolean booleanUnaryOperation(Boolean B, int kind)
    {
        boolean operand = B.booleanValue();
        switch(kind)
        {
            case Tree.NOT:
                return !operand;

            default:
                throw new InterpreterError("bad boolean unaryOperation");
        }
    }

    public static int intUnaryOperation(Integer I, int kind)
    {
        int operand = I.intValue();
        switch(kind)
        {
            case Tree.NOT:
                return operand == 0 ? 1 : 0;

            case Tree.POS:
                return operand;

            case Tree.NEG:
                return -operand;

            case Tree.COMPL:
                return ~operand;

             case Tree.POSTINC:
             case Tree.PREINC:
                return operand + 1;

            case Tree.POSTDEC:
            case Tree.PREDEC:
                return operand - 1;

            default:
                throw new InterpreterError("bad integer unaryOperation");
        }
    }

    public static float floatUnaryOperation(Float F, int kind)
    {
        float operand = F.floatValue();

        switch(kind)
        {
             case Tree.NOT:
                return operand == 0 ? 1 : 0;

           case Tree.POS:
                return operand;

            case Tree.NEG:
                return -operand;

             case Tree.POSTINC:
             case Tree.PREINC:
                return operand + 1;

            case Tree.POSTDEC:
            case Tree.PREDEC:
                return operand - 1;

            default:
                throw new InterpreterError("bad float unaryOperation:"+kind);
        }
    }

    public static double doubleUnaryOperation(Double D, int kind)
    {
        double operand = D.doubleValue();

        switch(kind)
        {
            case Tree.NOT:
                return operand == 0 ? 1 : 0;

            case Tree.POS:
                return operand;

            case Tree.NEG:
                return -operand;

             case Tree.POSTINC:
             case Tree.PREINC:
                return operand + 1;

            case Tree.POSTDEC:
            case Tree.PREDEC:
                return operand - 1;

           default:
                throw new InterpreterError("bad double unaryOperation:"+kind);
        }
    }

   public static Object vectorUnaryOperation(MelVector v, int kind)
    {
        float operand = v.length();

        switch(kind)
        {
			case Tree.NOT:
				return new Integer(operand == 0 ? 1 : 0);
			
			case Tree.POS:
				return v;
			
			case Tree.NEG:
				return v.mul(-1);

            default:
                throw new InterpreterError("bad float unaryOperation:"+kind);
        }
    }

   public static Object matrixUnaryOperation(MelMatrix m, int kind)
    {
        float operand = m.length();

        switch(kind)
        {
			case Tree.NOT:
				return new Integer(operand == 0 ? 1 : 0);
			
			case Tree.POS:
				return m;
			
			case Tree.NEG:
				return m.mul(-1);

            default:
                throw new InterpreterError("bad float unaryOperation:"+kind);
        }
    }

    public int intValue() throws EvalError
    {
        if(value instanceof Number)
            return((Number)value).intValue();
        else
            throw new EvalError("Primitive not a number");
    }

	/**
		Are we a numeric type:
		i.e. not boolean, null, or void
		(but including char)
	*/
	public boolean isNumber() {
		return ( !(value instanceof Boolean) 
			&& !(this == NULL) && !(this == VOID) );
	}

    public Number numberValue() throws EvalError
    {
		Object value = this.value;

        if (value instanceof Number)
            return (Number)value;
        else
            throw new EvalError("Primitive not a number");
    }

	public boolean equals( Object obj ) {
		if ( obj instanceof Primitive )
			return ((Primitive)obj).value.equals( this.value );
		else
			return obj.equals( this.value );
	}

	/**
		Unwrap primitive values and map voids to nulls.
		Normal (non Primitive) types remain unchanged.
		@param obj object type which may be bsh.Primitive
		@return corresponding "normal" Java type, "unwrapping" 
			any bsh.Primitive types to their wrapper types.
	*/
	public static Object unwrap( Object obj ) {
		if ( obj == null )
			return null;

        // map voids to nulls for the outside world
        if(obj == Primitive.VOID)
            return null;

        // unwrap primitives
        if(obj instanceof Primitive)
            return((Primitive)obj).getValue();
        else
            return obj;
	}


}
