package jmel.interp;

import jmel.tree.Tree;
import jmel.*;
import java.lang.reflect.InvocationTargetException;
import java.io.PrintStream;


/**
	TargetError is an EvalError that wraps an exception thrown by the script	
	(or by code called from the script).  TargetErrors indicate exceptions 
	which can be caught within the script itself, whereas a general EvalError 
	indicates that the script cannot be evaluated further for some reason.
	
	If the exception is caught within the script it is automatically unwrapped,
	so the code looks like normal Java code.  If the TargetError is thrown
	from the eval() or interpreter.eval() method it may be caught and unwrapped
	to determine what exception was thrown.
*/
/*
	Implementation note: 
	Although it is easy to catch exceptions generated by called Java code
	and wrap them (e.g. in the method dispatching code), we must be careful 
	do the same with arbitrary exceptions that we generate in interpreted
	code, e.g. ArithmeticException, ClassCastException.

	Also an important location to look at is BSHMethodInvocation.  There
	we catch eval errors and rethrow them to compound the location information
*/
public class TargetError extends EvalError 
{
	Throwable target;
	boolean inNativeCode;

	public TargetError(String msg, Throwable t, Tree tree, boolean inNativeCode )
	{
		super(msg, tree);
		target = t;
		this.inNativeCode = inNativeCode;
	}

	public TargetError(String msg, Throwable t, Tree tree )
	{
		super(msg, tree);
		target = t;
	}

	public TargetError(Throwable t, Tree tree )
	{
		this("TargetError", t, tree);
	}

	/**
		If you're going to use this please catch and re-throw the exception
		in an AST and add the node...
		@see reThrow()
	*/
	public TargetError( String s, Throwable t )
	{
		this(s, t, null, false);
	}

	public Throwable getTarget()
	{
		// check for easy mistake
		if(target instanceof InvocationTargetException)
			return((InvocationTargetException)target).getTargetException();
		else
			return target;
	}

	public String toString() {

		return super.toString() 
			+ "\nTarget exception: " + 
			printTargetError( target );
	}

    public void printStackTrace() { 
		printStackTrace( false, System.err );
	}

    public void printStackTrace( PrintStream out ) { 
		printStackTrace( false, out );
	}

    public void printStackTrace( boolean debug, PrintStream out ) {
		if ( debug ) {
			super.printStackTrace( out );
			System.out.println("--- Target Stack Trace ---");
		}
		target.printStackTrace( out );
	}

	/**
		Generate a printable string showing the wrapped target exception.
		If the proxy mechanism is available, allow the extended print to
		check for UndeclaredThrowableException and print that embedded error.
	*/
	public String printTargetError( Throwable t ) 
	{
		return target.toString();
	}

	/**
		Return true if the TargetError was generated from native code.
		e.g. if the script called into a compiled java class which threw
		the excpetion.  We distinguish so that we can print the stack trace
		for the native code case... the stack trace would not be useful if
		the exception was generated by the script.  e.g. if the script
		explicitly threw an exception... (the stack trace would simply point
		to the bsh internals which generated the exception).
	*/
	public boolean inNativeCode() { 
		return inNativeCode; 
	}
}

