package jmel.parser;

import jmel.tree.*;
import jmel.code.*;
import jmel.util.*;
import jmel.tree.Tree.*;


/**
 * The parser maps a token sequence into an abstract syntax
 *  tree. It operates by recursive descent, with code derived
 *  systematically from an LL(1) grammar. For efficiency reasons, an
 *  operator precedence scheme is used for parsing binary operation
 *  expressions.
 */
public class Parser implements Tokens {
	
	/**
	* The number of precedence levels of infix operators.
	*/
	private static final int infixPrecedenceLevels = 10;
	
 	/**
	* The scanner used for lexical analysis.
	*/
	protected Scanner S;
	
	/**
	* The factory to be used for abstract syntax tree construction.
	*/
	protected TreeMaker F;

    /**
     * The log to be used for error diagnostics.
     */
    protected Log log;

	/**
	* Construct a parser from a given scanner, tree factory and log.
	*/
	public Parser(Scanner S, TreeMaker F, Log log) {
		this.S = S;
		this.F = F;
		this.log = log;
	}
	
	/**
	* When terms are parsed, the mode determines which is expected:
	*     mode = EXPR        : an expression
	*     mode = TYPE        : a type
	*     mode = NOPARAMS    : no parameters allowed for type
	*/
	protected static final int EXPR = 1;
	protected static final int TYPE = 2;
	protected static final int NOPARAMS = 4;
	protected static final int COMMAND = 5;
	
	/**
	* The current mode.
	*/
	protected int mode = 0;

	/**
	* The mode of the term that was parsed last.
	*/
	protected int lastmode = 0;
	public static Tree errorTree = new Tree.Erroneous();
	
	private Hashtable methods = new Hashtable();

	/**
	* Skip forward until a suitable stop token is found.
	*/
	private void skip() {
		int nbraces = 0;
		int nparens = 0;
		while (true) {
			switch (S.token) {
				case EOF:
				
				case SEMI:
					if (nbraces == 0 && nparens == 0) return;
					break;
				
				case RBRACE:
					if (nbraces == 0) return;
					nbraces--;
					break;
				
				case RPAREN:
					if (nparens > 0) nparens--;
					break;
				
				case LBRACE:
					nbraces++;
					break;
				
				case LPAREN:
					nparens++;
					break;
			}
			S.nextToken();
		}
	}

	/**
	* Generate a syntax error at given position using the given argument
	*  unless one was already reported at the same position, then skip.
	*/
	protected Tree syntaxError(int pos, String key, String arg) {
		if (pos != S.errPos) log.error(pos, key, arg);
		skip();
		S.errPos = pos;
		if (matrixParsing) matrixParsing = false;
		return errorTree;
	}
	
	/**
	* Generate a syntax error at given position unless one was already
	*  reported at the same position, then skip.
	*/
	private Tree syntaxError(int pos, String key) {
		return syntaxError(pos, key, null);
	}

	/**
	* Generate a syntax error at current position unless one was already reported
	*  at the same position, then skip.
	*/
	private Tree syntaxError(String key) {
		return syntaxError(S.pos, key, null);
	}

    /**
	* Generate a syntax error at current position unless one was already reported
	*  at the same position, then skip.
	*/
	private Tree syntaxError(String key, String arg) {
		return syntaxError(S.pos, key, arg);
	}

    /**
      * If next input token matches given token, skip it, otherwise report an error.
      */
	protected void accept(int token) {
		if (S.token == token) {
			S.nextToken();
		} else {
			int pos = Position.line(S.pos) > Position.line(S.lastPos) ? S.lastPos : S.pos;
			syntaxError(pos, "expected", S.token2string(token));
			if (S.token == token) S.nextToken();
		}
	}

	/**
	* Report an illegal start of expression/type error at given position.
	*/
	Tree illegal(int pos) {
		if ((mode & EXPR) != 0)
			return syntaxError(pos, "illegal.start.of.expr");
		else
			return syntaxError(pos, "illegal.start.of.type");
	}

	/**
	* Report an illegal start of expression/type error at current position.
	*/
	Tree illegal() {
		return illegal(S.pos);
	}

	boolean isZero(String s) {
		char[] cs = s.toCharArray();
		int i = 0;
		while (i < cs.length && (cs[i] == '0' || cs[i] == '.')) i++;
		return !(i < cs.length && ('1' <= cs[i] && cs[i] <= '9'));
	}
	
	String strval(Name prefix) {
		String s = S.stringVal();
		return prefix.len == 0 ? s : prefix + s;
	}

	/**
	* Construct an operation or type test node.
	*/
	private Tree makeOp(int pos, int topOp, Tree od1, Tree od2) {
		return F.at(pos).Operation(optag(topOp), List.make(od1, od2));
	}
	
    /**
	* If tree is a concatenation of string literals, replace it
	*  by a single literal representing the concatenated string.
	*/
	private static StringBuffer foldStrings(Tree tree) {
		if (tree.tag == Tree.LITERAL) {
			Literal lit = (Literal) tree;
		} else if (tree.tag == Tree.PLUS) {
			Operation op = (Operation) tree;
			if (op.args.tail.nonEmpty()) {
				Tree fst = (Tree) op.args.head;
				Tree snd = (Tree) op.args.tail.head;
				if (snd.tag == Tree.LITERAL) {
					Literal lit = (Literal) snd;
					if (lit.typetag == Type.CLASS) {
						StringBuffer buf = foldStrings(fst);
						if (buf != null) {
							return buf.append(lit.value);
						}
					}
				}
			}
		}
		return null;
	}

	/**
	* optimization: To save allocating a new operand/operator stack
	*  for every binary operation, we use supplys.
	*/
	ListBuffer odStackSupply = new ListBuffer();
	ListBuffer opStackSupply = new ListBuffer();
	
	private Tree[] newOdStack() {
		if (odStackSupply.elems == odStackSupply.last)
			odStackSupply.append(new Tree[infixPrecedenceLevels + 1]);
		Tree[] odStack = (Tree[]) odStackSupply.elems.head;
		odStackSupply.elems = odStackSupply.elems.tail;
		return odStack;
	}
	
	private int[] newOpStack() {
		if (opStackSupply.elems == opStackSupply.last)
			opStackSupply.append(new int[infixPrecedenceLevels + 1]);
		int[] opStack = (int[]) opStackSupply.elems.head;
		opStackSupply.elems = opStackSupply.elems.tail;
		return opStack;
	}

	/**
	* Name = NAME
	*/
	protected Name name() {
		if (S.token == NAME) {
			Name name = S.name;
			S.nextToken();
			return name;
		} else {
			accept(NAME);
			return Names.error;
		}
	}

	/**
	* Ident = IDENTIFIER
	*/
	protected Name ident() {
		if (S.token == IDENTIFIER) {
			Name name = S.name;
			S.nextToken();
			return name;
		} else {
			accept(IDENTIFIER);
			return Names.error;
		}
	}
	
	protected Name var() {
		if (S.token == VARIABLE) {
			Name name = S.name;
			S.nextToken();
			return name;
		} else {
			accept(VARIABLE);
			return Names.error;
		}
	}
	
	/**
	* Qualident = Ident { DOT Ident }
	*/
	public Tree qualident() {
		Tree t = F.at(S.pos).Ident(ident());
		while (S.token == DOT) {
			int pos = S.pos;
			S.nextToken();
			t = F.at(pos).Select(t, ident());
		}
		return t;
	}
	/**
	* Qualname = Name { DOT Name }
	*/
	public Tree qualname() {
		Tree t = F.at(S.pos).Ident(name());
		while (S.token == DOT) {
			int pos = S.pos;
			S.nextToken();
			t = F.at(pos).Select(t, name());
		}
		return t;
	}

	/**
	* Literal =
	*	   INTLITERAL
	*	 | FLOATLITERAL
	*	 | DOUBLELITERAL
	*	 | STRINGLITERAL
	*    | TRUE
	*    | FALSE
	*    | ON
	*    | OFF
	*    | YES
	*    | NO
	*/
	Tree literal(Name prefix) {
		int pos = S.pos;
		Tree t = errorTree;
		switch (S.token) {
			case INTLITERAL:
				try {
					t = F.at(pos).Literal(Type.INT, new Integer(Convert.string2int(strval(prefix), S.radix)));
				} catch (NumberFormatException ex) {
					log.error(S.pos, "int.number.too.large", strval(prefix));
				}
				break;

			case FLOATLITERAL:
			{
				Float n = Float.valueOf(S.stringVal());
				if (n.floatValue() == 0.0F && !isZero(S.stringVal()))
					log.error(S.pos, "fp.number.too.small");
				else if (n.floatValue() == Float.POSITIVE_INFINITY)
					log.error(S.pos, "fp.number.too.large");
				else
					t = F.at(pos).Literal(Type.FLOAT, n);
				break;
			}

			case DOUBLELITERAL:
			{
				Double n = Double.valueOf(S.stringVal());
				if (n.doubleValue() == 0.0 && !isZero(S.stringVal()))
					log.error(S.pos, "fp.number.too.small");
				else if (n.doubleValue() == Double.POSITIVE_INFINITY)
					log.error(S.pos, "fp.number.too.large");
				else
					t = F.at(pos).Literal(Type.DOUBLE, n);
				break;
			}

			case STRINGLITERAL:
				t = F.at(pos).Literal(Type.CLASS, S.stringVal());
				break;
			
			case TRUE:			
			case ON:
			case YES:
				t = F.at(pos).Ident(Name.fromString("true"));
				break;

			
			case FALSE:			
			case OFF:
			case NO:
				t = F.at(pos).Ident(Name.fromString("false"));
				break;

			default:
				throw new InternalError();
		}
		S.nextToken();
		return t;
	}

	/**
	* terms can be either expressions or types.
	*/
	public Tree expression() {
		return term(EXPR);
	}

	public Tree type() {
		return term(TYPE);
	}

	public Tree command() {
		return term(COMMAND);
	}
	
	/**
	*  Expression = Expression1 [ExpressionRest]
	*  ExpressionRest = [AssignmentOperator Expression1]
	*  AssignmentOperator = "=" | "+=" | "-=" | "*=" | "/=" |  "&=" | "|=" | "^=" | "%=" | "<<=" | ">>=" | ">>>="
	*  Type = Type1
	*  TypeNoParams = TypeNoParams1
	*  StatementExpression = Expression
	*  ConstantExpression = Expression
	*/
	protected Tree term(int newmode) {
		int prevmode = mode;
		mode = newmode;
		Tree t = term();
		lastmode = mode;
		mode = prevmode;
		return t;
	}
	
	Tree term() {
		Tree t = term1();
		if ((mode & EXPR) != 0 && S.token == EQ || PLUSEQ <= S.token && S.token <= GTGTGTEQ)
			return termRest(t);
		else
			return t;
	}
	
	Tree termRest(Tree t) {
		switch (S.token) {
			case EQ:
			{
				int pos = S.pos;
				S.nextToken();
				mode = EXPR;
				Tree t1 = term();
				return F.at(pos).Assign(t, t1);
			}
			
			case PLUSEQ:
			case SUBEQ:
			case STAREQ:
			case SLASHEQ:
			case PERCENTEQ:
			case AMPEQ:
			case BAREQ:
			case CARETEQ:
			case LTLTEQ:
			case GTGTEQ:
			case GTGTGTEQ:
				int pos = S.pos;
				int token = S.token;
				S.nextToken();
				mode = EXPR;
				Tree t1 = term();
				return F.at(pos).Assignop(optag(token), t, t1);
			
			default:
				return t;		
		}
	}

	/**
	* Expression1   = Expression2 [Expression1Rest]
	*  Type1         = Type2
	*  TypeNoParams1 = TypeNoParams2
	*/
	Tree term1() {
		Tree t = term2();
		if ((mode & EXPR) != 0 & S.token == QUES) {
			mode = EXPR;
			return term1Rest(t);
		} else {
			return t;
		}
	}

	/**
	* Expression1Rest = ["?" Expression ":" Expression1]
	*/
	Tree term1Rest(Tree t) {
		if (S.token == QUES) {
			int pos = S.pos;
			S.nextToken();
			Tree t1 = term();
			accept(COLON);
			Tree t2 = term1();
			return F.at(pos).Conditional(Tree.CONDEXPR, t, t1, t2);
		} else {
			return t;
		}
	}

	/**
	* Expression2   = Expression3 [Expression2Rest]
	*  Type2         = Type3
	*  TypeNoParams2 = TypeNoParams3
	*/
	Tree term2() {
		Tree t = term3();
		if ((mode & EXPR) != 0 && prec(S.token) >= TreeInfo.orPrec) {
			mode = EXPR;
			return term2Rest(t, TreeInfo.orPrec);
		} else {
			return t;
		}
	}
	
	Tree term2Rest(Tree t, int minprec) {
		List savedOd = odStackSupply.elems;
		Tree[] odStack = newOdStack();
		List savedOp = opStackSupply.elems;
		int[] opStack = newOpStack();
		int top = 0;
		odStack[0] = t;
		int startPos = S.pos;
		int topOp = ERROR;
		while (prec(S.token) >= minprec) {
			opStack[top] = topOp;
			top++;
			topOp = S.token;
			int pos = S.pos;
			S.nextToken();
			odStack[top] = term3();
			while (top > 0 && prec(topOp) >= prec(S.token)) {
				odStack[top - 1] = makeOp(pos, topOp, odStack[top - 1], odStack[top]);
				top--;
				topOp = opStack[top];
			}
		}
		if (Util.checks)
			Util.assert(top == 0);
		t = odStack[0];
		if (t.tag == Tree.PLUS) {
			StringBuffer buf = foldStrings(t);
			if (buf != null) {
				t = F.at(startPos).Literal(Type.CLASS, buf.toString());
			}
		}
		odStackSupply.elems = savedOd;
		opStackSupply.elems = savedOp;
		return t;
	}

    /**
      * Expression3    = PrefixOp Expression3
      *                 | "(" Expr | TypeNoParams ")" Expression3
      *                 | Primary {Selector} {PostfixOp}
      *  Primary        = "(" Expression ")"
      *                 | Literal
      *                 | Ident { "." Ident } Arguments
      *  PrefixOp       = "++" | "--" | "!" | "~" | "+" | "-"
      *  PostfixOp      = "++" | "--"
      *  Type3          = Ident { "." Ident } [TypeArguments] {TypeSelector} BracketsOpt
      *                 | BasicType
      *  TypeNoParams3  = Ident { "." Ident } BracketsOpt
      *  Selector       = "." Ident [Arguments]
      *                 | "[" Expression "]"
      *  TypeSelector   = "." Ident [TypeArguments]
      *  SuperSuffix    = Arguments | "." Ident [Arguments]
      */
	public Tree term3() {
		int pos = S.pos;
		Tree t;
		switch (S.token) {
			case CATCH:
				return catchClause();
			
			case PLUSPLUS:
			case SUBSUB:
			case BANG:
			case TILDE:
			case PLUS:
			case SUB:
				if ((mode & EXPR) != 0) {
					mode = EXPR;
					int token = S.token;
					S.nextToken();
					if (token == SUB && (S.token == INTLITERAL || S.token == LONGLITERAL) && S.radix == 10) {
						t = literal(Names.hyphen);
					} else {
					    t = term3();
					    return F.at(pos).Operation(unoptag(token), List.make(t));
					}
				} else
					return illegal();
				break;

			case LPAREN:
				if ((mode & EXPR) != 0) {
					S.nextToken();
					mode = EXPR | TYPE | NOPARAMS;
					t = term3();
					t = termRest(term1Rest(term2Rest(t, TreeInfo.orPrec)));
					accept(RPAREN);
					lastmode = mode;
					mode = EXPR;
					if ((lastmode & EXPR) == 0) {
						Tree t1 = term3();
						return F.at(pos).TypeCast(t, t1);
					} else if ((lastmode & TYPE) != 0) {
						switch (S.token) {
							case BANG:
							case TILDE:
							case LPAREN:
							case INTLITERAL:
							case FLOATLITERAL:
							case DOUBLELITERAL:
							case STRINGLITERAL:
							case TRUE:
							case FALSE:
							case YES:
							case NO:
							case ON:
							case OFF:
							case NAME:
							case IDENTIFIER:
							case VARIABLE:
							case INT:
							case FLOAT:
							case DOUBLE:
								Tree t1 = term3();
								return F.at(pos).TypeCast(t, t1);
						}
					}
				} else
					return illegal();
				break;
			
			case INTLITERAL:
			case FLOATLITERAL:
			case DOUBLELITERAL:
			case STRINGLITERAL:
			case TRUE:
			case FALSE:
			case YES:
			case NO:
			case ON:
			case OFF:
				if ((mode & EXPR) != 0) {
					mode = EXPR;
					t = literal(Names.empty);
				} else
					return illegal();
				break;

			case IDENTIFIER:
			{
				Name id = ident();
				t = F.at(S.pos).Ident(id);
				loop:
				while (true) {
					pos = S.pos;
					switch (S.token) {
						case LBRACKET:
							S.nextToken();
							if (S.token == RBRACKET) {
								S.nextToken();
								t = bracketsSuffix(bracketsOpt(F.at(pos).TypeArray(t,0)));
							} else {
								if ((mode & EXPR) != 0) {
									mode = EXPR;
									Tree t1 = term();
									t = F.at(pos).Indexed(t, t1);
								}
								accept(RBRACKET);
							}
							break loop;
				
						case DOT:
							S.nextToken();
							if ((mode & EXPR) != 0) {
							}
							t = F.at(pos).Select(t, name());
							break;
								
						default:
							break loop;								
					}
				}
				break;
			}

			case LBRACE:
				return arrayInitializer(null);

			case LTLT:
				return matrixInitializer();
				
			case VARIABLE:
			{
				Name v = var();
				t = F.at(S.pos).Ident(v);
				loop:
				while (true) {
					pos = S.pos;
					switch (S.token) {
						case LBRACKET:
							S.nextToken();
							if (S.token == RBRACKET) {
								S.nextToken();
								t = bracketsSuffix(bracketsOpt(F.at(pos).TypeArray(t,0)));
							} else {
								if ((mode & EXPR) != 0) {
									mode = EXPR;
									Tree t1 = term();
									t = F.at(pos).Indexed(t, t1);
								}
								accept(RBRACKET);
							}
							break loop;
							
						case DOT:
							S.nextToken();
							if ((mode & EXPR) != 0) {
							}
							t = F.at(pos).Select(t, name());
							break;
								
						default:
							break loop;								
					}
				}
				break;
			}
			
			case NAME:
				t = qualname();
				boolean withDot = false;
				loop:
				while (true) {
					pos = S.pos;
					if (S.token == LPAREN) {
						if (!withDot && (mode & EXPR) != 0) {
							mode = EXPR;
							t = arguments(t);				
						} 
					} else {
						switch (S.token) {
							case LBRACKET:
								S.nextToken();
								if (S.token == RBRACKET) {
									S.nextToken();
									t = bracketsSuffix(bracketsOpt(F.at(pos).TypeArray(t,0)));
								} else {
									if ((mode & EXPR) != 0) {
										mode = EXPR;
										Tree t1 = term();
										t = F.at(pos).Indexed(t, t1);
									}
									accept(RBRACKET);
								}
								break loop;
								
							case DOT:
								withDot = true;
								S.nextToken();
								if ((mode & EXPR) != 0) {
								}
								t = F.at(pos).Select(t, name());
								break;
									
							default:
								if (!withDot && mode == COMMAND)
									t = arguments(t);
								break loop;
						}
					}
				}
				break;

			case QUOTE:
			{
				S.nextToken();
				if (S.token != NAME) return syntaxError(pos, "expected", S.token2string(NAME));
				withQuote = true;
				boolean isMethod = methods.get(S.name.toString()) != null;
				Tree cmd = command();
				if (isMethod) 
					t = cmd;
				else
					t = F.at(pos).Command(checkExprStat(cmd), false);
				if (S.token == RPAREN) S.nextToken();
				accept(QUOTE);
				withQuote = false;
				break;
			}

			case INT:
			case FLOAT:
			case DOUBLE:
			case STRING:
			case VECTOR:
			case MATRIX:
				t = basicType();
				if (S.token == LPAREN) {
					Tree t1 = term3();
					return F.at(pos).TypeCast(t, t1);
				}							
				break;

			case GLOBAL:
				S.nextToken();
				t = type();
				Name v = var();
				t = bracketsOpt(t);
				t = F.at(S.pos).GlobalRef(v, t);
				break;
				
			default:
				return illegal();			
		}	
		while (true) {
			int pos1 = S.pos;
			if (S.token == LBRACKET) {
				S.nextToken();
				if (S.token == RBRACKET && (mode & TYPE) != 0) {
					mode = TYPE;
					S.nextToken();
					return bracketsOpt(F.at(pos1).TypeArray(t,0));
				}
				if((mode & 1) != 0) {
					mode = EXPR;
					t = F.at(pos1).Indexed(t, term());
				}
				accept(RBRACKET);
            
			} else if (S.token == DOT) {
				S.nextToken();
				t = argumentsOpt(F.at(pos).Select(t, name()));
			} else {
				break;
			}
		}
		while ((S.token == PLUSPLUS || S.token == SUBSUB) && (mode & EXPR) != 0) {
			mode = EXPR;
			t = F.at(S.pos).Operation(S.token == PLUSPLUS ? Tree.POSTINC : Tree.POSTDEC, List.make(t));
			S.nextToken();
		}
		return t;
	}

    /**
	* BasicType = INT | FLOAT | DOUBLE | VECTOR | MATRIX | STRING
	*/
	Tree basicType() {
		Tree t = F.at(S.pos).TypeIdent(typetag(S.token));
		S.nextToken();
		return t;
	}
	
	/**
	* ArgumentsOpt = [ Arguments ]
	*/
	Tree argumentsOpt(Tree t) {
		if ((mode & EXPR) != 0 && S.token == LPAREN) {
			mode = EXPR;
			return arguments(t);
		} else {
			return t;
		}
	}

	/**
	* Arguments = [ "(" ] [Expression { COMMA Expression }] [ ")" ]
	*/
	private boolean withQuote = false;
	private int nbparens = 0;
//	if (nbparens > 0) syntaxError(S.pos, "invalid.parens.count", ""+nbparens);

	public List arguments() {
		int pos = S.pos;
		ListBuffer args = new ListBuffer();

		boolean comma = false;
		Tree tree;
		
loop:		
		while (true) {
			switch (S.token) {
				case SEMI: 
				case RPAREN: 
					break loop;

				case QUOTE: 
					if (withQuote) break loop;
					tree = expression();
					args.append(tree);
					if (tree instanceof Erroneous) break loop;
					comma = false;
					break;
					
				case LPAREN:
					S.nextToken();
					if (S.token == RPAREN) S.nextToken();
					else {
						tree = expression();
						args.append(tree);
						if (tree instanceof Erroneous) break loop;
						if (S.token == RPAREN) S.nextToken();
					}
					comma = false;
					break;					

				case COMMA:
					if (moreStatement) break loop;
					S.nextToken();
					comma = true;
					break;
					
				default:
					if (!comma && S.token != LTLT && optag(S.token) != -1) break loop;
					tree = expression();
					args.append(tree);
					if (tree instanceof Erroneous) break loop;
					comma = false;
					break;					
			}
		}

		return args.toList();
	}

	Tree arguments(Tree t) {
		int pos = S.pos;
		List args = arguments();
		return F.at(pos).Apply(t, args);
	}

	/**
	* BracketsOpt = {"[" "]"}
	*/
	protected Tree bracketsOpt(Tree t) {
		if (S.token == LBRACKET) {
			if (t instanceof TypeIdent && ((TypeIdent)t).typetag == Type.MATRIX) {
				S.nextToken();
				int pos = S.pos;
				Tree rows = expression();
				if (!(rows instanceof Literal) || ((Literal)rows).typetag != Type.INT) {
					syntaxError(pos, "expected", "integer");
					return errorTree;
				}
				accept(RBRACKET);
				accept(LBRACKET);
				pos = S.pos;
				Tree cols = expression();
				if (!(cols instanceof Literal) || ((Literal)cols).typetag != Type.INT) {
					syntaxError(pos, "expected", "integer");
					return errorTree;
				}
				accept(RBRACKET);
				return F.at(pos).TypeMatrix(((Integer)((Literal)rows).value).intValue(),((Integer)((Literal)cols).value).intValue());
			}
			while (S.token == LBRACKET) {
				int pos = S.pos;
				S.nextToken();
				Tree cols = S.token == RBRACKET ? null : expression();
				if (cols != null && (!(cols instanceof Literal) || ((Literal)cols).typetag != Type.INT)) {
					syntaxError(pos, "expected", "integer");
					return errorTree;
				}
				accept(RBRACKET);
				t = F.at(pos).TypeArray(t, cols != null ? ((Integer)((Literal)cols).value).intValue() : 0);
			}
		}
		return t;
	}

	/**
	*  BracketsSuffixType =
	*/
	Tree bracketsSuffix(Tree t) {
		if ((mode & TYPE) != 0) {
			mode = TYPE;
		} else {
			syntaxError(S.pos, "dot.class.expected");
		}
		return t;
	}
	
	/**
	* ArrayCreatorRest = "[" ( "]" BracketsOpt ArrayInitializer
	*                         | Expression "]" {"[" Expression "]"} BracketsOpt )
	*/
	Tree arrayCreatorRest(int newpos, Tree elemtype) {
		int pos = S.pos;
		accept(LBRACKET);
		if (S.token == RBRACKET) {
			S.nextToken();
			elemtype = bracketsOpt(elemtype);
			if (S.token == LBRACE) {
				return arrayInitializer(elemtype);
			} else {
				syntaxError(S.pos, "expected", S.token2string(LBRACE));
				return errorTree;
			}
		} else {
			ListBuffer dims = new ListBuffer();
			dims.append(expression());
			accept(RBRACKET);
			while (S.token == LBRACKET) {
				int pos1 = S.pos;
				S.nextToken();
				dims.append(expression());
				accept(RBRACKET);
			}
			return F.at(newpos).NewArray(elemtype, dims.toList(), null);
		}
	}

	/**
	* ArrayInitializer = "{" [VariableInitializer {"," VariableInitializer} [","]] "}"
	*/
	Tree arrayInitializer(Tree t) {
		int pos = S.pos;
		accept(LBRACE);
		ListBuffer elems = new ListBuffer();
		if (S.token != RBRACE) {
			elems.append(variableInitializer());
			while (S.token == COMMA) {
				S.nextToken();
				if (S.token == RBRACE)
					break;
				elems.append(variableInitializer());
			}
		}
		accept(RBRACE);
		return F.at(pos).NewArray(t, Tree.emptyList, elems.toList());
	}

	/**
	* MatrixInitializer = "<<" [VariableInitializer { [ "," |";" ] VariableInitializer} ] ">>"
	*/
	private static boolean matrixParsing = false;
	Tree matrixInitializer() {
		matrixParsing = true;
		int pos = S.pos;
		int rows = 1, cols = 1, firstcols = 0;
		boolean error = false, ismatrix = false;
		accept(LTLT);
		ListBuffer elems = new ListBuffer();
		ListBuffer l2 = new ListBuffer();
		if (S.token != GTGT) {
			elems.append(variableInitializer());
			while (S.token == COMMA || S.token == SEMI) {
				if (S.token == COMMA) 
					cols++;
				else {
					l2.append(elems.toList());
					elems = new ListBuffer();
					rows++;
					if (firstcols != 0 && firstcols != cols)
						syntaxError(S.pos, "bad.count", ""+firstcols);
					if (firstcols == 0) firstcols = cols;
					cols = 1;
					if (!ismatrix) ismatrix = true;
				}
				S.nextToken();
				if (S.token == GTGT) 
					break;
				elems.append(variableInitializer());
				pos = S.pos;
			}
		}
		accept(GTGT);
		if (firstcols != 0 && firstcols != cols)
			syntaxError(pos, "bad.count", ""+firstcols);
		if (ismatrix) l2.append(elems.toList());
		matrixParsing = false;
		if (!ismatrix)
			if (elems.length() <= 3) return F.at(pos).NewVector(elems.toList());
			else return F.at(pos).NewMatrix(rows, cols, elems.toList());
		return F.at(pos).NewMatrix(rows, cols, l2.toList());
	}

	/**
	* VariableInitializer = ArrayInitializer | Expression | MatrixInitializer
	*/
	Tree variableInitializer() {
		return expression();
	}
	
	/**
	* ParExpression = "(" Expression ")"
	*/
	Tree parExpression() {
		accept(LPAREN);
		Tree t = expression();
		accept(RPAREN);
		return t;
	}
	
	/**
	* Block = "{" BlockStatements "}"
	*/
	protected Block block(int flags) {
		int pos = S.pos;
		accept(LBRACE);
		List stats = blockStatements();
		Block t = F.at(pos).Block(flags, stats);
		while (S.token == CASE || S.token == DEFAULT) {
			syntaxError("orphaned", S.token2string(S.token));
			blockStatements();
		}
		t.endpos = S.pos;
		accept(RBRACE);
		return t;
	}
	
	protected Block block() {
		return block(0);
	}

	/**
	* BlockStatements = { BlockStatement }
	*  BlockStatement  = LocalVariableDeclarationStatement
	*                  | ClassOrInterfaceDeclaration
	*                  | [Ident ":"] Statement
	*  LocalVariableDeclarationStatement = [FINAL] Type VariableDeclarators ";"
	*/
	List blockStatements() {
        ListBuffer stats = new ListBuffer();
        while (true) {
			if (blockStatement(stats)) return stats.toList();
		}
	}

	public boolean blockStatement(ListBuffer stats) {
		int pos = S.pos;
		switch (S.token) {
			case RBRACE:
			case CASE:
			case DEFAULT:
			case EOF:
				return true;
				
			case LBRACE:
			case IF:
			case FOR:
			case WHILE:
			case DO:
			case SWITCH:
			case RETURN:
			case BREAK:
			case CONTINUE:
			case SEMI:
			case ELSE:
			case CATCH:
			case NAME:
			case ALIAS:
				stats.append(statement());
				break;
			
			default:
				Name name = S.name;
				boolean isMethod = methods.get(S.name.toString()) != null;
				Tree t = term(EXPR | TYPE);
				if ((lastmode & TYPE) != 0 && S.token == VARIABLE) {
					stats.appendList(variableDeclarators(0, t));
					if (S.token == RPAREN) S.nextToken();
					accept(SEMI);
				} else {
					if (isMethod) 
						stats.append( F.at(pos).Exec(checkExprStat(t)) );
					else
						stats.append( F.at(pos).Command(checkExprStat(t), true) );
					accept(SEMI);
				}				
		}
		return false;
	}

	/**
	* Statement =
	*       Block
	*     | IF ParExpression Statement [ELSE Statement]
	*     | FOR "(" ForInitOpt ";" [Expression] ";" ForUpdateOpt ")" Statement
	*     | WHILE ParExpression Statement
	*     | DO Statement WHILE ParExpression ";"
	*     | TRY Block ( Catches | [Catches] FinallyPart )
	*     | SWITCH ParExpression "{" SwitchBlockStatementGroups "}"
	*     | SYNCHRONIZED ParExpression Block
	*     | RETURN [Expression] ";"
	*     | THROW Expression ";"
	*     | BREAK [Ident]
	*     | CONTINUE [Ident]
	*     | ";"
	*     | ExpressionStatement
	*     | Ident ":" Statement
	*/
	public Tree statement() {
		if (S.token == RPAREN) S.nextToken();
		
		int pos = S.pos;
		switch (S.token) {
			case LBRACE:
				return block();
			
			case IF:
			{
				S.nextToken();
				Tree cond = parExpression();
				Tree thenpart = statement();
				Tree elsepart = null;
				if (S.token == ELSE) {
					S.nextToken();
					elsepart = statement();
				}
				return F.at(pos).Conditional(Tree.CONDSTAT, cond, thenpart, elsepart);
			}

			case FOR:
			{
				S.nextToken();
				accept(LPAREN);
				List inits = S.token == SEMI ? Tree.emptyList : forInit();
				Tree coll = (Tree)inits.head;
				if (coll == null || coll.tag != Tree.COLLECTION) {
					accept(SEMI);
					Tree cond = S.token == SEMI ? null : expression();
					accept(SEMI);
					List steps = S.token == RPAREN ? Tree.emptyList : forUpdate();
					accept(RPAREN);
					Tree body = statement();
					return F.at(pos).ForLoop(inits, cond, steps, body);
				} else {
					accept(RPAREN);
					Tree body = statement();
					return F.at(pos).ForIn(coll, body);
				}
			}

			case WHILE:
			{
				S.nextToken();
				Tree cond = parExpression();
				Tree body = statement();
				return F.at(pos).WhileLoop(cond, body);
			}

			case DO:
			{
				S.nextToken();
				Tree body = statement();
				accept(WHILE);
				Tree cond = parExpression();
				accept(SEMI);
				return F.at(pos).DoLoop(body, cond);
			}

			case SWITCH:
			{
				S.nextToken();
				Tree selector = parExpression();
				accept(LBRACE);
				List cases = switchBlockStatementGroups();
				accept(RBRACE);
				return F.at(pos).Switch(selector, cases);
			}

			case RETURN:
			{
				S.nextToken();
				Tree result = S.token == SEMI ? null : expression();
				accept(SEMI);
				return F.at(pos).Return(result);
			}
			
			case BREAK:
			{
				S.nextToken();
				Name label = S.token == IDENTIFIER ? ident() : null;
				accept(SEMI);
				return F.at(pos).Break(label);
			}
			
			case CONTINUE:
			{
				S.nextToken();
				Name label = S.token == IDENTIFIER ? ident() : null;
				accept(SEMI);
				return F.at(pos).Continue(label);
			}

			case SEMI:
				S.nextToken();
				return F.at(pos).Block(0, Tree.emptyList);
			
			case ELSE:
				return syntaxError("else.without.if");
			
			case NAME:
			{
				boolean isMethod = methods.get(S.name.toString()) != null;
				withQuote = false;	
				Tree cmd = command();
				Tree stat;
				if (isMethod) 
					stat = F.at(pos).Exec(checkExprStat(cmd));
				else
					stat = F.at(pos).Command(checkExprStat(cmd), true);
				if (S.token == RPAREN) S.nextToken();
				accept(SEMI);
				return stat;
			}
			
			case VARIABLE:
			{
				Tree stat = F.at(pos).Exec(checkExprStat(expression()));
				accept(SEMI);
				return stat;
			}
			
			case ALIAS:
			{
				Tree stat = F.at(pos).Exec(alias());
				accept(SEMI);
				return stat;
			}
						
			default:
				Name name = S.name;
				Tree expr = expression();
				if (S.token == COLON && expr.tag == Tree.IDENT) {
					S.nextToken();
					Tree stat = statement();
					return F.at(pos).Labelled(name, stat);
				} else {
					boolean isMethod = methods.get(S.name.toString()) != null;
					Tree stat;
					if (isMethod) 
						stat = F.at(pos).Exec(checkExprStat(expr));
					else
						stat = F.at(pos).Command(checkExprStat(expr), true);
					accept(SEMI);
					return stat;
				}				
		}
	}
	
	/**
	* Alias	= ALIAS name [ "{" BlockStatements "}" | Command ]
	*/
	Alias alias() {
		int pos = S.pos;
		accept(ALIAS);
		Name name = name();
		Tree cmd = null;
		ListBuffer args = new ListBuffer();
		if (S.token == LBRACE) {
			Block block = block();
			args.appendList(block.stats);
		} else {
			cmd = command();
		}
System.out.println("alias: " + name + " " + cmd);		
		return F.at(pos).Alias(name, cmd, args.toList());
	}
	
	/**
	* CatchClause	= CATCH "(" FormalParameter ")" 
	*/
	Catch catchClause() {
		int pos = S.pos;
		accept(CATCH);
		accept(LPAREN);
		Tree expr = expression();
		accept(RPAREN);
		return F.at(pos).Catch(null, expr);
	}
	
	/**
	* SwitchBlockStatementGroups = { SwitchBlockStatementGroup }
	*  SwitchBlockStatementGroup = SwitchLabel BlockStatements
	*  SwitchLabel = CASE ConstantExpression ":" | DEFAULT ":"
	*/
	List switchBlockStatementGroups() {
		ListBuffer cases = new ListBuffer();
		while (true) {
			int pos = S.pos;
			switch (S.token) {
				case CASE:
				{
					S.nextToken();
					Tree pat = expression();
					accept(COLON);
					List stats = blockStatements();
					cases.append(F.at(pos).Case(pat, stats));
					break;
				}
				
				case DEFAULT:
				{
					S.nextToken();
					accept(COLON);
					List stats = blockStatements();
					cases.append(F.at(pos).Case(null, stats));
					break;
				}
				
				case RBRACE:
				case EOF:
					return cases.toList();
				
				default:
					S.nextToken();
					syntaxError(pos, "case.default.or.right-brace.expected");				
			}
		}
	}

	/**
	* MoreStatementExpressions = { COMMA StatementExpression }
	*/
	private boolean moreStatement = false;
	List moreStatementExpressions(int pos, Tree first) {
		ListBuffer stats = new ListBuffer();
		stats.append(F.at(pos).Exec(checkExprStat(first)));
		moreStatement = true;
		while (S.token == COMMA) {
			S.nextToken();
			pos = S.pos;
			Tree t = expression();
			stats.append(F.at(pos).Exec(checkExprStat(t)));
		}
		moreStatement = false;
		return stats.toList();
	}
	
	/**
	* ForInit = StatementExpression MoreStatementExpressions
	*           |  [FINAL] Type VariableDeclarators
	*/
	List forInit() {
		int pos = S.pos;
		Tree t = term(EXPR | TYPE);
		if (t.tag == Tree.COLLECTION) return Tree.emptyList.prepend(t);
		if ((lastmode & TYPE) != 0 && S.token == IDENTIFIER)
			return variableDeclarators(0, t);
		else
			return moreStatementExpressions(pos, t);
	}

	/**
	* ForUpdate = StatementExpression MoreStatementExpressions
	*/
	List forUpdate() {
		return moreStatementExpressions(S.pos, expression());
	}

	/**
	* ModifiersOpt = { Modifier }
	*  Modifier = GLOBAL
	*/
	public int modifiersOpt() {
		int flag = 0;
		if (S.token == GLOBAL) {
			flag = Flags.PUBLIC;
			S.nextToken();
		}
		return flag;
	}

	/**
	* VariableDeclarators = VariableDeclarator { "," VariableDeclarator }
	*/
	protected List variableDeclarators(int flags, Tree type) {
		return variableDeclaratorsRest(S.pos, flags, type, var(), false);
	}

	/**
	* VariableDeclaratorsRest = VariableDeclaratorRest { "," VariableDeclarator }
	*  ConstantDeclaratorsRest = ConstantDeclaratorRest { "," ConstantDeclarator }
	*
	*  @param reqInit  Is an initializer always required?
	*/
	protected List variableDeclaratorsRest(int pos, int flags, Tree type, Name name, boolean reqInit) {
		ListBuffer vdefs = new ListBuffer();
		vdefs.append(variableDeclaratorRest(pos, flags, type, name, reqInit));
		while (S.token == COMMA) {
			S.nextToken();
			vdefs.append(variableDeclarator(flags, type, reqInit));
		}
		return vdefs.toList();
	}

	/**
	* VariableDeclarator = Ident VariableDeclaratorRest
	*  ConstantDeclarator = Ident ConstantDeclaratorRest
	*/
	VarDef variableDeclarator(int flags, Tree type, boolean reqInit) {
		return variableDeclaratorRest(S.pos, flags, type, var(), reqInit);
	}
	
	/**
	* VariableDeclaratorRest = BracketsOpt ["=" VariableInitializer]
	*  ConstantDeclaratorRest = BracketsOpt "=" VariableInitializer
	*
	*  @param reqInit  Is an initializer always required?
	*/
	protected VarDef variableDeclaratorRest(int pos, int flags, Tree type, Name name,	boolean reqInit) {
		type = bracketsOpt(type);
		Tree init = null;
		if (S.token == EQ) {
			S.nextToken();
			init = variableInitializer();
		} else if (reqInit)
			syntaxError(S.pos, "expected", S.token2string(EQ));
		VarDef result = F.at(pos).VarDef(flags, name, type, init);
		return result;
	}
	
	/**
	* VariableDeclaratorId = Ident BracketsOpt
	*/
	VarDef variableDeclaratorId(int flags, Tree type) {
		int pos = S.pos;
		Name name = var();
		type = bracketsOpt(type);
		return F.at(pos).VarDef(flags, name, type, null);
	}

	/**
	* CompilationUnit = {TypeDeclaration}
	*/
	public Tree.TopLevel compilationUnit(boolean expression) {
		int pos = S.pos;
		ListBuffer defs = new ListBuffer();
		while (S.token != EOF)
			defs.appendList(typeDeclaration(expression));
		Tree.TopLevel toplevel = F.at(pos).TopLevel(null, defs.toList());
		return toplevel;
	}

	/**
	* TypeDeclaration
	*/
	List typeDeclaration(boolean expression) {
		int pos = S.pos;
		ListBuffer defs = new ListBuffer();
		while (S.token != EOF) {
			defs.appendList(body(expression));
		}
		return defs.toList();
	}

	/**
	* body =
	*      ";"
	*    | ModifiersOpt
	*      ( Type Ident ( VariableDeclaratorsRest ";" | MethodDeclaratorRest )
	*      | VOID Ident MethodDeclaratorRest
	*      | TypeParameters (Type | VOID) Ident MethodDeclaratorRest
	*      | Ident ConstructorDeclaratorRest
	*      | TypeParameters Ident ConstructorDeclaratorRest
	*      | ClassOrInterfaceDeclaration
	*      )
	*/
	protected List body(boolean expression) {
		
		int pos = S.pos;
		switch (S.token) {
			case SEMI:
			case IF:
			case DO:
			case WHILE:
			case FOR:
			case SWITCH:
			case CATCH:
			case ALIAS:
				return blockStatements();

			case LBRACE:
				return Tree.emptyList.prepend(block());
				
			default:
				int flags = modifiersOpt();
				if (S.token == PROC) {
					if (expression) return Tree.emptyList.prepend(syntaxError(S.pos, "proc.not.allowed.here"));
					pos = S.pos;
					S.nextToken();
					Name name;		
					Tree type;
					boolean isVoid = S.token == NAME;
					if (isVoid) {
						type = F.at(pos).TypeIdent(Type.VOID);
						name = name();		
					} else {
						type = type();
						name = name();		
					}
					return Tree.emptyList.prepend(methodDeclaratorRest(pos, flags, type, name, TypeParameter.emptyList, false, isVoid));
				} else if (S.token == NAME || S.token == VARIABLE) {
					List defs = Tree.emptyList.prepend(statement());
					return defs;
				} else {
					Tree type = type();
					Name name = null;		
					switch (type.tag) {
						case Tree.TYPEIDENT:
							name = var();
							break;
							
						case IDENTIFIER:
							name = ((Ident)type).name;
							type = null;
							break;
							
						case NAME:
							statement();
							break;
							
						default:
							break;
					}
					List defs = variableDeclaratorsRest(pos, flags, type, name, false);
					if (S.token == RPAREN) S.nextToken();
					accept(SEMI);
					return defs;
				}	
		}
	}

	/**
	* MethodDeclaratorRest =
	*      FormalParameters BracketsOpt [Throws TypeList] ( MethodBody | ";")
	*  VoidMethodDeclaratorRest =
	*      FormalParameters [Throws TypeList] ( MethodBody | ";")
	*  InterfaceMethodDeclaratorRest =
	*      FormalParameters BracketsOpt [THROWS TypeList] ";"
	*  VoidInterfaceMethodDeclaratorRest =
	*      FormalParameters [THROWS TypeList] ";"
	*  ConstructorDeclaratorRest =
	*      "(" FormalParameterListOpt ")" [THROWS TypeList] MethodBody
	*/
	protected Tree methodDeclaratorRest(int pos, int flags, Tree type, Name name, List typarams, boolean isInterface, boolean isVoid) {
		List params = formalParameters();
		if (!isVoid) type = bracketsOpt(type);
		List thrown = Tree.emptyList;
		if (S.token == THROWS) {
			S.nextToken();
			thrown = qualidentList();
		}
		Block body;
		if (S.token == LBRACE) {
			if (isInterface) log.error(S.pos, "intf.meth.cant.have.body");
			body = block();
		} else {
			accept(SEMI);
			body = null;
		}
		Tree result = F.at(pos).MethodDef(flags, name, type, typarams, params, thrown, body);
		methods.put(name.toString(), Boolean.TYPE);
		return result;
	}

	/**
	* QualidentList = Qualident {"," Qualident}
	*/
	List qualidentList() {
		ListBuffer ts = new ListBuffer();
		ts.append(qualident());
		while (S.token == COMMA) {
			S.nextToken();
			ts.append(qualident());
		}
		return ts.toList();
	}

	/**
	* FormalParameters = "(" [FormalParameter {"," FormalParameter}] ")"
	*/
	protected List formalParameters() {
		ListBuffer params = new ListBuffer();
		accept(LPAREN);
		if (S.token != RPAREN) {
			params.append(formalParameter());
			while (S.token == COMMA) {
				S.nextToken();
				params.append(formalParameter());
			}
		}
		accept(RPAREN);
		return params.toList();
	}

	int optFinal() {
		if (S.token == FINAL) {
			S.nextToken();
			return Flags.FINAL;
		} else {
			return 0;
		}
	}

	/**
	* FormalParameter = [FINAL] Type VariableDeclaratorId
	*/
	VarDef formalParameter() {
		return variableDeclaratorId(optFinal(), type());
	}
	
	/**
	* Check that given tree is a legal expression statement.
	*/
	public Tree checkExprStat(Tree t) {
		switch (t.tag) {
			case Tree.PREINC:
			case Tree.PREDEC:
			case Tree.POSTINC:
			case Tree.POSTDEC:
			case Tree.ASSIGN:
			case Tree.BITOR_ASG:
			case Tree.BITXOR_ASG:
			case Tree.BITAND_ASG:
			case Tree.SL_ASG:
			case Tree.SR_ASG:
			case Tree.USR_ASG:
			case Tree.PLUS_ASG:
			case Tree.MINUS_ASG:
			case Tree.MUL_ASG:
			case Tree.DIV_ASG:
			case Tree.MOD_ASG:
			case Tree.APPLY:
			case Tree.NEWCLASS:
			case Tree.ERRONEOUS:
			case Tree.GLOBALREF:
			case Tree.CATCH:
				return t;
			
			default:
				log.error(t.pos, "not.stmt");
				return errorTree;		
		}
	}

	/**
	* Return precedence of operator represented by token,
	*  -1 if token is not a binary operator. @see TreeInfo.opPrec
	*/
	static int prec(int token) {
		int oc = optag(token);
		return oc >= 0 ? TreeInfo.opPrec(oc) : -1;
	}

	/**
	* Return operation tag of binary operator represented by token,
	*  -1 if token is not a binary operator.
	*/
	static int optag(int token) {
		switch (token) {
			case BARBAR:
				return Tree.OR;
			
			case AMPAMP:
				return Tree.AND;
			
			case BAR:
				return Tree.BITOR;
			
			case BAREQ:
				return Tree.BITOR_ASG;
			
			case CARET:
				return Tree.BITXOR;
			
			case CARETEQ:
				return Tree.BITXOR_ASG;
			
			case AMP:
				return Tree.BITAND;
			
			case AMPEQ:
				return Tree.BITAND_ASG;
			
			case EQEQ:
				return Tree.EQ;
			
			case BANGEQ:
				return Tree.NE;
			
			case LT:
				return Tree.LT;
			
			case GT:
				return Tree.GT;
			
			case LTEQ:
				return Tree.LE;
			
			case GTEQ:
				return Tree.GE;
			
			case LTLT:
				return Tree.SL;
			
			case LTLTEQ:
				return Tree.SL_ASG;
			
			case GTGT:
				return matrixParsing? -1 : Tree.SR;
			
			case GTGTEQ:
				return Tree.SR_ASG;
			
			case GTGTGT:
				return Tree.USR;
			
			case GTGTGTEQ:
				return Tree.USR_ASG;
			
			case PLUS:
				return Tree.PLUS;
			
			case PLUSEQ:
				return Tree.PLUS_ASG;
			
			case SUB:
				return Tree.MINUS;
			
			case SUBEQ:
				return Tree.MINUS_ASG;
			
			case STAR:
				return Tree.MUL;
			
			case STAREQ:
				return Tree.MUL_ASG;
			
			case SLASH:
				return Tree.DIV;
			
			case SLASHEQ:
				return Tree.DIV_ASG;
			
			case PERCENT:
				return Tree.MOD;
			
			case PERCENTEQ:
				return Tree.MOD_ASG;
			
			case INSTANCEOF:
				return Tree.TYPETEST;
			
			case IN:
				return Tree.COLLECTION;
			
			default:
				return -1;			
		}
	}

	/**
	* Return operation tag of unary operator represented by token,
	*  -1 if token is not a binary operator.
	*/
	static int unoptag(int token) {
		switch (token) {
			case PLUS:
				return Tree.POS;
			
			case SUB:
				return Tree.NEG;
			
			case BANG:
				return Tree.NOT;
			
			case TILDE:
				return Tree.COMPL;
			
			case PLUSPLUS:
				return Tree.PREINC;
			
			case SUBSUB:
				return Tree.PREDEC;
			
			default:
				return -1;		
		}
	}

	/**
	* Return type tag of basic type represented by token,
	*  -1 if token is not a basic type identifier.
	*/
	static int typetag(int token) {
		switch (token) {
			case INT:
				return Type.INT;
			
			case FLOAT:
				return Type.FLOAT;
			
			case DOUBLE:
				return Type.DOUBLE;
			
			case STRING:
				return Type.STRING;

			case VECTOR:
				return Type.VECTOR;

			case MATRIX:
				return Type.MATRIX;

			default:
				return -1;	
		}
    }
}
