package jmel.parser;

import java.io.*;
import jmel.util.*;

/**
 * The lexical analyzer maps an input stream consisting of
 *  ASCII characters and Unicode escapes into a token sequence.
 */
public class Scanner implements Tokens, LayoutCharacters {
	/**
	* The token, set by nextToken().
	*/
	public int token;
	
	/**
	* The token's position. pos = line << Position.LINESHIFT + col.
	*  Line and column numbers start at 1.
	*/
	public int pos;
	
	/**
	* The first character position after the previous token.
	*/
	protected int lastPos;
	
	/**
	* The position where a lexical error occurred;
	*/
	int errPos = Position.NOPOS;
	
	/**
	* The name of an identifier or token:
	*/
	public Name name;
	
	/**
	* The radix of a numeric literal token.
	*/
	int radix;
	
	/**
	* A character buffer for literals.
	*/
	protected char[] sbuf = new char[128];
	protected int sp;
	
	/**
	* The input buffer, index of next chacter to be read,
	*  index of one past last character in buffer.
	*/
	protected char[] buf;
	protected int bp;
	protected int buflen;
	
	/**
	* The current character.
	*/
	protected char ch;
	
	/**
	* The line number position of the current character.
	*/
	protected int line;
	
	/**
	* The column number position of the current character.
	*/
	protected int col;
	
	/**
	* The buffer index of the last converted unicode character
	*/
	private int unicodeConversionBp = 0;
	
	/**
	* Keyword array. Maps name indices to Tokens.
	*/
	protected static byte[] key;
	
	/**
	* The number of the last entered keyword.
	*/
	protected static int maxKey = 0;
	
	/**
	* The log to be used for error reporting.
	*/
	Log log;
	
	public int nparens = 0;
	
	/**
	* Construct a scanner from a file input stream, a log, and a
	*  character encoding.
	*/
	public Scanner(Log log) {
		this.log = log;
	}
	public Scanner(InputStream in, Log log, String encoding) {
		this(log);
		try {
			int bufsize = in.available() + 1;
			if (buf == null || buf.length < bufsize) buf = new char[bufsize];
			buflen = 0;
			InputStreamReader reader = encoding == null ? new InputStreamReader(in) : new InputStreamReader(in, encoding);
			if (jmel.script.Script.interactive.get()) {
				CommandLineReader clreader = new CommandLineReader(reader);
				while (true) {
					int nread = clreader.read(buf, buflen, buf.length - buflen);
					if (nread < 0) break;
					buflen = buflen + nread;
					char[] newbuf = new char[buflen * 2];
					System.arraycopy(buf, 0, newbuf, 0, buflen);
					buf = newbuf;
					if (buflen > 0) {
						if (buf[buflen-1] == '\n') {
							if (buflen > 1 && buf[buflen-2] != ';')
								buf[buflen-1] = ';';
							break;
						}
					} else 
						break;
				}
			} else {
				while (true) {
					int nread = reader.read(buf, buflen, buf.length - buflen);
					if (nread < 0) nread = 0;
					buflen = buflen + nread;
					if (buflen < buf.length) break;
					char[] newbuf = new char[buflen * 2];
					System.arraycopy(buf, 0, newbuf, 0, buflen);
					buf = newbuf;
				}
			}
		} catch (IOException e) {
			lexError(e.toString());
			buf = new char[1];
			buflen = 0;
		}
		buf[buflen] = EOB;
		line = 1;
		col = 0;
		bp = -1;
		scanChar();
		nextToken();
	}

	/**
	* Report an error at the given position using the provided argument.
	*/
	protected void lexError(int pos, String msg, String arg) {
		log.error(pos, msg, arg);
		token = ERROR;
		errPos = pos;
	}
	
	/**
	* Report an error at the given position.
	*/
	protected void lexError(int pos, String key) {
		lexError(pos, key, null);
	}
	
	/**
	* Report an error at the current token position.
	*/
	protected void lexError(String key) {
		lexError(pos, key, null);
	}
	
	/**
	* Report an error at the current token position using the provided
	*  argument.
	*/
	protected void lexError(String key, String arg) {
		lexError(pos, key, arg);
	}

	/**
	* Convert unicode escape; bp points to initial '\' character
	*  (Spec 3.3).
	*/
	protected void convertUnicode() {
		int startcol = col;
		if (ch == '\\') {
			bp++;
			ch = buf[bp];
			col++;
			if (ch == 'u') {
				do {
					bp++;
					ch = buf[bp];
					col++;
				} while (ch == 'u');
				int limit = bp + 3;
				if (limit < buflen) {
					int d = Character.digit(ch, 16);
					int code = d;
					while (bp < limit && d >= 0) {
						bp++;
						ch = buf[bp];
						col++;
						d = Character.digit(ch, 16);
						code = (code << 4) + d;
					}
					if (d >= 0) {
						ch = (char) code;
						unicodeConversionBp = bp;
						return;
					}
				}
				lexError(Position.make(line, startcol), "illegal.unicode.esc");
			} else {
				bp--;
				ch = '\\';
			}
		}
	}
	
	/**
	* Read next character.
	*/
	protected void scanChar() {
		bp++;
		ch = buf[bp];
		col++;
		if (ch == '\\') convertUnicode();
	}
	
	/**
	* Read next character in comment, skipping over double '\' characters.
	*/
	protected void scanCommentChar() {
		bp++;
		ch = buf[bp];
		col++;
		if (ch == '\\') {
			if (buf[bp + 1] == '\\') {
				bp++;
				col++;
			} else {
				convertUnicode();
			}
		}
	}

	/**
	* Append a character to sbuf.
	*/
	protected void putChar(char ch) {
		if (sp == sbuf.length) {
			char[] newsbuf = new char[sbuf.length * 2];
			System.arraycopy(sbuf, 0, newsbuf, 0, sbuf.length);
			sbuf = newsbuf;
		}
		sbuf[sp++] = ch;
	}
	
	/**
	* Read next character in character or string literal and copy into sbuf.
	*/
	protected void scanLitChar() {
		if (ch == '\\') {
			if (buf[bp + 1] == '\\' && unicodeConversionBp != bp) {
				bp++;
				col++;
				putChar('\\');
				scanChar();
			} else {
				scanChar();
				switch (ch) {
					case '0':					
					case '1':					
					case '2':					
					case '3':					
					case '4':					
					case '5':					
					case '6':					
					case '7':
						char leadch = ch;
						int oct = Character.digit(ch, 8);
						scanChar();
						if ('0' <= ch && ch <= '7') {
							oct = oct * 8 + Character.digit(ch, 8);
							scanChar();
							if (leadch <= '3' && '0' <= ch && ch <= '7') {
								oct = oct * 8 + Character.digit(ch, 8);
								scanChar();
							}
						}
						putChar((char) oct);
						break;
					
					case 'b':
						putChar('\b');
						scanChar();
						break;
					
					case 't':
						putChar('\t');
						scanChar();
						break;
					
					case 'n':
						putChar('\n');
						scanChar();
						break;
					
					case 'f':
						putChar('\f');
						scanChar();
						break;
					
					case 'r':
						putChar('\r');
						scanChar();
						break;
					
					case '\'':
						putChar('\'');
						scanChar();
						break;
					
					case '\"':
						putChar('\"');
						scanChar();
						break;
					
					case '\\':
						putChar('\\');
						scanChar();
						break;
					
					default:
						lexError(Position.make(line, col), "illegal.esc.char");			
				}
			}
		} else if (bp != buflen) {
			putChar(ch);
			scanChar();
		}
	}
	
	/**
	* Read fractional part of floating point number.
	*/
	protected void scanFraction() {
		while (Character.digit(ch, 10) >= 0) {
			putChar(ch);
			scanChar();
		}
		int sp1 = sp;
		if (ch == 'e' || ch == 'E') {
			putChar(ch);
			scanChar();
			if (ch == '+' || ch == '-') {
				putChar(ch);
				scanChar();
			}
			if ('0' <= ch && ch <= '9') {
				do {
					putChar(ch);
					scanChar();
				} while ('0' <= ch && ch <= '9');
				return;
			}
			lexError("malformed.fp.lit");
			sp = sp1;
		}
	}

	/**
	* Read fractional part and 'd' or 'f' suffix of floating point number.
	*/
	protected void scanFractionAndSuffix() {
		scanFraction();
		if (ch == 'd' || ch == 'D') {
			scanChar();
			token = DOUBLELITERAL;
		} else {
			if (ch == 'd' || ch == 'D') scanChar();
			token = FLOATLITERAL;
		}
	}

	/**
	* Read a number.
	*  @param radix  The radix of the number; one of 8, 10, 16.
	*/
	protected void scanNumber(int radix) {
		this.radix = radix;
		while (Character.digit(ch, radix == 8 ? 10 : radix) >= 0) {
			putChar(ch);
			scanChar();
		}
		if (radix <= 10 && ch == '.') {
			putChar(ch);
			scanChar();
			scanFractionAndSuffix();
		} else if (radix <= 10 && (ch == 'e' || ch == 'E' || ch == 'f' || ch == 'F' || ch == 'd' || ch == 'D')) {
			scanFractionAndSuffix();
		} else {
			if (ch == 'l' || ch == 'L') {
				scanChar();
				token = LONGLITERAL;
			} else {
				token = INTLITERAL;
			}
		}
	}
	
	/**
	* Read an identifier.
	*/
	protected void scanName() {
		int start = bp;
		do {
			if (sp == sbuf.length) 
				putChar(ch);
			else
			    sbuf[sp++] = ch;
			bp++;
			ch = buf[bp];
			col++;
			if (ch == '\\') convertUnicode();
			switch (ch) {
				case 'A':				
				case 'B':
				case 'C':
				case 'D':
				case 'E':
				case 'F':
				case 'G':
				case 'H':
				case 'I':
				case 'J':
				case 'K':
				case 'L':
				case 'M':
				case 'N':
				case 'O':
				case 'P':
				case 'Q':
				case 'R':
				case 'S':
				case 'T':
				case 'U':
				case 'V':
				case 'W':
				case 'X':
				case 'Y':
				case 'Z':
				case 'a':
				case 'b':
				case 'c':
				case 'd':
				case 'e':
				case 'f':
				case 'g':
				case 'h':
				case 'i':
				case 'j':
				case 'k':
				case 'l':
				case 'm':
				case 'n':
				case 'o':
				case 'p':
				case 'q':
				case 'r':
				case 's':
				case 't':
				case 'u':
				case 'v':
				case 'w':
				case 'x':
				case 'y':
				case 'z':
				case '_':
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
				case '|':
					break;
				
				default:
					if (ch <= 128 || !Character.isJavaIdentifierPart(ch)) {
						name = Name.fromChars(sbuf, 0, sp);
						if (name.index <= maxKey)
							token = key[name.index];
						else
							token = NAME;
						return;
					}			
			}
		} while (true);
	}
	
	/**
	* Read an string.
	*/
	protected void scanIdent() {
		int start = bp;
		do {
			if (sp == sbuf.length) 
				putChar(ch);
			else
			    sbuf[sp++] = ch;
			bp++;
			ch = buf[bp];
			col++;
			if (ch == '\\') convertUnicode();
			switch (ch) {
				case 'A':				
				case 'B':
				case 'C':
				case 'D':
				case 'E':
				case 'F':
				case 'G':
				case 'H':
				case 'I':
				case 'J':
				case 'K':
				case 'L':
				case 'M':
				case 'N':
				case 'O':
				case 'P':
				case 'Q':
				case 'R':
				case 'S':
				case 'T':
				case 'U':
				case 'V':
				case 'W':
				case 'X':
				case 'Y':
				case 'Z':
				case 'a':
				case 'b':
				case 'c':
				case 'd':
				case 'e':
				case 'f':
				case 'g':
				case 'h':
				case 'i':
				case 'j':
				case 'k':
				case 'l':
				case 'm':
				case 'n':
				case 'o':
				case 'p':
				case 'q':
				case 'r':
				case 's':
				case 't':
				case 'u':
				case 'v':
				case 'w':
				case 'x':
				case 'y':
				case 'z':
				case '_':
					break;
					
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					if (bp == start+1) lexError(pos+1, "illegal.identifier");
					break;
				
				default:
					if (ch <= 128 || !Character.isJavaIdentifierPart(ch)) {
						name = Name.fromChars(sbuf, 0, sp);
						token = VARIABLE;
						return;
					}			
			}
		} while (true);
	}
	
	/**
	* Return true if ch can be part of an operator.
	*/
	public boolean isSpecial(char ch) {
		switch (ch) {
			case '!':
			case '%':
			case '&':
			case '*':
			case '?':
			case '+':
			case '-':
			case ':':
			case '<':
			case '=':
			case '>':
			case '^':
			case '|':
			case '~':
			case '`':
				return true;
			
			default:
				return false;
		}
	}

	/**
	* Read longest possible sequence of special characters and convert
	*  to token.
	*/
	public void scanOperator() {
		if (ch == '`') {
			token = QUOTE;
			scanChar();
			return;
		}
		if (ch == '-' || ch == '+') {
			char oldch = ch;
			int oldsp = sp, oldbp = bp, oldcol = col, oldline = line;
			char[] oldsbuf = new char[sbuf.length];
			System.arraycopy(sbuf, 0, oldsbuf, 0, sbuf.length);
			Name oldname = name;
			boolean isNumber = true;
			do {
				if ((ch < '0' || ch > '9') && ch != '.' && ch != '+' && ch != '-') isNumber = false;
				putChar(ch);
				name = Name.fromChars(sbuf, 0, sp);
				scanChar();
			} while ((ch >= '0' && ch <= '9') || (ch == '.') || (ch >= 'A' && ch <= 'Z') || (ch >= 'a' &&  ch <= 'z'));
			if (!isNumber && name.length() >= 2 && ch != '[' && ch != '(' && ch != ')') return;

			sp = oldsp; bp = oldbp;
			col = oldcol; line = oldline;
			sbuf = oldsbuf;
			name = oldname;
			ch = oldch;
		}
		while (true) {
			putChar(ch);
			Name newname = Name.fromChars(sbuf, 0, sp);
			if (newname.index > maxKey || key[newname.index] == IDENTIFIER) {
				sp--;
				break;
			}
			name = newname;
			token = key[newname.index];
			scanChar();
			if (!isSpecial(ch)) break;
		}
	}

	/**
	* Skip a non-documentation comment. This method should be called once
	* the initial /, * and the next character have been read.
	*/
	protected void skipComment() {
		while (bp < buflen) {
			switch (ch) {
				case EOI:
					lexError(pos, "unclosed.comment");
					return;
				
				case '*':
					scanChar();
					if (ch == '/') return;
					break;
				
				case '\t':
					col = (col - 1) / TabInc * TabInc + TabInc;
					scanCommentChar();
					break;
				
				case FF:
					col = 0;
					scanCommentChar();
					break;
				
				case CR:
					line++;
					col = 0;
					scanCommentChar();
					if (ch == LF) {
						col = 0;
						scanCommentChar();
					}
					break;
				
				case LF:
					line++;
					col = 0;
					scanCommentChar();
					break;
				
				default:
					bp++;
					ch = buf[bp];
					col++;
					if (ch == '\\') {
						if (buf[bp + 1] == '\\') {
							bp++;
							col++;
						} else {
							convertUnicode();
						}
					}
					break;
				
			}
		}
	}

	/**
	* The value of a literal token, recorded as a string.
	*  For numbers, leading 0x and 'f', 'd', 'l' suffixes are suppressed.
	*/
	public String stringVal() {
		return new String(sbuf, 0, sp);
	}

	public int getPosition() { return bp; }
	public void setPosition(int bp) { this.bp = bp; }

	/**
	* Read token.
	*/
	public void nextToken() {
		lastPos = (line << Position.LINESHIFT) + col;
		sp = 0;
		while (true) {
			pos = (line << Position.LINESHIFT) + col;
			int start = bp;
			switch (ch) {
				case ' ':
					bp++;
					ch = buf[bp];
					col++;
					if (ch == '\\') convertUnicode();
					break;
				
				case '\t':
					col = (col - 1) / TabInc * TabInc + TabInc;
					scanChar();
					break;
				
				case FF:
					col = 0;
					scanChar();
					break;

				case CR:
					line++;
					col = 0;
					scanChar();
					if (ch == LF) {
						col = 0;
						scanChar();
					}
					break;
				
				case LF:
					line++;
					col = 0;
					scanChar();
					break;

				case 'A':
				case 'B':
				case 'C':
				case 'D':
				case 'E':
				case 'F':
				case 'G':
				case 'H':
				case 'I':
				case 'J':
				case 'K':
				case 'L':
				case 'M':
				case 'N':
				case 'O':
				case 'P':
				case 'Q':
				case 'R':
				case 'S':
				case 'T':
				case 'U':
				case 'V':
				case 'W':
				case 'X':
				case 'Y':
				case 'Z':
				case 'a':
				case 'b':
				case 'c':
				case 'd':
				case 'e':
				case 'f':
				case 'g':
				case 'h':
				case 'i':
				case 'j':
				case 'k':
				case 'l':
				case 'm':
				case 'n':
				case 'o':
				case 'p':
				case 'q':
				case 'r':
				case 's':
				case 't':
				case 'u':
				case 'v':
				case 'w':
				case 'x':
				case 'y':
				case 'z':
				case '_':
				case '|':
					scanName();
					return;

				case '$':
					scanIdent();
					return;

				case '0':
					scanChar();
					if (ch == 'x' || ch == 'X') {
						scanChar();
						if (Character.digit(ch, 16) < 0) {
							lexError("invalid.hex.number");
						}
						scanNumber(16);
					} else {
						putChar('0');
						scanNumber(8);
					}
					return;
				
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					scanNumber(10);
				return;

				case '.':
					scanChar();
					if ('0' <= ch && ch <= '9') {
						putChar('.');
						scanFractionAndSuffix();
					} else if (ch == '.') {
						putChar('.');
						putChar('.');
						scanChar();
						name = Name.fromChars(sbuf, 0, sp);
						token = NAME;
					} else {
						token = DOT;
					}
					return;
				
				case ',':
					scanChar();
					token = COMMA;
					return;

				case ';':
					scanChar();
					token = SEMI;
					return;
			
				case '(':
					nparens++;
					scanChar();
					token = LPAREN;
					return;
					
				case ')':
					nparens--;
					scanChar();
					token = RPAREN;
					return;

				case '[':
					scanChar();
					token = LBRACKET;
					return;
					
				case ']':
					scanChar();
					token = RBRACKET;
					return;
					
				case '{':
					scanChar();
					token = LBRACE;
					return;
					
				case '}':
					scanChar();
					token = RBRACE;
					return;

				case '/':
					scanChar();
					if (ch == '/') {
						do {
							bp++;
							ch = buf[bp];
							col++;
							if (ch == '\\') {
								if (buf[bp + 1] == '\\') {
									bp++;
									col++;
								} else {
									convertUnicode();
								}
							}
						} while (ch != CR && ch != LF && bp < buflen);
						break;
					} else if (ch == '*') {
						scanChar();
						skipComment();
						if (ch == '/') {
							scanChar();
							break;
						} else {
							lexError("unclosed.comment");
							return;
						}
					} else if (ch == '=') {
						name = Names.slashequals;
						token = SLASHEQ;
						scanChar();
					} else {
						name = Names.slash;
						token = SLASH;
					}
					return;

				case '\'':
					scanChar();
					if (ch == '\'') {
						lexError("empty.char.lit");
					} else {
						if (ch == CR || ch == LF) lexError(pos, "illegal.line.end.in.char.lit");
						scanLitChar();
						if (ch == '\'') {
							scanChar();
							token = CHARLITERAL;
						} else {
							lexError(pos, "unclosed.char.lit");
						}
					}
					return;

				case '\"':
					scanChar();
					while (ch != '\"' && ch != CR && ch != LF && bp < buflen) scanLitChar();
					if (ch == '\"') {
						token = STRINGLITERAL;
						scanChar();
					} else {
						lexError(pos, "unclosed.str.lit");
					}
					return;

				default:
					if (isSpecial(ch)) {
						scanOperator();
					} else if (Character.isJavaIdentifierStart(ch)) {
						scanIdent();
					} else if (bp == buflen || ch == EOI && bp + 1 == buflen) {
						token = EOF;
					} else {
						lexError("illegal.char", String.valueOf((int) ch));
						scanChar();
					}
					return;					
            }
        }
    }

	/**
	* The names of all tokens.
	*/
	private static Name[] tokenName = new Name[Tokens.TokenCount];

	public String token2string(int token) {
		switch (token) {
			case IDENTIFIER:
				return log.getLocalizedString("token.identifier");
			
			case VARIABLE:
				return log.getLocalizedString("token.variable");
				
			case NAME:
				return log.getLocalizedString("token.name");
			
			case CHARLITERAL:
				return log.getLocalizedString("token.character");
			
			case STRINGLITERAL:
				return log.getLocalizedString("token.string");
			
			case INTLITERAL:
				return log.getLocalizedString("token.integer");
			
			case LONGLITERAL:
				return log.getLocalizedString("token.long-integer");
			
			case FLOATLITERAL:
				return log.getLocalizedString("token.float");
			
			case DOUBLELITERAL:
				return log.getLocalizedString("token.double");
			
			case DOT:
				return "\'.\'";
			
			case COMMA:
				return "\',\'";
			
			case SEMI:
				return "\';\'";
			
			case LPAREN:
				return "\'(\'";
			
			case RPAREN:
				return "\')\'";
			
			case LBRACKET:
				return "\'[\'";
			
			case RBRACKET:
				return "\']\'";
			
			case LBRACE:
				return "\'{\'";
			
			case RBRACE:
				return "\'}\'";
			
			case ERROR:
				return log.getLocalizedString("token.bad-symbol");
			
			case EOF:
				return log.getLocalizedString("token.end-of-input");
			
			default:
				return tokenName[token].toJava();			
		}
	}
	
	private static void enterKeyword(String s, int token) {
		Name n = Name.fromString(s);
		tokenName[token] = n;
		if (n.index > maxKey) maxKey = n.index;
	}
	
	static {
		for (int j = 0; j < Tokens.TokenCount; j++) tokenName[j] = null;
		enterKeyword("`", QUOTE);
		enterKeyword("+", PLUS);
		enterKeyword("-", SUB);
		enterKeyword("!", BANG);
		enterKeyword("%", PERCENT);
		enterKeyword("^", CARET);
		enterKeyword("&", AMP);
		enterKeyword("*", STAR);
		enterKeyword("|", BAR);
		enterKeyword("~", TILDE);
		enterKeyword("/", SLASH);
		enterKeyword(">", GT);
		enterKeyword("<", LT);
		enterKeyword("?", QUES);
		enterKeyword(":", COLON);
		enterKeyword("=", EQ);
		enterKeyword("++", PLUSPLUS);
		enterKeyword("--", SUBSUB);
		enterKeyword("==", EQEQ);
		enterKeyword("<=", LTEQ);
		enterKeyword(">=", GTEQ);
		enterKeyword("!=", BANGEQ);
		enterKeyword("<<", LTLT);
		enterKeyword(">>", GTGT);
		enterKeyword(">>>", GTGTGT);
		enterKeyword("+=", PLUSEQ);
		enterKeyword("-=", SUBEQ);
		enterKeyword("*=", STAREQ);
		enterKeyword("/=", SLASHEQ);
		enterKeyword("&=", AMPEQ);
		enterKeyword("|=", BAREQ);
		enterKeyword("^=", CARETEQ);
		enterKeyword("%=", PERCENTEQ);
		enterKeyword("<<=", LTLTEQ);
		enterKeyword(">>=", GTGTEQ);
		enterKeyword(">>>=", GTGTGTEQ);
		enterKeyword("||", BARBAR);
		enterKeyword("&&", AMPAMP);
		enterKeyword("boolean", BOOLEAN);
		enterKeyword("break", BREAK);
		enterKeyword("case", CASE);
		enterKeyword("catch", CATCH);
		enterKeyword("continue", CONTINUE);
		enterKeyword("default", DEFAULT);
		enterKeyword("do", DO);
		enterKeyword("double", DOUBLE);
		enterKeyword("else", ELSE);
		enterKeyword("float", FLOAT);
		enterKeyword("for", FOR);
		enterKeyword("if", IF);
		enterKeyword("int", INT);
		enterKeyword("return", RETURN);
		enterKeyword("switch", SWITCH);
		enterKeyword("while", WHILE);
		enterKeyword("true", TRUE);
		enterKeyword("false", FALSE);

		enterKeyword("global", GLOBAL);
		enterKeyword("in", IN);
		enterKeyword("off", OFF);
		enterKeyword("on", ON);
		enterKeyword("matrix", MATRIX);
		enterKeyword("no", NO);
		enterKeyword("proc", PROC);
		enterKeyword("string", STRING);
		enterKeyword("vector", VECTOR);
		enterKeyword("yes", YES);
		enterKeyword("alias", ALIAS);
		
		key = new byte[maxKey + 1];
		for (int i = 0; i <= maxKey; i++) key[i] = IDENTIFIER;
		for (byte j = 0; j < Tokens.TokenCount; j++)
			if (tokenName[j] != null)
				key[tokenName[j].index] = j;
				
		// Dereference Java keywords
		Name oldname;
		oldname = Name.fromString("public");
		if (oldname.index <= maxKey) key[oldname.index] = NAME;
		oldname = Name.fromString("private");
		if (oldname.index <= maxKey) key[oldname.index] = NAME;
		oldname = Name.fromString("protected");
		if (oldname.index <= maxKey) key[oldname.index] = NAME;
		oldname = Name.fromString("length");
		if (oldname.index <= maxKey) key[oldname.index] = NAME;
 		oldname = Name.fromString("class");
		if (oldname.index <= maxKey) key[oldname.index] = NAME;
 		oldname = Name.fromString("interface");
		if (oldname.index <= maxKey) key[oldname.index] = NAME;
   }
}
