package jmel.tree;

import jmel.code.*;
import jmel.code.Symbol.*;
import jmel.util.*;

/**
 * Root class for abstract syntax tree nodes. It provides
 *  definitions for specific tree nodes as subclasses nested inside
 *  There are 40 such subclasses.
 *
 *  Each subclass is highly standardized.  It generally contains only tree
 *  fields for the syntactic subcomponents of the node.  Some classes that
 *  represent identifier uses or definitions also define a
 *  Symbol field that denotes the represented identifier.  Classes
 *  for non-local jumps also carry the jump target as a field.  The root
 *  class Tree itself defines fields for the tree's type and
 *  position.  No other fields are kept in a tree node; instead parameters
 *  are passed to methods accessing the node.
 *
 *  The only method defined in subclasses is `visit' which applies a
 *  given visitor to the tree. The actual tree processing is done by
 *  visitor classes in other packages. The abstract class
 *  Visitor, as well as an Factory interface for trees, are
 *  defined as inner classes in Tree.
 *  @see TreeMaker
 *  @see TreeInfo
 *  @see TreeTranslator
 *  @see Pretty
 */
public abstract class Tree implements Nodes {
	/**
	* An empty list of trees.
	*/
	public static final List emptyList = new List();

	public int pos;
	public Type type;
	public int tag;
	public boolean statement;
	
	/**
	* Initialize tree with given tag.
	*/
	public Tree(int tag) {
		this.tag = tag;
		if (Util.statistics) count++;
	}

	/**
	* Set position field and return this tree.
	*/
	public Tree setPos(int pos) {
		this.pos = pos;
		return this;
	}

	/**
	* Set type field and return this tree.
	*/
	public Tree setType(Type type) {
		this.type = type;
		return this;
	}

	/**
	* Visit this tree with a given visitor.
	*/
	public void visit(Visitor v) {
		v._case(this);
	}

	/**
	* Everything in one source file is kept in a TopLevel structure.
	* @param pid              The tree representing the package clause.
	* @param sourcefile       The source file name.
	* @param defs             All definitions in this file.
	* @param packge           The package it belongs to.
	* @param namedImportScope A scope for all named imports.
	* @param starImportScope  A scope for all import-on-demands.
	* @param docComments      A hashtable that stores all documentation comments
	*                         indexed by the tree nodes they refer to.
	*                         defined only if option -s is set.
	*/
	public static class TopLevel extends Tree {
		public Tree pid;
		public List defs;
		public Name sourcefile;
		public PackageSymbol packge;
		public Scope namedImportScope;
		public Scope starImportScope;
		
		public TopLevel(Tree pid, List defs, Name sourcefile, PackageSymbol packge, Scope namedImportScope, Scope starImportScope) {
			super(TOPLEVEL);
			this.pid = pid;
			this.defs = defs;
			this.sourcefile = sourcefile;
			this.packge = packge;
			this.namedImportScope = namedImportScope;
			this.starImportScope = starImportScope;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* An import clause.
	* @param qualid    The imported class(es).
	*/
	public static class Import extends Tree {
		public Tree qualid;
		
		public Import(Tree qualid) {
			super(IMPORT);
			this.qualid = qualid;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A class definition.
	* @param flags class flags
	* @param name the name of the class
	* @param typarams formal class parameters
	* @param extending the classes this class extends
	* @param implementing the interfaces implemented by this class
	* @param defs all variables and methods defined in this class
	* @param sym the symbol
	*/
	public static class ClassDef extends Tree {
		public int flags;
		public Name name;
		public List typarams;
		public Tree extending;
		public List implementing;
		public List defs;
		public ClassSymbol sym;
		
		public ClassDef(int flags, Name name, List typarams, Tree extending, List implementing, List defs, ClassSymbol sym) {
			super(CLASSDEF);
			this.flags = flags;
			this.name = name;
			this.typarams = typarams;
			this.extending = extending;
			this.implementing = implementing;
			this.defs = defs;
			this.sym = sym;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A method definition.
	* @param flags method flags
	* @param name method name
	* @param restype type of method return value
	* @param typarams type parameters
	* @param params value parameters
	* @param thrown exceptions thrown by this method
	* @param stats statements in the method
	* @param sym method symbol
	*/
	public static class MethodDef extends Tree {
		public int flags;
		public Name name;
		public Tree restype;
		public List typarams;
		public List params;
		public List thrown;
		public Block body;
		public MethodSymbol sym;
		public ListBuffer buffer = new ListBuffer();
		
		public MethodDef(int flags, Name name, Tree restype, List typarams, List params, List thrown, Block body, MethodSymbol sym) {
			super(METHODDEF);
			this.flags = flags;
			this.name = name;
			this.restype = restype;
			this.typarams = typarams;
			this.params = params;
			this.thrown = thrown;
			this.body = body;
			this.sym = sym;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A variable definition.
	* @param flags variable flags
	* @param name variable name
	* @param vartype type of the variable
	* @param init variables initial value
	* @param sym symbol
	*/
	public static class VarDef extends Tree {
		public int flags;
		public Name name;
		public Tree vartype;
		public Tree init;
		public VarSymbol sym;
		
		public VarDef(int flags, Name name, Tree vartype, Tree init, VarSymbol sym) {
			super(VARDEF);
			this.flags = flags;
			this.name = name;
			this.vartype = vartype;
			this.init = init;
			this.sym = sym;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
		public static final List emptyList = new List();
	}

	/**
	* A statement block.
	* @param stats statements
	* @param flags modifier
	*/
	public static class Block extends Tree {
		public int flags;
		public List stats;
		public int endpos = Position.NOPOS;
		
		public Block(int flags, List stats) {
			super(BLOCK);
			this.stats = stats;
			this.flags = flags;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	/**
	* A do loop
	*/
	public static class DoLoop extends Tree {
		public Tree body;
		public Tree cond;
		
		public DoLoop(Tree body, Tree cond) {
			super(DOLOOP);
			this.body = body;
			this.cond = cond;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	/**
	* A while loop
	*/
	public static class WhileLoop extends Tree {
		public Tree cond;
		public Tree body;
		
		public WhileLoop(Tree cond, Tree body) {
			super(WHILELOOP);
			this.cond = cond;
			this.body = body;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A for loop.
	*/
	public static class ForLoop extends Tree {
		public List init;
		public Tree cond;
		public List step;
		public Tree body;
		
		public ForLoop(List init, Tree cond, List step, Tree body) {
			super(FORLOOP);
			this.init = init;
			this.cond = cond;
			this.step = step;
			this.body = body;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A labelled expression or statement.
	*/
	public static class Labelled extends Tree {
		public Name label;
		public Tree body;
		
		public Labelled(Name label, Tree body) {
			super(LABELLED);
			this.label = label;
			this.body = body;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	/**
	* A "switch ( ) { }" construction.
	*/
	public static class Switch extends Tree {
		public Tree selector;
		public List cases;
		
		public Switch(Tree selector, List cases) {
			super(SWITCH);
			this.selector = selector;
			this.cases = cases;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A "case  :" of a switch.
	*/
	public static class Case extends Tree {
		public Tree pat;
		public List stats;
		
		public Case(Tree pat, List stats) {
			super(CASE);
			this.pat = pat;
			this.stats = stats;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A synchronized block.
	*/
	public static class Synchronized extends Tree {
		public Tree lock;
		public Tree body;
		
		public Synchronized(Tree lock, Tree body) {
			super(SYNCHRONIZED);
			this.lock = lock;
			this.body = body;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A "try { } catch ( ) { } finally { }" block.
	*/
	public static class Try extends Tree {
		public Tree body;
		public List catchers;
		public Tree finalizer;
		
		public Try(Tree body, List catchers, Tree finalizer) {
			super(TRY);
			this.body = body;
			this.catchers = catchers;
			this.finalizer = finalizer;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	/**
	* A catch block.
	*/
	public static class Catch extends Tree {
		public VarDef param;
		public Tree body;
		
		public Catch(VarDef param, Tree body) {
			super(CATCH);
			this.param = param;
			this.body = body;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}

		public static List emptyList = new List();		
	}

	/**
	* An "if ( ) { } else { }" block or a ( ) ? ( ) : ( )
	* conditional expression
	*/
	public static class Conditional extends Tree {
		public Tree cond;
		public Tree thenpart;
		public Tree elsepart;
		
		public Conditional(int tag, Tree cond, Tree thenpart, Tree elsepart) {
			super(tag);
			this.cond = cond;
			this.thenpart = thenpart;
			this.elsepart = elsepart;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* an expression statement
	* @param expr expression structure
	*/
	public static class Exec extends Tree {
		public Tree expr;
		
		public Exec(Tree expr) {
			super(EXEC);
			this.expr = expr;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* a command invocation
	* @param cmd expression structure
	* @param statement if returns a value
	*/
	public static class Command extends Tree {
		public Tree cmd;
		
		public Command(Tree cmd, boolean statement) {
			super(COMMAND);
			this.cmd = cmd;
			cmd.statement = statement;
			this.statement = statement;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* an alias definition
	*/
	public static class Alias extends Tree {
		public Name name;
		public Tree cmd;
		public List args;

		public Alias(Name name, Tree cmd, List args) {
			super(ALIAS);
			this.name = name;
			this.cmd = cmd;
			this.args = args;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A break from a loop or switch.
	*/
	public static class Break extends Tree {
		public Name label;
		public Tree target;
		
		public Break(Name label, Tree target) {
			super(BREAK);
			this.label = label;
			this.target = target;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	/**
	* A continue of a loop.
	*/
	public static class Continue extends Tree {
		public Name label;
		public Tree target;
		
		public Continue(Name label, Tree target) {
			super(CONTINUE);
			this.label = label;
			this.target = target;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A return statement.
	*/
	public static class Return extends Tree {
		public Tree expr;
		
		public Return(Tree expr) {
			super(RETURN);
			this.expr = expr;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	/**
	* A throw statement.
	*/
	public static class Throw extends Tree {
		public Tree expr;
		
		public Throw(Tree expr) {
			super(THROW);
			this.expr = expr;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A method invocation
	*/
	public static class Apply extends Tree {
		public Tree meth;
		public List args;
		
		public Apply(Tree meth, List args) {
			super(APPLY);
			this.meth = meth;
			this.args = args;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	/**
	* A new(...) operation.
	*/
	public static class NewClass extends Tree {
		public Tree encl;
		public Tree clazz;
		public List args;
		public ClassDef def;
		public Symbol constructor;
		
		public NewClass(Tree encl, Tree clazz, List args, ClassDef def, Symbol constructor) {
			super(NEWCLASS);
			this.encl = encl;
			this.clazz = clazz;
			this.args = args;
			this.def = def;
			this.constructor = constructor;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A new[...] operation.
	*/
	public static class NewArray extends Tree {
		public Tree elemtype;
		public List dims;
		public List elems;
		
		public NewArray(Tree elemtype, List dims, List elems) {
			super(NEWARRAY);
			this.elemtype = elemtype;
			this.dims = dims;
			this.elems = elems;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A assignment with "=".
	*/
	public static class Assign extends Tree {
		public Tree lhs;
		public Tree rhs;
		
		public Assign(Tree lhs, Tree rhs) {
			super(ASSIGN);
			this.lhs = lhs;
			this.rhs = rhs;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* An assignment with "+=", "|=" ...
	*/
	public static class Assignop extends Tree {
		public Tree lhs;
		public Tree rhs;
		public Symbol operator;
		
		public Assignop(int opcode, Tree lhs, Tree rhs, Symbol operator) {
			super(opcode);
			this.lhs = lhs;
			this.rhs = rhs;
			this.operator = operator;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A binary or unary opperation.
	*/
	public static class Operation extends Tree {
		public List args;
		public Symbol operator;
		
		public Operation(int opcode, List args, Symbol operator) {
			super(opcode);
			this.args = args;
			this.operator = operator;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A type cast.
	*/
	public static class TypeCast extends Tree {
		public Tree clazz;
		public Tree expr;
		
		public TypeCast(Tree clazz, Tree expr) {
			super(TYPECAST);
			this.clazz = clazz;
			this.expr = expr;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	/**
	* A type test.
	*/
	public static class TypeTest extends Tree {
		public Tree expr;
		public Tree clazz;
		
		public TypeTest(Tree expr, Tree clazz) {
			super(TYPETEST);
			this.expr = expr;
			this.clazz = clazz;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	/**
	* An array selection
	*/
	public static class Indexed extends Tree {
		public Tree indexed;
		public Tree index;
	
		public Indexed(Tree indexed, Tree index) {
			super(INDEXED);
			this.indexed = indexed;
			this.index = index;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* Selects through packages and classes
	* @param selected selected Tree hierarchie
	* @param selector name of field to select thru
	* @param sym symbol of the selected class
	*/
	public static class Select extends Tree {
		public Tree selected;
		public Name name;
		public Symbol sym;
		
		public Select(Tree selected, Name name, Symbol sym) {
			super(SELECT);
			this.selected = selected;
			this.name = name;
			this.sym = sym;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* An identifier
	* @param idname the name
	* @param sym the symbol
	*/
	public static class Ident extends Tree {
		public Name name;
		public Symbol sym;
		
		public Ident(Name name, Symbol sym) {
			super(IDENT);
			this.name = name;
			this.sym = sym;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A constant value given literally.
	* @param value value representation
	*/
	public static class Literal extends Tree {
		public int typetag;
		public Object value;
		
		public Literal(int typetag, Object value) {
			super(LITERAL);
			this.typetag = typetag;
			this.value = value;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	public static class NewVector extends Tree {
		public Tree x, y, z;
		public List elems;
		
		public NewVector(List elements) {
			super(VECTOR);
			for (List l = elements; l.nonEmpty(); l = l.tail) {
				if (x == null) x = (Tree) l.head;
				else if (y == null) y = (Tree) l.head;
				else if (z == null) z = (Tree) l.head;
			}
			this.elems = elements;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	public static class NewMatrix extends Tree {
		public int rows, cols;
		public Tree elements[][];
		
		public NewMatrix(int rows, int cols, List elements) {
			super(MATRIX);
			this.rows = rows;
			this.cols = cols;
			this.elements = new Tree[rows][cols];
			int i = 0, j = 0;
			for (List l1 = elements; l1.nonEmpty(); l1 = l1.tail) {
				if (l1.head instanceof List) {
					j = 0;
					for (List l2 = (List) l1.head; l2.nonEmpty(); l2 = l2.tail, j++) {
						if (j < this.elements[i].length) this.elements[i][j] = (Tree) l2.head;
					}
					i++;
				} else
					if (j < this.elements[i].length) this.elements[i][j++] = (Tree) l1.head;
			}
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	public static class TypeMatrix extends Tree {
		public int rows, cols;
		
		public TypeMatrix(int rows, int cols) {
			super(TYPEMATRIX);
			this.rows = rows;
			this.cols = cols;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	public static class GlobalRef extends Tree {
		public Tree type;
		public Name name;
		
		public GlobalRef(Name name, Tree type) {
			super(GLOBALREF);
			this.name = name;
			this.type = type;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	public static class Collection extends Tree {
		public Tree expr;
		public Tree coll;
		
		public Collection(Tree expr, Tree coll) {
			super(COLLECTION);
			this.expr = expr;
			this.coll = coll;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	public static class ForIn extends Tree {
		public Tree coll;
		public Tree body;
		
		public ForIn(Tree coll, Tree body) {
			super(FORIN);
			this.coll = coll;
			this.body = body;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}
	
	/**
	* Identifies a basic type.
	* @param tag the basic type id
	* @see SemanticConstants
	*/
	public static class TypeIdent extends Tree {
		public int typetag;
		
		public TypeIdent(int typetag) {
			super(TYPEIDENT);
			this.typetag = typetag;
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* An array type, A[]
	*/
	public static class TypeArray extends Tree {
		public Tree elemtype;
		public int size;
		
		public TypeArray(Tree elemtype, int size) {
			super(TYPEARRAY);
			this.elemtype = elemtype;
			this.size = size;		
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* A formal class parameter.
	* @param name name
	* @param extBound bound given in "extends ..." clause
	* @param implBound bound given in "implements ..." clause
	* only one of extBound, implBound can be non-null, and they might
	* both be null.
	*/
	public static class TypeParameter extends Tree {		
		TypeParameter() {
			super(0);
		}
		public static final List emptyList = new List();
	}

	public static class Erroneous extends Tree {	
		public Erroneous() {
			super(ERRONEOUS);
		}
		
		public void visit(Visitor v) {
			v._case(this);
		}
	}

	/**
	* An interface for tree factories
	*/
	public interface Factory {
		
		TopLevel TopLevel(Tree pid, List defs);
		
		Import Import(Tree qualid);
		
		ClassDef ClassDef(int flags, Name name, List typarams, Tree extending, List implementing, List defs);
		
		MethodDef MethodDef(int flags, Name name, Tree restype, List typarams, List params, List thrown, Block body);
		
		VarDef VarDef(int flags, Name name, Tree vartype, Tree init);
		
		Block Block(int flags, List stats);
		
		DoLoop DoLoop(Tree body, Tree cond);
		
		WhileLoop WhileLoop(Tree cond, Tree body);
		
		ForLoop ForLoop(List init, Tree cond, List step, Tree body);
		
		Labelled Labelled(Name label, Tree body);
		
		Switch Switch(Tree selector, List cases);
		
		Case Case(Tree pat, List stats);
		
		Synchronized Synchronized(Tree lock, Tree body);
		
		Try Try(Tree body, List catchers, Tree finalizer);
		
		Catch Catch(VarDef param, Tree body);
		
		Conditional Conditional(int tag, Tree cond, Tree thenpart, Tree elsepart);
		
		Exec Exec(Tree expr);
		
		Break Break(Name label);
		
		Continue Continue(Name label);
		
		Return Return(Tree expr);
		
		Throw Throw(Tree expr);
		
		Apply Apply(Tree fn, List args);
		
		Alias Alias(Name name, Tree cmd, List args);
		
		NewClass NewClass(Tree encl, Tree clazz, List args, ClassDef def);
		
		NewArray NewArray(Tree elemtype, List dims, List elems);
		
		Assign Assign(Tree lhs, Tree rhs);
		
		Assignop Assignop(int opcode, Tree lhs, Tree rhs);
		
		Operation Operation(int opcode, List args);
		
		TypeCast TypeCast(Tree expr, Tree type);
		
		TypeTest TypeTest(Tree expr, Tree clazz);
		
		Indexed Indexed(Tree indexed, Tree index);
		
		Select Select(Tree selected, Name selector);
		
		Ident Ident(Name idname);
		
		Literal Literal(int tag, Object value);
		
		TypeIdent TypeIdent(int typetag);
		
		TypeArray TypeArray(Tree elemtype, int size);
		
		Erroneous Erroneous();
		
		NewVector NewVector(List elems);
		
		NewMatrix NewMatrix(int rows, int cols, List elems);
		
		TypeMatrix TypeMatrix(int rows, int cols);
		
		Command Command(Tree cmd, boolean statement);
		
		GlobalRef GlobalRef(Name name, Tree type);
		
		Collection Collection(Tree expr, Tree coll);
		
		ForIn ForIn(Tree coll, Tree body);
	}

	/**
	  * A generic visitor class for trees.
	  */
	public static abstract class Visitor {
		public void _case(TopLevel that) {
			_case((Tree) that);
		}
		
		public void _case(Import that) {
			_case((Tree) that);
		}
		
		public void _case(ClassDef that) {
			_case((Tree) that);
		}
		
		public void _case(MethodDef that) {
			_case((Tree) that);
		}
		
		public void _case(VarDef that) {
			_case((Tree) that);
		}
		
		public void _case(Block that) {
			_case((Tree) that);
		}
		
		public void _case(DoLoop that) {
			_case((Tree) that);
		}

		public void _case(WhileLoop that) {
			_case((Tree) that);
		}
		
		public void _case(ForLoop that) {
			_case((Tree) that);
		}
		
		public void _case(Labelled that) {
			_case((Tree) that);
		}
		
		public void _case(Switch that) {
			_case((Tree) that);
		}
		
		public void _case(Case that) {
			_case((Tree) that);
		}
		
		public void _case(Synchronized that) {
			_case((Tree) that);
		}
		
		public void _case(Try that) {
			_case((Tree) that);
		}
		
		public void _case(Catch that) {
			_case((Tree) that);
		}
		
		public void _case(Conditional that) {
			_case((Tree) that);
		}
		
		public void _case(Exec that) {
			_case((Tree) that);
		}
		
		public void _case(Break that) {
			_case((Tree) that);
		}
		
		public void _case(Continue that) {
			_case((Tree) that);
		}

		public void _case(Return that) {
			_case((Tree) that);
		}
		
		public void _case(Throw that) {
			_case((Tree) that);
		}
		
		public void _case(Apply that) {
			_case((Tree) that);
		}
		
		public void _case(Alias that) {
			_case((Tree) that);
		}
		
		public void _case(NewClass that) {
			_case((Tree) that);
		}
		
		public void _case(NewArray that) {
			_case((Tree) that);
		}
		
		public void _case(Assign that) {
			_case((Tree) that);
		}
		
		public void _case(Assignop that) {
			_case((Tree) that);
		}
		
		public void _case(Operation that) {
			_case((Tree) that);
		}
		
		public void _case(TypeCast that) {
			_case((Tree) that);
		}
		
		public void _case(TypeTest that) {
			_case((Tree) that);
		}
		
		public void _case(Indexed that) {
			_case((Tree) that);
		}
		
		public void _case(Select that) {
			_case((Tree) that);
		}
		
		public void _case(Ident that) {
			_case((Tree) that);
		}
		
		public void _case(Literal that) {
			_case((Tree) that);
		}
		
		public void _case(TypeIdent that) {
			_case((Tree) that);
		}
		
		public void _case(TypeArray that) {
			_case((Tree) that);
		}
		
		public void _case(Erroneous that) {
			_case((Tree) that);
		}
		
		public void _case(NewVector that) {
			_case((Tree) that);
		}
		
		public void _case(NewMatrix that) {
			_case((Tree) that);
		}
		
		public void _case(TypeMatrix that) {
			_case((Tree) that);
		}
		
		public void _case(Command that) {
			_case((Tree) that);
		}
		
		public void _case(GlobalRef that) {
			_case((Tree) that);
		}
		
		public void _case(Collection that) {
			_case((Tree) that);
		}
		
		public void _case(ForIn that) {
			_case((Tree) that);
		}
		
		public void _case(Tree that) {
			throw new InternalError("unexpected: " + that);
		}
	}
    public static int count = 0;
}
